/*
 * Copyright 2014-2022 Thingsquare AB.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
* NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
* BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
* ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
*/

/**
 * Parts of this code (the HTML5 ajax code) is from Framework7 and is
 * licensed under the following license:
 *
 * Framework7 1.6.0
 * Full featured mobile HTML framework for building iOS & Android apps
 *
 * http://framework7.io/
 *
 * Copyright 2017, Vladimir Kharlampidi
 * The iDangero.us
 * http://www.idangero.us/
 *
 * Licensed under MIT
 *
 * Released on: April 10, 2017
 */

/* jslint nomen: true, node: true */
/* jslint esversion: 8, esnext: false */

'use strict';

var thsq;
var checkLoaded, fcmMessage, fcmToken, hostMessage, udpCallback, bleCallback, bleEnabled;

(function (root) {

    var nodejs = false;
    var browser = false;

    if (typeof require !== 'undefined') {
        var request = require('request');
        var jar = request.jar();
        var WebSocketClient = require('websocket').client;
        var crypto = require('crypto');
        var npm_pkg_version = require('../package.json').version;
        var fs = require('fs');

        nodejs = true;
    } else {
        browser = true;
    }

    /*jslint nomen: true, browser: true */
    /*global AndroidUDPSender, AndroidUDPListener, AndroidBLEListener,AndroidFCMService, XMLHttpRequest */

    function each(obj, callback) {
            // Check it's iterable
            // TODO: Should probably raise a value error here
            if (typeof obj !== 'object') return;
            // Don't bother continuing without a callback
            if (!callback) return;
            if (Array.isArray(obj)) {
                var i;
                // Array
                for (i = 0; i < obj.length; i++) {
                    // If callback returns false
                    if (callback(i, obj[i]) === false) {
                        // Break out of the loop
                        return;
                    }
                }
            }
            else {
                var prop;
                // Object
                for (prop in obj) {
                    // Check the propertie belongs to the object
                    // not it's prototype
                    if (obj.hasOwnProperty(prop)) {
                        // If the callback returns false
                        if (callback(prop, obj[prop]) === false) {
                            // Break out of the loop;
                            return;
                        }
                    }
                }
            }
    }
    function serializeObject(obj, parents) {
        if (typeof obj === 'string') return obj;
        var resultArray = [];
        var separator = '&';
        parents = parents || [];
        var newParents;
        function var_name(name) {
            if (parents.length > 0) {
                var _parents = '';
                for (var j = 0; j < parents.length; j++) {
                    if (j === 0) _parents += parents[j];
                    else _parents += '[' + encodeURIComponent(parents[j]) + ']';
                }
                return _parents + '[' + encodeURIComponent(name) + ']';
            }
            else {
                return encodeURIComponent(name);
            }
        }
        function var_value(value) {
            return encodeURIComponent(value);
        }
        for (var prop in obj) {
            if (obj.hasOwnProperty(prop)) {
                var toPush;
                if (Array.isArray(obj[prop])) {
                    toPush = [];
                    for (var i = 0; i < obj[prop].length; i ++) {
                        if (!Array.isArray(obj[prop][i]) && typeof obj[prop][i] === 'object') {
                            newParents = parents.slice();
                            newParents.push(prop);
                            newParents.push(i + '');
                            toPush.push(serializeObject(obj[prop][i], newParents));
                        }
                        else {
                            toPush.push(var_name(prop) + '[]=' + var_value(obj[prop][i]));
                        }

                    }
                    if (toPush.length > 0) resultArray.push(toPush.join(separator));
                }
                else if (obj[prop] === null) {
                    resultArray.push(var_name(prop) + '=');
                }
                else if (typeof obj[prop] === 'object') {
                        // Object, convert to named array
                        newParents = parents.slice();
                        newParents.push(prop);
                        toPush = serializeObject(obj[prop], newParents);
                        if (toPush !== '') resultArray.push(toPush);
                    }
                    else if (typeof obj[prop] !== 'undefined' && obj[prop] !== '') {
                        // Should be string or plain value
                        resultArray.push(var_name(prop) + '=' + var_value(obj[prop]));
                    }
                    else if (obj[prop] === '') resultArray.push(var_name(prop));
                }
            }
            return resultArray.join(separator);
        }
    var globalAjaxOptions = {};

    function ajaxNode(options) {
        var reqoptions, qs, form;

        if (options.method === 'GET') {
            qs = options.data;
        } else {
            if (options.formData) {
                qs = options.data;
            } else {
                form = options.data;
            }
        }

        let agent = 'thsq.js-' + npm_pkg_version;
        if (thsq.options.application) {
            agent += ' (' + thsq.options.application + ')';
        }

        reqoptions = {
            method: options.method,
            url: options.url,
            qs: qs,
            form: form,
            formData: options.formData,
            timeout: options.timeout,
            jar: jar,
            headers: {
                'User-Agent': agent
            }
        };
        request(reqoptions, function (error, response, body) {
            if (error || response.statusCode > 300) {
                if (typeof options.error === 'function') {
                    options.error(body, response);
                }
            } else {
                if (typeof options.success === 'function') {
                    options.success(body, response);
                }
            }
            if (typeof options.complete === 'function') {
                options.complete(body, response);
            }
        });
    }

    function ajaxWeb(options) {
            var defaults = {
                method: 'GET',
                data: false,
                async: true,
                cache: true,
                user: '',
                password: '',
                headers: {},
                xhrFields: {},
                statusCode: {},
                processData: true,
                dataType: 'text',
                contentType: 'application/x-www-form-urlencoded',
                timeout: 0
            };
            var callbacks = ['beforeSend', 'error', 'complete', 'success', 'statusCode'];


            //For jQuery guys
            if (options.type) options.method = options.type;

            // Merge global and defaults
            each(globalAjaxOptions, function (globalOptionName, globalOptionValue) {
                if (callbacks.indexOf(globalOptionName) < 0) defaults[globalOptionName] = globalOptionValue;
            });

            // Function to run XHR callbacks and events
            function fireAjaxCallback (eventName, eventData, callbackName) {
                var a = arguments;
                //if (eventName) $(document).trigger(eventName, eventData);
                if (callbackName) {
                    // Global callback
                    if (callbackName in globalAjaxOptions) globalAjaxOptions[callbackName](a[3], a[4], a[5], a[6]);
                    // Options callback
                    if (options[callbackName]) options[callbackName](a[3], a[4], a[5], a[6]);
                }
            }

            // Merge options and defaults
            each(defaults, function (prop, defaultValue) {
                if (!(prop in options)) options[prop] = defaultValue;
            });

            // Default URL
            if (!options.url) {
                options.url = window.location.toString();
            }
            // Parameters Prefix
            var paramsPrefix = options.url.indexOf('?') >= 0 ? '&' : '?';

            // UC method
            var _method = options.method.toUpperCase();
            // Data to modify GET URL
            if ((_method === 'GET' || _method === 'HEAD' || _method === 'OPTIONS' || _method === 'DELETE') && options.data) {
                var stringData;
                if (typeof options.data === 'string') {
                    // Should be key=value string
                    if (options.data.indexOf('?') >= 0) stringData = options.data.split('?')[1];
                    else stringData = options.data;
                }
                else {
                    // Should be key=value object
                    stringData = serializeObject(options.data);
                }
                if (stringData.length) {
                    options.url += paramsPrefix + stringData;
                    if (paramsPrefix === '?') paramsPrefix = '&';
                }
            }

            // Cache for GET/HEAD requests
            if (_method === 'GET' || _method === 'HEAD' || _method === 'OPTIONS' || _method === 'DELETE') {
                if (options.cache === false) {
                    options.url += (paramsPrefix + '_nocache=' + Date.now());
                }
            }

            // Create XHR
            var xhr = new XMLHttpRequest();

            // Save Request URL
            xhr.requestUrl = options.url;
            xhr.requestParameters = options;

            // Open XHR
            xhr.open(_method, options.url, options.async, options.user, options.password);

            // Create POST Data
            var postData = null;

            if ((_method === 'POST' || _method === 'PUT' || _method === 'PATCH') && options.data) {
                if (options.processData) {
                    var postDataInstances = [ArrayBuffer, Blob, Document, FormData];
                    // Post Data
                    if (postDataInstances.indexOf(options.data.constructor) >= 0) {
                        postData = options.data;
                    }
                    else {
                        // POST Headers
                        var boundary = '---------------------------' + Date.now().toString(16);

                        if (options.contentType === 'multipart\/form-data') {
                            xhr.setRequestHeader('Content-Type', 'multipart\/form-data; boundary=' + boundary);
                        }
                        else {
                            xhr.setRequestHeader('Content-Type', options.contentType);
                        }
                        postData = '';
                        var _data = serializeObject(options.data);
                        if (options.contentType === 'multipart\/form-data') {
                            boundary = '---------------------------' + Date.now().toString(16);
                            _data = _data.split('&');
                            var _newData = [];
                            for (var i = 0; i < _data.length; i++) {
                                _newData.push('Content-Disposition: form-data; name="' + _data[i].split('=')[0] + '"\r\n\r\n' + _data[i].split('=')[1] + '\r\n');
                            }
                            postData = '--' + boundary + '\r\n' + _newData.join('--' + boundary + '\r\n') + '--' + boundary + '--\r\n';
                        }
                        else {
                            postData = _data;
                        }
                    }
                }
                else {
                    postData = options.data;
                }

            }

            // Additional headers
            if (options.headers) {
                each(options.headers, function (headerName, headerCallback) {
                    xhr.setRequestHeader(headerName, headerCallback);
                });
            }

            // Check for crossDomain
            if (typeof options.crossDomain === 'undefined') {
                options.crossDomain = /^([\w-]+:)?\/\/([^\/]+)/.test(options.url) && RegExp.$2 !== window.location.host;
            }

            if (!options.crossDomain) {
                xhr.setRequestHeader('X-Requested-With', 'XMLHttpRequest');
            }

            if (options.xhrFields) {
                each(options.xhrFields, function (fieldName, fieldValue) {
                    xhr[fieldName] = fieldValue;
                });
            }

            var xhrTimeout;
            // Handle XHR
            xhr.onload = function (e) {
                if (xhrTimeout) clearTimeout(xhrTimeout);
                if ((xhr.status >= 200 && xhr.status < 300) || xhr.status === 0) {
                    var responseData;
                    if (options.dataType === 'json') {
                        try {
                            responseData = JSON.parse(xhr.responseText);
                            fireAjaxCallback('ajaxSuccess ajax:success', {xhr: xhr}, 'success', responseData, xhr.status, xhr);
                        }
                        catch (err) {
                            fireAjaxCallback('ajaxError ajax:error', {xhr: xhr, parseerror: true}, 'error', xhr, 'parseerror');
                        }
                    }
                    else {
                        responseData = xhr.responseType === 'text' || xhr.responseType === '' ? xhr.responseText : xhr.response;
                        fireAjaxCallback('ajaxSuccess ajax:success', {xhr: xhr}, 'success', responseData, xhr.status, xhr);
                    }
                }
                else {
                    fireAjaxCallback('ajaxError ajax:error', {xhr: xhr}, 'error', xhr, xhr.status);
                }
                if (options.statusCode) {
                    if (globalAjaxOptions.statusCode && globalAjaxOptions.statusCode[xhr.status]) globalAjaxOptions.statusCode[xhr.status](xhr);
                    if (options.statusCode[xhr.status]) options.statusCode[xhr.status](xhr);
                }
                fireAjaxCallback('ajaxComplete ajax:complete', {xhr: xhr}, 'complete', xhr, xhr.status);
            };

            xhr.onerror = function (e) {
                if (xhrTimeout) clearTimeout(xhrTimeout);
                fireAjaxCallback('ajaxError ajax:error', {xhr: xhr}, 'error', xhr, xhr.status);
                fireAjaxCallback('ajaxComplete ajax:complete', {xhr: xhr, error: true}, 'complete', xhr, 'error');
            };

            // Ajax start callback
            fireAjaxCallback('ajaxStart ajax:start', {xhr: xhr}, 'start', xhr);
            fireAjaxCallback(undefined, undefined, 'beforeSend', xhr);

            // Timeout
            if (options.timeout > 0) {
                xhr.onabort = function () {
                    if (xhrTimeout) clearTimeout(xhrTimeout);
                };
                xhrTimeout = setTimeout(function () {
                    xhr.abort();
                    fireAjaxCallback('ajaxError ajax:error', {xhr: xhr, timeout: true}, 'error', xhr, 'timeout');
                    fireAjaxCallback('ajaxComplete ajax:complete', {xhr: xhr, timeout: true}, 'complete', xhr, 'timeout');
                }, options.timeout);
            }

            // Send XHR
            xhr.send(postData);

            // Return XHR object
            return xhr;
        }

    var ajax;

    if (typeof XMLHttpRequest === 'undefined') {
        ajax = ajaxNode;
    } else {
        ajax = ajaxWeb;
    }

    function EventEmitter() {
        this.events = {};
    }

    EventEmitter.prototype.on = function (event, options, listener) {
        if (this.events[event] === undefined) {
            this.events[event] = [];
        }

        if (typeof options === 'function') {
            listener = options;
            options = {};
        }

        this.events[event].push(listener);

        if (event === 'device-variable-history') {
            throw new Error('obsolete device-variable-history event, see thsq.addHistoryListener()');
        }
    };

    EventEmitter.prototype.removeListener = function (event, listener) {
        var index;

        if (this.events[event] !== undefined) {
            index = this.events[event].indexOf(listener);

            if (index > -1) {
                this.events[event].splice(index, 1);
            }
        }
        return this;
    };

    EventEmitter.prototype.off = function (event, listener) {
        return this.removeListener(event, listener);
    };

    EventEmitter.prototype.removeAllListeners = function (event) {
        if (event) {
            delete this.events[event];
        } else {
            this.events = {};
        }
        return this;
    };

    EventEmitter.prototype.listeners = function (event) {
        if (this.events[event] !== undefined) {
            return this.events[event].slice();
        }
    };

    EventEmitter.prototype.emit = function (event) {
        var i, listeners, length, args;

        args = [].slice.call(arguments, 1);

        if (this.events[event] !== undefined) {
            listeners = this.events[event].slice();
            length = listeners.length;

            for (i = 0; i < length; i += 1) {
                listeners[i].apply(this, args);
            }
            return true;
        }
        return false;
    };

    EventEmitter.prototype.once = function (event, listener) {
        this.on(event, function callback() {
            this.removeListener(event, callback);
            listener.apply(this, arguments);
        });
    };

    EventEmitter.prototype.listenerCount = function (event) {
        if (this.events[event] === undefined) {
            return 0;
        }
        return this.events[event].length;
    };

    EventEmitter.prototype.eventNames = function (event) {
        return Object.keys(this.events);
    };
    /*
     * JavaScript MD5
     * https://github.com/blueimp/JavaScript-MD5
     *
     * Copyright 2011, Sebastian Tschan
     * https://blueimp.net
     *
     * Licensed under the MIT license:
     * https://opensource.org/licenses/MIT
     *
     * Based on
     * A JavaScript implementation of the RSA Data Security, Inc. MD5 Message
     * Digest Algorithm, as defined in RFC 1321.
     * Version 2.2 Copyright (C) Paul Johnston 1999 - 2009
     * Other contributors: Greg Holt, Andrew Kepert, Ydnar, Lostinet
     * Distributed under the BSD License
     * See http://pajhome.org.uk/crypt/md5 for more info.
     */

    var md5;
    (function() {

        /**
         * Add integers, wrapping at 2^32.
         * This uses 16-bit operations internally to work around bugs in interpreters.
         *
         * @param {number} x First integer
         * @param {number} y Second integer
         * @returns {number} Sum
         */
        function safeAdd(x, y) {
            var lsw = (x & 0xffff) + (y & 0xffff);
            var msw = (x >> 16) + (y >> 16) + (lsw >> 16);
            return (msw << 16) | (lsw & 0xffff);
        }

        /**
         * Bitwise rotate a 32-bit number to the left.
         *
         * @param {number} num 32-bit number
         * @param {number} cnt Rotation count
         * @returns {number} Rotated number
         */
        function bitRotateLeft(num, cnt) {
            return (num << cnt) | (num >>> (32 - cnt));
        }

        /**
         * Basic operation the algorithm uses.
         *
         * @param {number} q q
         * @param {number} a a
         * @param {number} b b
         * @param {number} x x
         * @param {number} s s
         * @param {number} t t
         * @returns {number} Result
         */
        function md5cmn(q, a, b, x, s, t) {
            return safeAdd(bitRotateLeft(safeAdd(safeAdd(a, q), safeAdd(x, t)), s), b);
        }
        /**
         * Basic operation the algorithm uses.
         *
         * @param {number} a a
         * @param {number} b b
         * @param {number} c c
         * @param {number} d d
         * @param {number} x x
         * @param {number} s s
         * @param {number} t t
         * @returns {number} Result
         */
        function md5ff(a, b, c, d, x, s, t) {
            return md5cmn((b & c) | (~b & d), a, b, x, s, t);
        }
        /**
         * Basic operation the algorithm uses.
         *
         * @param {number} a a
         * @param {number} b b
         * @param {number} c c
         * @param {number} d d
         * @param {number} x x
         * @param {number} s s
         * @param {number} t t
         * @returns {number} Result
         */
        function md5gg(a, b, c, d, x, s, t) {
            return md5cmn((b & d) | (c & ~d), a, b, x, s, t);
        }
        /**
         * Basic operation the algorithm uses.
         *
         * @param {number} a a
         * @param {number} b b
         * @param {number} c c
         * @param {number} d d
         * @param {number} x x
         * @param {number} s s
         * @param {number} t t
         * @returns {number} Result
         */
        function md5hh(a, b, c, d, x, s, t) {
            return md5cmn(b ^ c ^ d, a, b, x, s, t);
        }
        /**
         * Basic operation the algorithm uses.
         *
         * @param {number} a a
         * @param {number} b b
         * @param {number} c c
         * @param {number} d d
         * @param {number} x x
         * @param {number} s s
         * @param {number} t t
         * @returns {number} Result
         */
        function md5ii(a, b, c, d, x, s, t) {
            return md5cmn(c ^ (b | ~d), a, b, x, s, t);
        }

        /**
         * Calculate the MD5 of an array of little-endian words, and a bit length.
         *
         * @param {Array} x Array of little-endian words
         * @param {number} len Bit length
         * @returns {Array<number>} MD5 Array
         */
        function binlMD5(x, len) {
            /* append padding */
            x[len >> 5] |= 0x80 << len % 32;
            x[(((len + 64) >>> 9) << 4) + 14] = len;

            var i;
            var olda;
            var oldb;
            var oldc;
            var oldd;
            var a = 1732584193;
            var b = -271733879;
            var c = -1732584194;
            var d = 271733878;

            for (i = 0; i < x.length; i += 16) {
                olda = a;
                oldb = b;
                oldc = c;
                oldd = d;

                a = md5ff(a, b, c, d, x[i], 7, -680876936);
                d = md5ff(d, a, b, c, x[i + 1], 12, -389564586);
                c = md5ff(c, d, a, b, x[i + 2], 17, 606105819);
                b = md5ff(b, c, d, a, x[i + 3], 22, -1044525330);
                a = md5ff(a, b, c, d, x[i + 4], 7, -176418897);
                d = md5ff(d, a, b, c, x[i + 5], 12, 1200080426);
                c = md5ff(c, d, a, b, x[i + 6], 17, -1473231341);
                b = md5ff(b, c, d, a, x[i + 7], 22, -45705983);
                a = md5ff(a, b, c, d, x[i + 8], 7, 1770035416);
                d = md5ff(d, a, b, c, x[i + 9], 12, -1958414417);
                c = md5ff(c, d, a, b, x[i + 10], 17, -42063);
                b = md5ff(b, c, d, a, x[i + 11], 22, -1990404162);
                a = md5ff(a, b, c, d, x[i + 12], 7, 1804603682);
                d = md5ff(d, a, b, c, x[i + 13], 12, -40341101);
                c = md5ff(c, d, a, b, x[i + 14], 17, -1502002290);
                b = md5ff(b, c, d, a, x[i + 15], 22, 1236535329);

                a = md5gg(a, b, c, d, x[i + 1], 5, -165796510);
                d = md5gg(d, a, b, c, x[i + 6], 9, -1069501632);
                c = md5gg(c, d, a, b, x[i + 11], 14, 643717713);
                b = md5gg(b, c, d, a, x[i], 20, -373897302);
                a = md5gg(a, b, c, d, x[i + 5], 5, -701558691);
                d = md5gg(d, a, b, c, x[i + 10], 9, 38016083);
                c = md5gg(c, d, a, b, x[i + 15], 14, -660478335);
                b = md5gg(b, c, d, a, x[i + 4], 20, -405537848);
                a = md5gg(a, b, c, d, x[i + 9], 5, 568446438);
                d = md5gg(d, a, b, c, x[i + 14], 9, -1019803690);
                c = md5gg(c, d, a, b, x[i + 3], 14, -187363961);
                b = md5gg(b, c, d, a, x[i + 8], 20, 1163531501);
                a = md5gg(a, b, c, d, x[i + 13], 5, -1444681467);
                d = md5gg(d, a, b, c, x[i + 2], 9, -51403784);
                c = md5gg(c, d, a, b, x[i + 7], 14, 1735328473);
                b = md5gg(b, c, d, a, x[i + 12], 20, -1926607734);

                a = md5hh(a, b, c, d, x[i + 5], 4, -378558);
                d = md5hh(d, a, b, c, x[i + 8], 11, -2022574463);
                c = md5hh(c, d, a, b, x[i + 11], 16, 1839030562);
                b = md5hh(b, c, d, a, x[i + 14], 23, -35309556);
                a = md5hh(a, b, c, d, x[i + 1], 4, -1530992060);
                d = md5hh(d, a, b, c, x[i + 4], 11, 1272893353);
                c = md5hh(c, d, a, b, x[i + 7], 16, -155497632);
                b = md5hh(b, c, d, a, x[i + 10], 23, -1094730640);
                a = md5hh(a, b, c, d, x[i + 13], 4, 681279174);
                d = md5hh(d, a, b, c, x[i], 11, -358537222);
                c = md5hh(c, d, a, b, x[i + 3], 16, -722521979);
                b = md5hh(b, c, d, a, x[i + 6], 23, 76029189);
                a = md5hh(a, b, c, d, x[i + 9], 4, -640364487);
                d = md5hh(d, a, b, c, x[i + 12], 11, -421815835);
                c = md5hh(c, d, a, b, x[i + 15], 16, 530742520);
                b = md5hh(b, c, d, a, x[i + 2], 23, -995338651);

                a = md5ii(a, b, c, d, x[i], 6, -198630844);
                d = md5ii(d, a, b, c, x[i + 7], 10, 1126891415);
                c = md5ii(c, d, a, b, x[i + 14], 15, -1416354905);
                b = md5ii(b, c, d, a, x[i + 5], 21, -57434055);
                a = md5ii(a, b, c, d, x[i + 12], 6, 1700485571);
                d = md5ii(d, a, b, c, x[i + 3], 10, -1894986606);
                c = md5ii(c, d, a, b, x[i + 10], 15, -1051523);
                b = md5ii(b, c, d, a, x[i + 1], 21, -2054922799);
                a = md5ii(a, b, c, d, x[i + 8], 6, 1873313359);
                d = md5ii(d, a, b, c, x[i + 15], 10, -30611744);
                c = md5ii(c, d, a, b, x[i + 6], 15, -1560198380);
                b = md5ii(b, c, d, a, x[i + 13], 21, 1309151649);
                a = md5ii(a, b, c, d, x[i + 4], 6, -145523070);
                d = md5ii(d, a, b, c, x[i + 11], 10, -1120210379);
                c = md5ii(c, d, a, b, x[i + 2], 15, 718787259);
                b = md5ii(b, c, d, a, x[i + 9], 21, -343485551);

                a = safeAdd(a, olda);
                b = safeAdd(b, oldb);
                c = safeAdd(c, oldc);
                d = safeAdd(d, oldd);
            }
            return [a, b, c, d];
        }

        /**
         * Convert an array of little-endian words to a string
         *
         * @param {Array<number>} input MD5 Array
         * @returns {string} MD5 string
         */
        function binl2rstr(input) {
            var i;
            var output = '';
            var length32 = input.length * 32;
            for (i = 0; i < length32; i += 8) {
                output += String.fromCharCode((input[i >> 5] >>> i % 32) & 0xff);
            }
            return output;
        }

        /**
         * Convert a raw string to an array of little-endian words
         * Characters >255 have their high-byte silently ignored.
         *
         * @param {string} input Raw input string
         * @returns {Array<number>} Array of little-endian words
         */
        function rstr2binl(input) {
            var i;
            var output = [];
            output[(input.length >> 2) - 1] = undefined;
            for (i = 0; i < output.length; i += 1) {
                output[i] = 0;
            }
            var length8 = input.length * 8;
            for (i = 0; i < length8; i += 8) {
                output[i >> 5] |= (input.charCodeAt(i / 8) & 0xff) << i % 32;
            }
            return output;
        }

        /**
         * Calculate the MD5 of a raw string
         *
         * @param {string} s Input string
         * @returns {string} Raw MD5 string
         */
        function rstrMD5(s) {
            return binl2rstr(binlMD5(rstr2binl(s), s.length * 8));
        }

        /**
         * Calculates the HMAC-MD5 of a key and some data (raw strings)
         *
         * @param {string} key HMAC key
         * @param {string} data Raw input string
         * @returns {string} Raw MD5 string
         */
        function rstrHMACMD5(key, data) {
            var i;
            var bkey = rstr2binl(key);
            var ipad = [];
            var opad = [];
            var hash;
            ipad[15] = opad[15] = undefined;
            if (bkey.length > 16) {
                bkey = binlMD5(bkey, key.length * 8);
            }
            for (i = 0; i < 16; i += 1) {
                ipad[i] = bkey[i] ^ 0x36363636;
                opad[i] = bkey[i] ^ 0x5c5c5c5c;
            }
            hash = binlMD5(ipad.concat(rstr2binl(data)), 512 + data.length * 8);
            return binl2rstr(binlMD5(opad.concat(hash), 512 + 128));
        }

        /**
         * Convert a raw string to a hex string
         *
         * @param {string} input Raw input string
         * @returns {string} Hex encoded string
         */
        function rstr2hex(input) {
            var hexTab = '0123456789abcdef';
            var output = '';
            var x;
            var i;
            for (i = 0; i < input.length; i += 1) {
                x = input.charCodeAt(i);
                output += hexTab.charAt((x >>> 4) & 0x0f) + hexTab.charAt(x & 0x0f);
            }
            return output;
        }

        /**
         * Encode a string as UTF-8
         *
         * @param {string} input Input string
         * @returns {string} UTF8 string
         */
        function str2rstrUTF8(input) {
            return unescape(encodeURIComponent(input));
        }

        /**
         * Encodes input string as raw MD5 string
         *
         * @param {string} s Input string
         * @returns {string} Raw MD5 string
         */
        function rawMD5(s) {
            return rstrMD5(str2rstrUTF8(s));
        }
        /**
         * Encodes input string as Hex encoded string
         *
         * @param {string} s Input string
         * @returns {string} Hex encoded string
         */
        function hexMD5(s) {
            return rstr2hex(rawMD5(s));
        }
        /**
         * Calculates the raw HMAC-MD5 for the given key and data
         *
         * @param {string} k HMAC key
         * @param {string} d Input string
         * @returns {string} Raw MD5 string
         */
        function rawHMACMD5(k, d) {
        return rstrHMACMD5(str2rstrUTF8(k), str2rstrUTF8(d));
    }
    /**
     * Calculates the Hex encoded HMAC-MD5 for the given key and data
     *
     * @param {string} k HMAC key
     * @param {string} d Input string
     * @returns {string} Raw MD5 string
     */
    function hexHMACMD5(k, d) {
        return rstr2hex(rawHMACMD5(k, d));
    }

    /**
     * Calculates MD5 value for a given string.
     * If a key is provided, calculates the HMAC-MD5 value.
     * Returns a Hex encoded string unless the raw argument is given.
     *
     * @param {string} string Input string
     * @param {string} [key] HMAC key
     * @param {boolean} [raw] Raw output switch
     * @returns {string} MD5 output
     */
    function fmd5(string, key, raw) {
        if (!key) {
            if (!raw) {
                return hexMD5(string);
            }
            return rawMD5(string);
        }
        if (!raw) {
            return hexHMACMD5(key, string);
        }
        return rawHMACMD5(key, string);
    }

    md5 = fmd5;
    })();

    /*
     * [js-sha1]{@link https://github.com/emn178/js-sha1}
     *
     * @version 0.6.0
     * @author Chen, Yi-Cyuan [emn178@gmail.com]
     * @copyright Chen, Yi-Cyuan 2014-2017
     * @license MIT
     */
         var root = typeof window === 'object' ? window : {};
         var NODE_JS = !root.JS_SHA1_NO_NODE_JS && typeof process === 'object' && process.versions && process.versions.node;
         if (NODE_JS) {
            root = global;
        }
        var HEX_CHARS = '0123456789abcdef'.split('');
        var EXTRA = [-2147483648, 8388608, 32768, 128];
        var SHIFT = [24, 16, 8, 0];
        var OUTPUT_TYPES = ['hex', 'array', 'digest', 'arrayBuffer'];

        var blocks = [];

        var createOutputMethod = function (outputType) {
            return function (message) {
              return new Sha1(true).update(message)[outputType]();
          };
      };

      var createMethod = function () {
        var method = createOutputMethod('hex');
        if (NODE_JS) {
          method = nodeWrap(method);
      }
      method.create = function () {
          return new Sha1();
      };
      method.update = function (message) {
          return method.create().update(message);
      };
      for (var i = 0; i < OUTPUT_TYPES.length; ++i) {
          var type = OUTPUT_TYPES[i];
          method[type] = createOutputMethod(type);
      }
      return method;
    };

    var nodeWrap = function (method) {
        var crypto = eval("require('crypto')");
        var Buffer = eval("require('buffer').Buffer");
        var nodeMethod = function (message) {
          if (typeof message === 'string') {
            return crypto.createHash('sha1').update(message, 'utf8').digest('hex');
        } else if (message.constructor === ArrayBuffer) {
            message = new Uint8Array(message);
        } else if (message.length === undefined) {
            return method(message);
        }
        return crypto.createHash('sha1').update(Buffer.from(message)).digest('hex');
    };
    return nodeMethod;
    };

    function Sha1(sharedMemory) {
        if (sharedMemory) {
          blocks[0] = blocks[16] = blocks[1] = blocks[2] = blocks[3] =
          blocks[4] = blocks[5] = blocks[6] = blocks[7] =
          blocks[8] = blocks[9] = blocks[10] = blocks[11] =
          blocks[12] = blocks[13] = blocks[14] = blocks[15] = 0;
          this.blocks = blocks;
      } else {
          this.blocks = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
      }

      this.h0 = 0x67452301;
      this.h1 = 0xEFCDAB89;
      this.h2 = 0x98BADCFE;
      this.h3 = 0x10325476;
      this.h4 = 0xC3D2E1F0;

      this.block = this.start = this.bytes = this.hBytes = 0;
      this.finalized = this.hashed = false;
      this.first = true;
    }

    Sha1.prototype.update = function (message) {
        if (this.finalized) {
          return;
      }
      var notString = typeof(message) !== 'string';
      if (notString && message.constructor === root.ArrayBuffer) {
          message = new Uint8Array(message);
      }
      var code, index = 0, i, length = message.length || 0, blocks = this.blocks;

      while (index < length) {
          if (this.hashed) {
            this.hashed = false;
            blocks[0] = this.block;
            blocks[16] = blocks[1] = blocks[2] = blocks[3] =
            blocks[4] = blocks[5] = blocks[6] = blocks[7] =
            blocks[8] = blocks[9] = blocks[10] = blocks[11] =
            blocks[12] = blocks[13] = blocks[14] = blocks[15] = 0;
        }

        if(notString) {
            for (i = this.start; index < length && i < 64; ++index) {
              blocks[i >> 2] |= message[index] << SHIFT[i++ & 3];
          }
      } else {
        for (i = this.start; index < length && i < 64; ++index) {
          code = message.charCodeAt(index);
          if (code < 0x80) {
            blocks[i >> 2] |= code << SHIFT[i++ & 3];
        } else if (code < 0x800) {
            blocks[i >> 2] |= (0xc0 | (code >> 6)) << SHIFT[i++ & 3];
            blocks[i >> 2] |= (0x80 | (code & 0x3f)) << SHIFT[i++ & 3];
        } else if (code < 0xd800 || code >= 0xe000) {
            blocks[i >> 2] |= (0xe0 | (code >> 12)) << SHIFT[i++ & 3];
            blocks[i >> 2] |= (0x80 | ((code >> 6) & 0x3f)) << SHIFT[i++ & 3];
            blocks[i >> 2] |= (0x80 | (code & 0x3f)) << SHIFT[i++ & 3];
        } else {
            code = 0x10000 + (((code & 0x3ff) << 10) | (message.charCodeAt(++index) & 0x3ff));
            blocks[i >> 2] |= (0xf0 | (code >> 18)) << SHIFT[i++ & 3];
            blocks[i >> 2] |= (0x80 | ((code >> 12) & 0x3f)) << SHIFT[i++ & 3];
            blocks[i >> 2] |= (0x80 | ((code >> 6) & 0x3f)) << SHIFT[i++ & 3];
            blocks[i >> 2] |= (0x80 | (code & 0x3f)) << SHIFT[i++ & 3];
        }
    }
    }

    this.lastByteIndex = i;
    this.bytes += i - this.start;
    if (i >= 64) {
        this.block = blocks[16];
        this.start = i - 64;
        this.hash();
        this.hashed = true;
    } else {
        this.start = i;
    }
    }
    if (this.bytes > 4294967295) {
        this.hBytes += this.bytes / 4294967296 << 0;
        this.bytes = this.bytes % 4294967296;
    }
    return this;
    };

    Sha1.prototype.finalize = function () {
        if (this.finalized) {
          return;
      }
      this.finalized = true;
      var blocks = this.blocks, i = this.lastByteIndex;
      blocks[16] = this.block;
      blocks[i >> 2] |= EXTRA[i & 3];
      this.block = blocks[16];
      if (i >= 56) {
          if (!this.hashed) {
            this.hash();
        }
        blocks[0] = this.block;
        blocks[16] = blocks[1] = blocks[2] = blocks[3] =
        blocks[4] = blocks[5] = blocks[6] = blocks[7] =
        blocks[8] = blocks[9] = blocks[10] = blocks[11] =
        blocks[12] = blocks[13] = blocks[14] = blocks[15] = 0;
    }
    blocks[14] = this.hBytes << 3 | this.bytes >>> 29;
    blocks[15] = this.bytes << 3;
    this.hash();
    };

    Sha1.prototype.hash = function () {
        var a = this.h0, b = this.h1, c = this.h2, d = this.h3, e = this.h4;
        var f, j, t, blocks = this.blocks;

        for(j = 16; j < 80; ++j) {
          t = blocks[j - 3] ^ blocks[j - 8] ^ blocks[j - 14] ^ blocks[j - 16];
          blocks[j] =  (t << 1) | (t >>> 31);
      }

      for(j = 0; j < 20; j += 5) {
          f = (b & c) | ((~b) & d);
          t = (a << 5) | (a >>> 27);
          e = t + f + e + 1518500249 + blocks[j] << 0;
          b = (b << 30) | (b >>> 2);

          f = (a & b) | ((~a) & c);
          t = (e << 5) | (e >>> 27);
          d = t + f + d + 1518500249 + blocks[j + 1] << 0;
          a = (a << 30) | (a >>> 2);

          f = (e & a) | ((~e) & b);
          t = (d << 5) | (d >>> 27);
          c = t + f + c + 1518500249 + blocks[j + 2] << 0;
          e = (e << 30) | (e >>> 2);

          f = (d & e) | ((~d) & a);
          t = (c << 5) | (c >>> 27);
          b = t + f + b + 1518500249 + blocks[j + 3] << 0;
          d = (d << 30) | (d >>> 2);

          f = (c & d) | ((~c) & e);
          t = (b << 5) | (b >>> 27);
          a = t + f + a + 1518500249 + blocks[j + 4] << 0;
          c = (c << 30) | (c >>> 2);
      }

      for(; j < 40; j += 5) {
          f = b ^ c ^ d;
          t = (a << 5) | (a >>> 27);
          e = t + f + e + 1859775393 + blocks[j] << 0;
          b = (b << 30) | (b >>> 2);

          f = a ^ b ^ c;
          t = (e << 5) | (e >>> 27);
          d = t + f + d + 1859775393 + blocks[j + 1] << 0;
          a = (a << 30) | (a >>> 2);

          f = e ^ a ^ b;
          t = (d << 5) | (d >>> 27);
          c = t + f + c + 1859775393 + blocks[j + 2] << 0;
          e = (e << 30) | (e >>> 2);

          f = d ^ e ^ a;
          t = (c << 5) | (c >>> 27);
          b = t + f + b + 1859775393 + blocks[j + 3] << 0;
          d = (d << 30) | (d >>> 2);

          f = c ^ d ^ e;
          t = (b << 5) | (b >>> 27);
          a = t + f + a + 1859775393 + blocks[j + 4] << 0;
          c = (c << 30) | (c >>> 2);
      }

      for(; j < 60; j += 5) {
          f = (b & c) | (b & d) | (c & d);
          t = (a << 5) | (a >>> 27);
          e = t + f + e - 1894007588 + blocks[j] << 0;
          b = (b << 30) | (b >>> 2);

          f = (a & b) | (a & c) | (b & c);
          t = (e << 5) | (e >>> 27);
          d = t + f + d - 1894007588 + blocks[j + 1] << 0;
          a = (a << 30) | (a >>> 2);

          f = (e & a) | (e & b) | (a & b);
          t = (d << 5) | (d >>> 27);
          c = t + f + c - 1894007588 + blocks[j + 2] << 0;
          e = (e << 30) | (e >>> 2);

          f = (d & e) | (d & a) | (e & a);
          t = (c << 5) | (c >>> 27);
          b = t + f + b - 1894007588 + blocks[j + 3] << 0;
          d = (d << 30) | (d >>> 2);

          f = (c & d) | (c & e) | (d & e);
          t = (b << 5) | (b >>> 27);
          a = t + f + a - 1894007588 + blocks[j + 4] << 0;
          c = (c << 30) | (c >>> 2);
      }

      for(; j < 80; j += 5) {
          f = b ^ c ^ d;
          t = (a << 5) | (a >>> 27);
          e = t + f + e - 899497514 + blocks[j] << 0;
          b = (b << 30) | (b >>> 2);

          f = a ^ b ^ c;
          t = (e << 5) | (e >>> 27);
          d = t + f + d - 899497514 + blocks[j + 1] << 0;
          a = (a << 30) | (a >>> 2);

          f = e ^ a ^ b;
          t = (d << 5) | (d >>> 27);
          c = t + f + c - 899497514 + blocks[j + 2] << 0;
          e = (e << 30) | (e >>> 2);

          f = d ^ e ^ a;
          t = (c << 5) | (c >>> 27);
          b = t + f + b - 899497514 + blocks[j + 3] << 0;
          d = (d << 30) | (d >>> 2);

          f = c ^ d ^ e;
          t = (b << 5) | (b >>> 27);
          a = t + f + a - 899497514 + blocks[j + 4] << 0;
          c = (c << 30) | (c >>> 2);
      }

      this.h0 = this.h0 + a << 0;
      this.h1 = this.h1 + b << 0;
      this.h2 = this.h2 + c << 0;
      this.h3 = this.h3 + d << 0;
      this.h4 = this.h4 + e << 0;
    };

    Sha1.prototype.hex = function () {
        this.finalize();

        var h0 = this.h0, h1 = this.h1, h2 = this.h2, h3 = this.h3, h4 = this.h4;

        return HEX_CHARS[(h0 >> 28) & 0x0F] + HEX_CHARS[(h0 >> 24) & 0x0F] +
        HEX_CHARS[(h0 >> 20) & 0x0F] + HEX_CHARS[(h0 >> 16) & 0x0F] +
        HEX_CHARS[(h0 >> 12) & 0x0F] + HEX_CHARS[(h0 >> 8) & 0x0F] +
        HEX_CHARS[(h0 >> 4) & 0x0F] + HEX_CHARS[h0 & 0x0F] +
        HEX_CHARS[(h1 >> 28) & 0x0F] + HEX_CHARS[(h1 >> 24) & 0x0F] +
        HEX_CHARS[(h1 >> 20) & 0x0F] + HEX_CHARS[(h1 >> 16) & 0x0F] +
        HEX_CHARS[(h1 >> 12) & 0x0F] + HEX_CHARS[(h1 >> 8) & 0x0F] +
        HEX_CHARS[(h1 >> 4) & 0x0F] + HEX_CHARS[h1 & 0x0F] +
        HEX_CHARS[(h2 >> 28) & 0x0F] + HEX_CHARS[(h2 >> 24) & 0x0F] +
        HEX_CHARS[(h2 >> 20) & 0x0F] + HEX_CHARS[(h2 >> 16) & 0x0F] +
        HEX_CHARS[(h2 >> 12) & 0x0F] + HEX_CHARS[(h2 >> 8) & 0x0F] +
        HEX_CHARS[(h2 >> 4) & 0x0F] + HEX_CHARS[h2 & 0x0F] +
        HEX_CHARS[(h3 >> 28) & 0x0F] + HEX_CHARS[(h3 >> 24) & 0x0F] +
        HEX_CHARS[(h3 >> 20) & 0x0F] + HEX_CHARS[(h3 >> 16) & 0x0F] +
        HEX_CHARS[(h3 >> 12) & 0x0F] + HEX_CHARS[(h3 >> 8) & 0x0F] +
        HEX_CHARS[(h3 >> 4) & 0x0F] + HEX_CHARS[h3 & 0x0F] +
        HEX_CHARS[(h4 >> 28) & 0x0F] + HEX_CHARS[(h4 >> 24) & 0x0F] +
        HEX_CHARS[(h4 >> 20) & 0x0F] + HEX_CHARS[(h4 >> 16) & 0x0F] +
        HEX_CHARS[(h4 >> 12) & 0x0F] + HEX_CHARS[(h4 >> 8) & 0x0F] +
        HEX_CHARS[(h4 >> 4) & 0x0F] + HEX_CHARS[h4 & 0x0F];
    };

    Sha1.prototype.toString = Sha1.prototype.hex;

    Sha1.prototype.digest = function () {
        this.finalize();

        var h0 = this.h0, h1 = this.h1, h2 = this.h2, h3 = this.h3, h4 = this.h4;

        return [
        (h0 >> 24) & 0xFF, (h0 >> 16) & 0xFF, (h0 >> 8) & 0xFF, h0 & 0xFF,
        (h1 >> 24) & 0xFF, (h1 >> 16) & 0xFF, (h1 >> 8) & 0xFF, h1 & 0xFF,
        (h2 >> 24) & 0xFF, (h2 >> 16) & 0xFF, (h2 >> 8) & 0xFF, h2 & 0xFF,
        (h3 >> 24) & 0xFF, (h3 >> 16) & 0xFF, (h3 >> 8) & 0xFF, h3 & 0xFF,
        (h4 >> 24) & 0xFF, (h4 >> 16) & 0xFF, (h4 >> 8) & 0xFF, h4 & 0xFF
        ];
    };

    Sha1.prototype.array = Sha1.prototype.digest;

    Sha1.prototype.arrayBuffer = function () {
        this.finalize();

        var buffer = new ArrayBuffer(20);
        var dataView = new DataView(buffer);
        dataView.setUint32(0, this.h0);
        dataView.setUint32(4, this.h1);
        dataView.setUint32(8, this.h2);
        dataView.setUint32(12, this.h3);
        dataView.setUint32(16, this.h4);
        return buffer;
    };

    /**
     * [js-sha256]{@link https://github.com/emn178/js-sha256}
     *
     * @version 0.9.0
     * @author Chen, Yi-Cyuan [emn178@gmail.com]
     * @copyright Chen, Yi-Cyuan 2014-2017
     * @license MIT
     */
    (function() {
        var ERROR = 'input is invalid type';
        var WINDOW = typeof window === 'object';
        var root = WINDOW ? window : {};
        if (root.JS_SHA256_NO_WINDOW) {
            WINDOW = false;
        }
        var WEB_WORKER = !WINDOW && typeof self === 'object';
        var NODE_JS = !root.JS_SHA256_NO_NODE_JS && typeof process === 'object' && process.versions && process.versions.node;
        if (NODE_JS) {
            root = global;
        } else if (WEB_WORKER) {
            root = self;
        }
        var COMMON_JS = !root.JS_SHA256_NO_COMMON_JS && typeof module === 'object' && module.exports;
        var ARRAY_BUFFER = !root.JS_SHA256_NO_ARRAY_BUFFER && typeof ArrayBuffer !== 'undefined';
        var HEX_CHARS = '0123456789abcdef'.split('');
        var EXTRA = [-2147483648, 8388608, 32768, 128];
        var SHIFT = [24, 16, 8, 0];
        var K = [
            0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5, 0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
            0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3, 0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
            0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc, 0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
            0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7, 0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
            0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13, 0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
            0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3, 0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
            0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5, 0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
            0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208, 0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
        ];
        var OUTPUT_TYPES = ['hex', 'array', 'digest', 'arrayBuffer'];

        var blocks = [];

        if (root.JS_SHA256_NO_NODE_JS || !Array.isArray) {
            Array.isArray = function(obj) {
                return Object.prototype.toString.call(obj) === '[object Array]';
            };
        }

        if (ARRAY_BUFFER && (root.JS_SHA256_NO_ARRAY_BUFFER_IS_VIEW || !ArrayBuffer.isView)) {
            ArrayBuffer.isView = function(obj) {
                return typeof obj === 'object' && obj.buffer && obj.buffer.constructor === ArrayBuffer;
            };
        }

        var createOutputMethod = function(outputType, is224) {
            return function(message) {
                return new Sha256(is224, true).update(message)[outputType]();
            };
        };

        var createMethod = function(is224) {
            var method = createOutputMethod('hex', is224);
            if (NODE_JS) {
                method = nodeWrap(method, is224);
            }
            method.create = function() {
                return new Sha256(is224);
            };
            method.update = function(message) {
                return method.create().update(message);
            };
            for (var i = 0; i < OUTPUT_TYPES.length; ++i) {
                var type = OUTPUT_TYPES[i];
                method[type] = createOutputMethod(type, is224);
            }
            return method;
        };

        var nodeWrap = function(method, is224) {
            var crypto = eval("require('crypto')");
            var Buffer = eval("require('buffer').Buffer");
            var algorithm = is224 ? 'sha224' : 'sha256';
            var nodeMethod = function(message) {
                if (typeof message === 'string') {
                    return crypto.createHash(algorithm).update(message, 'utf8').digest('hex');
                } else {
                    if (message === null || message === undefined) {
                        throw new Error(ERROR);
                    } else if (message.constructor === ArrayBuffer) {
                        message = new Uint8Array(message);
                    }
                }
                if (Array.isArray(message) || ArrayBuffer.isView(message) ||
                    message.constructor === Buffer) {
                    return crypto.createHash(algorithm).update(Buffer.from(message)).digest('hex');
                } else {
                    return method(message);
                }
            };
            return nodeMethod;
        };

        var createHmacOutputMethod = function(outputType, is224) {
            return function(key, message) {
                return new HmacSha256(key, is224, true).update(message)[outputType]();
            };
        };

        var createHmacMethod = function(is224) {
            var method = createHmacOutputMethod('hex', is224);
            method.create = function(key) {
                return new HmacSha256(key, is224);
            };
            method.update = function(key, message) {
                return method.create(key).update(message);
            };
            for (var i = 0; i < OUTPUT_TYPES.length; ++i) {
                var type = OUTPUT_TYPES[i];
                method[type] = createHmacOutputMethod(type, is224);
            }
            return method;
        };

        function Sha256(is224, sharedMemory) {
            if (sharedMemory) {
                blocks[0] = blocks[16] = blocks[1] = blocks[2] = blocks[3] =
                    blocks[4] = blocks[5] = blocks[6] = blocks[7] =
                    blocks[8] = blocks[9] = blocks[10] = blocks[11] =
                    blocks[12] = blocks[13] = blocks[14] = blocks[15] = 0;
                this.blocks = blocks;
            } else {
                this.blocks = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0];
            }

            if (is224) {
                this.h0 = 0xc1059ed8;
                this.h1 = 0x367cd507;
                this.h2 = 0x3070dd17;
                this.h3 = 0xf70e5939;
                this.h4 = 0xffc00b31;
                this.h5 = 0x68581511;
                this.h6 = 0x64f98fa7;
                this.h7 = 0xbefa4fa4;
            } else { // 256
                this.h0 = 0x6a09e667;
                this.h1 = 0xbb67ae85;
                this.h2 = 0x3c6ef372;
                this.h3 = 0xa54ff53a;
                this.h4 = 0x510e527f;
                this.h5 = 0x9b05688c;
                this.h6 = 0x1f83d9ab;
                this.h7 = 0x5be0cd19;
            }

            this.block = this.start = this.bytes = this.hBytes = 0;
            this.finalized = this.hashed = false;
            this.first = true;
            this.is224 = is224;
        }

        Sha256.prototype.update = function(message) {
            if (this.finalized) {
                return;
            }
            var notString, type = typeof message;
            if (type !== 'string') {
                if (type === 'object') {
                    if (message === null) {
                        throw new Error(ERROR);
                    } else if (ARRAY_BUFFER && message.constructor === ArrayBuffer) {
                        message = new Uint8Array(message);
                    } else if (!Array.isArray(message)) {
                        if (!ARRAY_BUFFER || !ArrayBuffer.isView(message)) {
                            throw new Error(ERROR);
                        }
                    }
                } else {
                    throw new Error(ERROR);
                }
                notString = true;
            }
            var code, index = 0,
                i, length = message.length,
                blocks = this.blocks;

            while (index < length) {
                if (this.hashed) {
                    this.hashed = false;
                    blocks[0] = this.block;
                    blocks[16] = blocks[1] = blocks[2] = blocks[3] =
                        blocks[4] = blocks[5] = blocks[6] = blocks[7] =
                        blocks[8] = blocks[9] = blocks[10] = blocks[11] =
                        blocks[12] = blocks[13] = blocks[14] = blocks[15] = 0;
                }

                if (notString) {
                    for (i = this.start; index < length && i < 64; ++index) {
                        blocks[i >> 2] |= message[index] << SHIFT[i++ & 3];
                    }
                } else {
                    for (i = this.start; index < length && i < 64; ++index) {
                        code = message.charCodeAt(index);
                        if (code < 0x80) {
                            blocks[i >> 2] |= code << SHIFT[i++ & 3];
                        } else if (code < 0x800) {
                            blocks[i >> 2] |= (0xc0 | (code >> 6)) << SHIFT[i++ & 3];
                            blocks[i >> 2] |= (0x80 | (code & 0x3f)) << SHIFT[i++ & 3];
                        } else if (code < 0xd800 || code >= 0xe000) {
                            blocks[i >> 2] |= (0xe0 | (code >> 12)) << SHIFT[i++ & 3];
                            blocks[i >> 2] |= (0x80 | ((code >> 6) & 0x3f)) << SHIFT[i++ & 3];
                            blocks[i >> 2] |= (0x80 | (code & 0x3f)) << SHIFT[i++ & 3];
                        } else {
                            code = 0x10000 + (((code & 0x3ff) << 10) | (message.charCodeAt(++index) & 0x3ff));
                            blocks[i >> 2] |= (0xf0 | (code >> 18)) << SHIFT[i++ & 3];
                            blocks[i >> 2] |= (0x80 | ((code >> 12) & 0x3f)) << SHIFT[i++ & 3];
                            blocks[i >> 2] |= (0x80 | ((code >> 6) & 0x3f)) << SHIFT[i++ & 3];
                            blocks[i >> 2] |= (0x80 | (code & 0x3f)) << SHIFT[i++ & 3];
                        }
                    }
                }

                this.lastByteIndex = i;
                this.bytes += i - this.start;
                if (i >= 64) {
                    this.block = blocks[16];
                    this.start = i - 64;
                    this.hash();
                    this.hashed = true;
                } else {
                    this.start = i;
                }
            }
            if (this.bytes > 4294967295) {
                this.hBytes += this.bytes / 4294967296 << 0;
                this.bytes = this.bytes % 4294967296;
            }
            return this;
        };

        Sha256.prototype.finalize = function() {
            if (this.finalized) {
                return;
            }
            this.finalized = true;
            var blocks = this.blocks,
                i = this.lastByteIndex;
            blocks[16] = this.block;
            blocks[i >> 2] |= EXTRA[i & 3];
            this.block = blocks[16];
            if (i >= 56) {
                if (!this.hashed) {
                    this.hash();
                }
                blocks[0] = this.block;
                blocks[16] = blocks[1] = blocks[2] = blocks[3] =
                    blocks[4] = blocks[5] = blocks[6] = blocks[7] =
                    blocks[8] = blocks[9] = blocks[10] = blocks[11] =
                    blocks[12] = blocks[13] = blocks[14] = blocks[15] = 0;
            }
            blocks[14] = this.hBytes << 3 | this.bytes >>> 29;
            blocks[15] = this.bytes << 3;
            this.hash();
        };

        Sha256.prototype.hash = function() {
            var a = this.h0,
                b = this.h1,
                c = this.h2,
                d = this.h3,
                e = this.h4,
                f = this.h5,
                g = this.h6,
                h = this.h7,
                blocks = this.blocks,
                j, s0, s1, maj, t1, t2, ch, ab, da, cd, bc;

            for (j = 16; j < 64; ++j) {
                // rightrotate
                t1 = blocks[j - 15];
                s0 = ((t1 >>> 7) | (t1 << 25)) ^ ((t1 >>> 18) | (t1 << 14)) ^ (t1 >>> 3);
                t1 = blocks[j - 2];
                s1 = ((t1 >>> 17) | (t1 << 15)) ^ ((t1 >>> 19) | (t1 << 13)) ^ (t1 >>> 10);
                blocks[j] = blocks[j - 16] + s0 + blocks[j - 7] + s1 << 0;
            }

            bc = b & c;
            for (j = 0; j < 64; j += 4) {
                if (this.first) {
                    if (this.is224) {
                        ab = 300032;
                        t1 = blocks[0] - 1413257819;
                        h = t1 - 150054599 << 0;
                        d = t1 + 24177077 << 0;
                    } else {
                        ab = 704751109;
                        t1 = blocks[0] - 210244248;
                        h = t1 - 1521486534 << 0;
                        d = t1 + 143694565 << 0;
                    }
                    this.first = false;
                } else {
                    s0 = ((a >>> 2) | (a << 30)) ^ ((a >>> 13) | (a << 19)) ^ ((a >>> 22) | (a << 10));
                    s1 = ((e >>> 6) | (e << 26)) ^ ((e >>> 11) | (e << 21)) ^ ((e >>> 25) | (e << 7));
                    ab = a & b;
                    maj = ab ^ (a & c) ^ bc;
                    ch = (e & f) ^ (~e & g);
                    t1 = h + s1 + ch + K[j] + blocks[j];
                    t2 = s0 + maj;
                    h = d + t1 << 0;
                    d = t1 + t2 << 0;
                }
                s0 = ((d >>> 2) | (d << 30)) ^ ((d >>> 13) | (d << 19)) ^ ((d >>> 22) | (d << 10));
                s1 = ((h >>> 6) | (h << 26)) ^ ((h >>> 11) | (h << 21)) ^ ((h >>> 25) | (h << 7));
                da = d & a;
                maj = da ^ (d & b) ^ ab;
                ch = (h & e) ^ (~h & f);
                t1 = g + s1 + ch + K[j + 1] + blocks[j + 1];
                t2 = s0 + maj;
                g = c + t1 << 0;
                c = t1 + t2 << 0;
                s0 = ((c >>> 2) | (c << 30)) ^ ((c >>> 13) | (c << 19)) ^ ((c >>> 22) | (c << 10));
                s1 = ((g >>> 6) | (g << 26)) ^ ((g >>> 11) | (g << 21)) ^ ((g >>> 25) | (g << 7));
                cd = c & d;
                maj = cd ^ (c & a) ^ da;
                ch = (g & h) ^ (~g & e);
                t1 = f + s1 + ch + K[j + 2] + blocks[j + 2];
                t2 = s0 + maj;
                f = b + t1 << 0;
                b = t1 + t2 << 0;
                s0 = ((b >>> 2) | (b << 30)) ^ ((b >>> 13) | (b << 19)) ^ ((b >>> 22) | (b << 10));
                s1 = ((f >>> 6) | (f << 26)) ^ ((f >>> 11) | (f << 21)) ^ ((f >>> 25) | (f << 7));
                bc = b & c;
                maj = bc ^ (b & d) ^ cd;
                ch = (f & g) ^ (~f & h);
                t1 = e + s1 + ch + K[j + 3] + blocks[j + 3];
                t2 = s0 + maj;
                e = a + t1 << 0;
                a = t1 + t2 << 0;
            }

            this.h0 = this.h0 + a << 0;
            this.h1 = this.h1 + b << 0;
            this.h2 = this.h2 + c << 0;
            this.h3 = this.h3 + d << 0;
            this.h4 = this.h4 + e << 0;
            this.h5 = this.h5 + f << 0;
            this.h6 = this.h6 + g << 0;
            this.h7 = this.h7 + h << 0;
        };

        Sha256.prototype.hex = function() {
            this.finalize();

            var h0 = this.h0,
                h1 = this.h1,
                h2 = this.h2,
                h3 = this.h3,
                h4 = this.h4,
                h5 = this.h5,
                h6 = this.h6,
                h7 = this.h7;

            var hex = HEX_CHARS[(h0 >> 28) & 0x0F] + HEX_CHARS[(h0 >> 24) & 0x0F] +
                HEX_CHARS[(h0 >> 20) & 0x0F] + HEX_CHARS[(h0 >> 16) & 0x0F] +
                HEX_CHARS[(h0 >> 12) & 0x0F] + HEX_CHARS[(h0 >> 8) & 0x0F] +
                HEX_CHARS[(h0 >> 4) & 0x0F] + HEX_CHARS[h0 & 0x0F] +
                HEX_CHARS[(h1 >> 28) & 0x0F] + HEX_CHARS[(h1 >> 24) & 0x0F] +
                HEX_CHARS[(h1 >> 20) & 0x0F] + HEX_CHARS[(h1 >> 16) & 0x0F] +
                HEX_CHARS[(h1 >> 12) & 0x0F] + HEX_CHARS[(h1 >> 8) & 0x0F] +
                HEX_CHARS[(h1 >> 4) & 0x0F] + HEX_CHARS[h1 & 0x0F] +
                HEX_CHARS[(h2 >> 28) & 0x0F] + HEX_CHARS[(h2 >> 24) & 0x0F] +
                HEX_CHARS[(h2 >> 20) & 0x0F] + HEX_CHARS[(h2 >> 16) & 0x0F] +
                HEX_CHARS[(h2 >> 12) & 0x0F] + HEX_CHARS[(h2 >> 8) & 0x0F] +
                HEX_CHARS[(h2 >> 4) & 0x0F] + HEX_CHARS[h2 & 0x0F] +
                HEX_CHARS[(h3 >> 28) & 0x0F] + HEX_CHARS[(h3 >> 24) & 0x0F] +
                HEX_CHARS[(h3 >> 20) & 0x0F] + HEX_CHARS[(h3 >> 16) & 0x0F] +
                HEX_CHARS[(h3 >> 12) & 0x0F] + HEX_CHARS[(h3 >> 8) & 0x0F] +
                HEX_CHARS[(h3 >> 4) & 0x0F] + HEX_CHARS[h3 & 0x0F] +
                HEX_CHARS[(h4 >> 28) & 0x0F] + HEX_CHARS[(h4 >> 24) & 0x0F] +
                HEX_CHARS[(h4 >> 20) & 0x0F] + HEX_CHARS[(h4 >> 16) & 0x0F] +
                HEX_CHARS[(h4 >> 12) & 0x0F] + HEX_CHARS[(h4 >> 8) & 0x0F] +
                HEX_CHARS[(h4 >> 4) & 0x0F] + HEX_CHARS[h4 & 0x0F] +
                HEX_CHARS[(h5 >> 28) & 0x0F] + HEX_CHARS[(h5 >> 24) & 0x0F] +
                HEX_CHARS[(h5 >> 20) & 0x0F] + HEX_CHARS[(h5 >> 16) & 0x0F] +
                HEX_CHARS[(h5 >> 12) & 0x0F] + HEX_CHARS[(h5 >> 8) & 0x0F] +
                HEX_CHARS[(h5 >> 4) & 0x0F] + HEX_CHARS[h5 & 0x0F] +
                HEX_CHARS[(h6 >> 28) & 0x0F] + HEX_CHARS[(h6 >> 24) & 0x0F] +
                HEX_CHARS[(h6 >> 20) & 0x0F] + HEX_CHARS[(h6 >> 16) & 0x0F] +
                HEX_CHARS[(h6 >> 12) & 0x0F] + HEX_CHARS[(h6 >> 8) & 0x0F] +
                HEX_CHARS[(h6 >> 4) & 0x0F] + HEX_CHARS[h6 & 0x0F];
            if (!this.is224) {
                hex += HEX_CHARS[(h7 >> 28) & 0x0F] + HEX_CHARS[(h7 >> 24) & 0x0F] +
                    HEX_CHARS[(h7 >> 20) & 0x0F] + HEX_CHARS[(h7 >> 16) & 0x0F] +
                    HEX_CHARS[(h7 >> 12) & 0x0F] + HEX_CHARS[(h7 >> 8) & 0x0F] +
                    HEX_CHARS[(h7 >> 4) & 0x0F] + HEX_CHARS[h7 & 0x0F];
            }
            return hex;
        };

        Sha256.prototype.toString = Sha256.prototype.hex;

        Sha256.prototype.digest = function() {
            this.finalize();

            var h0 = this.h0,
                h1 = this.h1,
                h2 = this.h2,
                h3 = this.h3,
                h4 = this.h4,
                h5 = this.h5,
                h6 = this.h6,
                h7 = this.h7;

            var arr = [
                (h0 >> 24) & 0xFF, (h0 >> 16) & 0xFF, (h0 >> 8) & 0xFF, h0 & 0xFF,
                (h1 >> 24) & 0xFF, (h1 >> 16) & 0xFF, (h1 >> 8) & 0xFF, h1 & 0xFF,
                (h2 >> 24) & 0xFF, (h2 >> 16) & 0xFF, (h2 >> 8) & 0xFF, h2 & 0xFF,
                (h3 >> 24) & 0xFF, (h3 >> 16) & 0xFF, (h3 >> 8) & 0xFF, h3 & 0xFF,
                (h4 >> 24) & 0xFF, (h4 >> 16) & 0xFF, (h4 >> 8) & 0xFF, h4 & 0xFF,
                (h5 >> 24) & 0xFF, (h5 >> 16) & 0xFF, (h5 >> 8) & 0xFF, h5 & 0xFF,
                (h6 >> 24) & 0xFF, (h6 >> 16) & 0xFF, (h6 >> 8) & 0xFF, h6 & 0xFF
            ];
            if (!this.is224) {
                arr.push((h7 >> 24) & 0xFF, (h7 >> 16) & 0xFF, (h7 >> 8) & 0xFF, h7 & 0xFF);
            }
            return arr;
        };

        Sha256.prototype.array = Sha256.prototype.digest;

        Sha256.prototype.arrayBuffer = function() {
            this.finalize();

            var buffer = new ArrayBuffer(this.is224 ? 28 : 32);
            var dataView = new DataView(buffer);
            dataView.setUint32(0, this.h0);
            dataView.setUint32(4, this.h1);
            dataView.setUint32(8, this.h2);
            dataView.setUint32(12, this.h3);
            dataView.setUint32(16, this.h4);
            dataView.setUint32(20, this.h5);
            dataView.setUint32(24, this.h6);
            if (!this.is224) {
                dataView.setUint32(28, this.h7);
            }
            return buffer;
        };

        function HmacSha256(key, is224, sharedMemory) {
            var i, type = typeof key;
            if (type === 'string') {
                var bytes = [],
                    length = key.length,
                    index = 0,
                    code;
                for (i = 0; i < length; ++i) {
                    code = key.charCodeAt(i);
                    if (code < 0x80) {
                        bytes[index++] = code;
                    } else if (code < 0x800) {
                        bytes[index++] = (0xc0 | (code >> 6));
                        bytes[index++] = (0x80 | (code & 0x3f));
                    } else if (code < 0xd800 || code >= 0xe000) {
                        bytes[index++] = (0xe0 | (code >> 12));
                        bytes[index++] = (0x80 | ((code >> 6) & 0x3f));
                        bytes[index++] = (0x80 | (code & 0x3f));
                    } else {
                        code = 0x10000 + (((code & 0x3ff) << 10) | (key.charCodeAt(++i) & 0x3ff));
                        bytes[index++] = (0xf0 | (code >> 18));
                        bytes[index++] = (0x80 | ((code >> 12) & 0x3f));
                        bytes[index++] = (0x80 | ((code >> 6) & 0x3f));
                        bytes[index++] = (0x80 | (code & 0x3f));
                    }
                }
                key = bytes;
            } else {
                if (type === 'object') {
                    if (key === null) {
                        throw new Error(ERROR);
                    } else if (ARRAY_BUFFER && key.constructor === ArrayBuffer) {
                        key = new Uint8Array(key);
                    } else if (!Array.isArray(key)) {
                        if (!ARRAY_BUFFER || !ArrayBuffer.isView(key)) {
                            throw new Error(ERROR);
                        }
                    }
                } else {
                    throw new Error(ERROR);
                }
            }

            if (key.length > 64) {
                key = (new Sha256(is224, true)).update(key).array();
            }

            var oKeyPad = [],
                iKeyPad = [];
            for (i = 0; i < 64; ++i) {
                var b = key[i] || 0;
                oKeyPad[i] = 0x5c ^ b;
                iKeyPad[i] = 0x36 ^ b;
            }

            Sha256.call(this, is224, sharedMemory);

            this.update(iKeyPad);
            this.oKeyPad = oKeyPad;
            this.inner = true;
            this.sharedMemory = sharedMemory;
        }
        HmacSha256.prototype = new Sha256();

        HmacSha256.prototype.finalize = function() {
            Sha256.prototype.finalize.call(this);
            if (this.inner) {
                this.inner = false;
                var innerHash = this.array();
                Sha256.call(this, this.is224, this.sharedMemory);
                this.update(this.oKeyPad);
                this.update(innerHash);
                Sha256.prototype.finalize.call(this);
            }
        };

        var exports = createMethod();
        exports.sha256 = exports;
        exports.sha224 = createMethod(true);
        exports.sha256.hmac = createHmacMethod();
        exports.sha224.hmac = createHmacMethod(true);

        root.sha256 = exports.sha256;
        root.sha224 = exports.sha224;
    })();


    /* aes-js.js: MIT License. Copyright 2015-2018 Richard Moore <me@ricmoo.com>. See LICENSE.txt. */
    (function(root) {

        function checkInt(value) {
            return (parseInt(value) === value);
        }

        function checkInts(arrayish) {
            if (!checkInt(arrayish.length)) { return false; }

            for (var i = 0; i < arrayish.length; i++) {
                if (!checkInt(arrayish[i]) || arrayish[i] < 0 || arrayish[i] > 255) {
                    return false;
                }
            }

            return true;
        }

        function coerceArray(arg, copy) {

            // ArrayBuffer view
            if (arg.buffer && ArrayBuffer.isView(arg) && arg.name === 'Uint8Array') {

                if (copy) {
                    if (arg.slice) {
                        arg = arg.slice();
                    } else {
                        arg = Array.prototype.slice.call(arg);
                    }
                }

                return arg;
            }

            // It's an array; check it is a valid representation of a byte
            if (Array.isArray(arg)) {
                if (!checkInts(arg)) {
                    throw new Error('Array contains invalid value: ' + arg);
                }

                return new Uint8Array(arg);
            }

            // Something else, but behaves like an array (maybe a Buffer? Arguments?)
            if (checkInt(arg.length) && checkInts(arg)) {
                return new Uint8Array(arg);
            }

            throw new Error('unsupported array-like object');
        }

        function createArray(length) {
            return new Uint8Array(length);
        }

        function copyArray(sourceArray, targetArray, targetStart, sourceStart, sourceEnd) {
            if (sourceStart != null || sourceEnd != null) {
                if (sourceArray.slice) {
                    sourceArray = sourceArray.slice(sourceStart, sourceEnd);
                } else {
                    sourceArray = Array.prototype.slice.call(sourceArray, sourceStart, sourceEnd);
                }
            }
            targetArray.set(sourceArray, targetStart);
        }



        var convertUtf8 = (function() {
            function toBytes(text) {
                var result = [], i = 0;
                text = encodeURI(text);
                while (i < text.length) {
                    var c = text.charCodeAt(i++);

                    // if it is a % sign, encode the following 2 bytes as a hex value
                    if (c === 37) {
                        result.push(parseInt(text.substr(i, 2), 16))
                        i += 2;

                    // otherwise, just the actual byte
                    } else {
                        result.push(c)
                    }
                }

                return coerceArray(result);
            }

            function fromBytes(bytes) {
                var result = [], i = 0;

                while (i < bytes.length) {
                    var c = bytes[i];

                    if (c < 128) {
                        result.push(String.fromCharCode(c));
                        i++;
                    } else if (c > 191 && c < 224) {
                        result.push(String.fromCharCode(((c & 0x1f) << 6) | (bytes[i + 1] & 0x3f)));
                        i += 2;
                    } else {
                        result.push(String.fromCharCode(((c & 0x0f) << 12) | ((bytes[i + 1] & 0x3f) << 6) | (bytes[i + 2] & 0x3f)));
                        i += 3;
                    }
                }

                return result.join('');
            }

            return {
                toBytes: toBytes,
                fromBytes: fromBytes,
            }
        })();

        var convertHex = (function() {
            function toBytes(text) {
                var result = [];
                for (var i = 0; i < text.length; i += 2) {
                    result.push(parseInt(text.substr(i, 2), 16));
                }

                return result;
            }

            // http://ixti.net/development/javascript/2011/11/11/base64-encodedecode-of-utf8-in-browser-with-js.html
            var Hex = '0123456789abcdef';

            function fromBytes(bytes) {
                    var result = [];
                    for (var i = 0; i < bytes.length; i++) {
                        var v = bytes[i];
                        result.push(Hex[(v & 0xf0) >> 4] + Hex[v & 0x0f]);
                    }
                    return result.join('');
            }

            return {
                toBytes: toBytes,
                fromBytes: fromBytes,
            }
        })();


        // Number of rounds by keysize
        var numberOfRounds = {16: 10, 24: 12, 32: 14}

        // Round constant words
        var rcon = [0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36, 0x6c, 0xd8, 0xab, 0x4d, 0x9a, 0x2f, 0x5e, 0xbc, 0x63, 0xc6, 0x97, 0x35, 0x6a, 0xd4, 0xb3, 0x7d, 0xfa, 0xef, 0xc5, 0x91];

        // S-box and Inverse S-box (S is for Substitution)
        var S = [0x63, 0x7c, 0x77, 0x7b, 0xf2, 0x6b, 0x6f, 0xc5, 0x30, 0x01, 0x67, 0x2b, 0xfe, 0xd7, 0xab, 0x76, 0xca, 0x82, 0xc9, 0x7d, 0xfa, 0x59, 0x47, 0xf0, 0xad, 0xd4, 0xa2, 0xaf, 0x9c, 0xa4, 0x72, 0xc0, 0xb7, 0xfd, 0x93, 0x26, 0x36, 0x3f, 0xf7, 0xcc, 0x34, 0xa5, 0xe5, 0xf1, 0x71, 0xd8, 0x31, 0x15, 0x04, 0xc7, 0x23, 0xc3, 0x18, 0x96, 0x05, 0x9a, 0x07, 0x12, 0x80, 0xe2, 0xeb, 0x27, 0xb2, 0x75, 0x09, 0x83, 0x2c, 0x1a, 0x1b, 0x6e, 0x5a, 0xa0, 0x52, 0x3b, 0xd6, 0xb3, 0x29, 0xe3, 0x2f, 0x84, 0x53, 0xd1, 0x00, 0xed, 0x20, 0xfc, 0xb1, 0x5b, 0x6a, 0xcb, 0xbe, 0x39, 0x4a, 0x4c, 0x58, 0xcf, 0xd0, 0xef, 0xaa, 0xfb, 0x43, 0x4d, 0x33, 0x85, 0x45, 0xf9, 0x02, 0x7f, 0x50, 0x3c, 0x9f, 0xa8, 0x51, 0xa3, 0x40, 0x8f, 0x92, 0x9d, 0x38, 0xf5, 0xbc, 0xb6, 0xda, 0x21, 0x10, 0xff, 0xf3, 0xd2, 0xcd, 0x0c, 0x13, 0xec, 0x5f, 0x97, 0x44, 0x17, 0xc4, 0xa7, 0x7e, 0x3d, 0x64, 0x5d, 0x19, 0x73, 0x60, 0x81, 0x4f, 0xdc, 0x22, 0x2a, 0x90, 0x88, 0x46, 0xee, 0xb8, 0x14, 0xde, 0x5e, 0x0b, 0xdb, 0xe0, 0x32, 0x3a, 0x0a, 0x49, 0x06, 0x24, 0x5c, 0xc2, 0xd3, 0xac, 0x62, 0x91, 0x95, 0xe4, 0x79, 0xe7, 0xc8, 0x37, 0x6d, 0x8d, 0xd5, 0x4e, 0xa9, 0x6c, 0x56, 0xf4, 0xea, 0x65, 0x7a, 0xae, 0x08, 0xba, 0x78, 0x25, 0x2e, 0x1c, 0xa6, 0xb4, 0xc6, 0xe8, 0xdd, 0x74, 0x1f, 0x4b, 0xbd, 0x8b, 0x8a, 0x70, 0x3e, 0xb5, 0x66, 0x48, 0x03, 0xf6, 0x0e, 0x61, 0x35, 0x57, 0xb9, 0x86, 0xc1, 0x1d, 0x9e, 0xe1, 0xf8, 0x98, 0x11, 0x69, 0xd9, 0x8e, 0x94, 0x9b, 0x1e, 0x87, 0xe9, 0xce, 0x55, 0x28, 0xdf, 0x8c, 0xa1, 0x89, 0x0d, 0xbf, 0xe6, 0x42, 0x68, 0x41, 0x99, 0x2d, 0x0f, 0xb0, 0x54, 0xbb, 0x16];
        var Si =[0x52, 0x09, 0x6a, 0xd5, 0x30, 0x36, 0xa5, 0x38, 0xbf, 0x40, 0xa3, 0x9e, 0x81, 0xf3, 0xd7, 0xfb, 0x7c, 0xe3, 0x39, 0x82, 0x9b, 0x2f, 0xff, 0x87, 0x34, 0x8e, 0x43, 0x44, 0xc4, 0xde, 0xe9, 0xcb, 0x54, 0x7b, 0x94, 0x32, 0xa6, 0xc2, 0x23, 0x3d, 0xee, 0x4c, 0x95, 0x0b, 0x42, 0xfa, 0xc3, 0x4e, 0x08, 0x2e, 0xa1, 0x66, 0x28, 0xd9, 0x24, 0xb2, 0x76, 0x5b, 0xa2, 0x49, 0x6d, 0x8b, 0xd1, 0x25, 0x72, 0xf8, 0xf6, 0x64, 0x86, 0x68, 0x98, 0x16, 0xd4, 0xa4, 0x5c, 0xcc, 0x5d, 0x65, 0xb6, 0x92, 0x6c, 0x70, 0x48, 0x50, 0xfd, 0xed, 0xb9, 0xda, 0x5e, 0x15, 0x46, 0x57, 0xa7, 0x8d, 0x9d, 0x84, 0x90, 0xd8, 0xab, 0x00, 0x8c, 0xbc, 0xd3, 0x0a, 0xf7, 0xe4, 0x58, 0x05, 0xb8, 0xb3, 0x45, 0x06, 0xd0, 0x2c, 0x1e, 0x8f, 0xca, 0x3f, 0x0f, 0x02, 0xc1, 0xaf, 0xbd, 0x03, 0x01, 0x13, 0x8a, 0x6b, 0x3a, 0x91, 0x11, 0x41, 0x4f, 0x67, 0xdc, 0xea, 0x97, 0xf2, 0xcf, 0xce, 0xf0, 0xb4, 0xe6, 0x73, 0x96, 0xac, 0x74, 0x22, 0xe7, 0xad, 0x35, 0x85, 0xe2, 0xf9, 0x37, 0xe8, 0x1c, 0x75, 0xdf, 0x6e, 0x47, 0xf1, 0x1a, 0x71, 0x1d, 0x29, 0xc5, 0x89, 0x6f, 0xb7, 0x62, 0x0e, 0xaa, 0x18, 0xbe, 0x1b, 0xfc, 0x56, 0x3e, 0x4b, 0xc6, 0xd2, 0x79, 0x20, 0x9a, 0xdb, 0xc0, 0xfe, 0x78, 0xcd, 0x5a, 0xf4, 0x1f, 0xdd, 0xa8, 0x33, 0x88, 0x07, 0xc7, 0x31, 0xb1, 0x12, 0x10, 0x59, 0x27, 0x80, 0xec, 0x5f, 0x60, 0x51, 0x7f, 0xa9, 0x19, 0xb5, 0x4a, 0x0d, 0x2d, 0xe5, 0x7a, 0x9f, 0x93, 0xc9, 0x9c, 0xef, 0xa0, 0xe0, 0x3b, 0x4d, 0xae, 0x2a, 0xf5, 0xb0, 0xc8, 0xeb, 0xbb, 0x3c, 0x83, 0x53, 0x99, 0x61, 0x17, 0x2b, 0x04, 0x7e, 0xba, 0x77, 0xd6, 0x26, 0xe1, 0x69, 0x14, 0x63, 0x55, 0x21, 0x0c, 0x7d];

        // Transformations for encryption
        var T1 = [0xc66363a5, 0xf87c7c84, 0xee777799, 0xf67b7b8d, 0xfff2f20d, 0xd66b6bbd, 0xde6f6fb1, 0x91c5c554, 0x60303050, 0x02010103, 0xce6767a9, 0x562b2b7d, 0xe7fefe19, 0xb5d7d762, 0x4dababe6, 0xec76769a, 0x8fcaca45, 0x1f82829d, 0x89c9c940, 0xfa7d7d87, 0xeffafa15, 0xb25959eb, 0x8e4747c9, 0xfbf0f00b, 0x41adadec, 0xb3d4d467, 0x5fa2a2fd, 0x45afafea, 0x239c9cbf, 0x53a4a4f7, 0xe4727296, 0x9bc0c05b, 0x75b7b7c2, 0xe1fdfd1c, 0x3d9393ae, 0x4c26266a, 0x6c36365a, 0x7e3f3f41, 0xf5f7f702, 0x83cccc4f, 0x6834345c, 0x51a5a5f4, 0xd1e5e534, 0xf9f1f108, 0xe2717193, 0xabd8d873, 0x62313153, 0x2a15153f, 0x0804040c, 0x95c7c752, 0x46232365, 0x9dc3c35e, 0x30181828, 0x379696a1, 0x0a05050f, 0x2f9a9ab5, 0x0e070709, 0x24121236, 0x1b80809b, 0xdfe2e23d, 0xcdebeb26, 0x4e272769, 0x7fb2b2cd, 0xea75759f, 0x1209091b, 0x1d83839e, 0x582c2c74, 0x341a1a2e, 0x361b1b2d, 0xdc6e6eb2, 0xb45a5aee, 0x5ba0a0fb, 0xa45252f6, 0x763b3b4d, 0xb7d6d661, 0x7db3b3ce, 0x5229297b, 0xdde3e33e, 0x5e2f2f71, 0x13848497, 0xa65353f5, 0xb9d1d168, 0x00000000, 0xc1eded2c, 0x40202060, 0xe3fcfc1f, 0x79b1b1c8, 0xb65b5bed, 0xd46a6abe, 0x8dcbcb46, 0x67bebed9, 0x7239394b, 0x944a4ade, 0x984c4cd4, 0xb05858e8, 0x85cfcf4a, 0xbbd0d06b, 0xc5efef2a, 0x4faaaae5, 0xedfbfb16, 0x864343c5, 0x9a4d4dd7, 0x66333355, 0x11858594, 0x8a4545cf, 0xe9f9f910, 0x04020206, 0xfe7f7f81, 0xa05050f0, 0x783c3c44, 0x259f9fba, 0x4ba8a8e3, 0xa25151f3, 0x5da3a3fe, 0x804040c0, 0x058f8f8a, 0x3f9292ad, 0x219d9dbc, 0x70383848, 0xf1f5f504, 0x63bcbcdf, 0x77b6b6c1, 0xafdada75, 0x42212163, 0x20101030, 0xe5ffff1a, 0xfdf3f30e, 0xbfd2d26d, 0x81cdcd4c, 0x180c0c14, 0x26131335, 0xc3ecec2f, 0xbe5f5fe1, 0x359797a2, 0x884444cc, 0x2e171739, 0x93c4c457, 0x55a7a7f2, 0xfc7e7e82, 0x7a3d3d47, 0xc86464ac, 0xba5d5de7, 0x3219192b, 0xe6737395, 0xc06060a0, 0x19818198, 0x9e4f4fd1, 0xa3dcdc7f, 0x44222266, 0x542a2a7e, 0x3b9090ab, 0x0b888883, 0x8c4646ca, 0xc7eeee29, 0x6bb8b8d3, 0x2814143c, 0xa7dede79, 0xbc5e5ee2, 0x160b0b1d, 0xaddbdb76, 0xdbe0e03b, 0x64323256, 0x743a3a4e, 0x140a0a1e, 0x924949db, 0x0c06060a, 0x4824246c, 0xb85c5ce4, 0x9fc2c25d, 0xbdd3d36e, 0x43acacef, 0xc46262a6, 0x399191a8, 0x319595a4, 0xd3e4e437, 0xf279798b, 0xd5e7e732, 0x8bc8c843, 0x6e373759, 0xda6d6db7, 0x018d8d8c, 0xb1d5d564, 0x9c4e4ed2, 0x49a9a9e0, 0xd86c6cb4, 0xac5656fa, 0xf3f4f407, 0xcfeaea25, 0xca6565af, 0xf47a7a8e, 0x47aeaee9, 0x10080818, 0x6fbabad5, 0xf0787888, 0x4a25256f, 0x5c2e2e72, 0x381c1c24, 0x57a6a6f1, 0x73b4b4c7, 0x97c6c651, 0xcbe8e823, 0xa1dddd7c, 0xe874749c, 0x3e1f1f21, 0x964b4bdd, 0x61bdbddc, 0x0d8b8b86, 0x0f8a8a85, 0xe0707090, 0x7c3e3e42, 0x71b5b5c4, 0xcc6666aa, 0x904848d8, 0x06030305, 0xf7f6f601, 0x1c0e0e12, 0xc26161a3, 0x6a35355f, 0xae5757f9, 0x69b9b9d0, 0x17868691, 0x99c1c158, 0x3a1d1d27, 0x279e9eb9, 0xd9e1e138, 0xebf8f813, 0x2b9898b3, 0x22111133, 0xd26969bb, 0xa9d9d970, 0x078e8e89, 0x339494a7, 0x2d9b9bb6, 0x3c1e1e22, 0x15878792, 0xc9e9e920, 0x87cece49, 0xaa5555ff, 0x50282878, 0xa5dfdf7a, 0x038c8c8f, 0x59a1a1f8, 0x09898980, 0x1a0d0d17, 0x65bfbfda, 0xd7e6e631, 0x844242c6, 0xd06868b8, 0x824141c3, 0x299999b0, 0x5a2d2d77, 0x1e0f0f11, 0x7bb0b0cb, 0xa85454fc, 0x6dbbbbd6, 0x2c16163a];
        var T2 = [0xa5c66363, 0x84f87c7c, 0x99ee7777, 0x8df67b7b, 0x0dfff2f2, 0xbdd66b6b, 0xb1de6f6f, 0x5491c5c5, 0x50603030, 0x03020101, 0xa9ce6767, 0x7d562b2b, 0x19e7fefe, 0x62b5d7d7, 0xe64dabab, 0x9aec7676, 0x458fcaca, 0x9d1f8282, 0x4089c9c9, 0x87fa7d7d, 0x15effafa, 0xebb25959, 0xc98e4747, 0x0bfbf0f0, 0xec41adad, 0x67b3d4d4, 0xfd5fa2a2, 0xea45afaf, 0xbf239c9c, 0xf753a4a4, 0x96e47272, 0x5b9bc0c0, 0xc275b7b7, 0x1ce1fdfd, 0xae3d9393, 0x6a4c2626, 0x5a6c3636, 0x417e3f3f, 0x02f5f7f7, 0x4f83cccc, 0x5c683434, 0xf451a5a5, 0x34d1e5e5, 0x08f9f1f1, 0x93e27171, 0x73abd8d8, 0x53623131, 0x3f2a1515, 0x0c080404, 0x5295c7c7, 0x65462323, 0x5e9dc3c3, 0x28301818, 0xa1379696, 0x0f0a0505, 0xb52f9a9a, 0x090e0707, 0x36241212, 0x9b1b8080, 0x3ddfe2e2, 0x26cdebeb, 0x694e2727, 0xcd7fb2b2, 0x9fea7575, 0x1b120909, 0x9e1d8383, 0x74582c2c, 0x2e341a1a, 0x2d361b1b, 0xb2dc6e6e, 0xeeb45a5a, 0xfb5ba0a0, 0xf6a45252, 0x4d763b3b, 0x61b7d6d6, 0xce7db3b3, 0x7b522929, 0x3edde3e3, 0x715e2f2f, 0x97138484, 0xf5a65353, 0x68b9d1d1, 0x00000000, 0x2cc1eded, 0x60402020, 0x1fe3fcfc, 0xc879b1b1, 0xedb65b5b, 0xbed46a6a, 0x468dcbcb, 0xd967bebe, 0x4b723939, 0xde944a4a, 0xd4984c4c, 0xe8b05858, 0x4a85cfcf, 0x6bbbd0d0, 0x2ac5efef, 0xe54faaaa, 0x16edfbfb, 0xc5864343, 0xd79a4d4d, 0x55663333, 0x94118585, 0xcf8a4545, 0x10e9f9f9, 0x06040202, 0x81fe7f7f, 0xf0a05050, 0x44783c3c, 0xba259f9f, 0xe34ba8a8, 0xf3a25151, 0xfe5da3a3, 0xc0804040, 0x8a058f8f, 0xad3f9292, 0xbc219d9d, 0x48703838, 0x04f1f5f5, 0xdf63bcbc, 0xc177b6b6, 0x75afdada, 0x63422121, 0x30201010, 0x1ae5ffff, 0x0efdf3f3, 0x6dbfd2d2, 0x4c81cdcd, 0x14180c0c, 0x35261313, 0x2fc3ecec, 0xe1be5f5f, 0xa2359797, 0xcc884444, 0x392e1717, 0x5793c4c4, 0xf255a7a7, 0x82fc7e7e, 0x477a3d3d, 0xacc86464, 0xe7ba5d5d, 0x2b321919, 0x95e67373, 0xa0c06060, 0x98198181, 0xd19e4f4f, 0x7fa3dcdc, 0x66442222, 0x7e542a2a, 0xab3b9090, 0x830b8888, 0xca8c4646, 0x29c7eeee, 0xd36bb8b8, 0x3c281414, 0x79a7dede, 0xe2bc5e5e, 0x1d160b0b, 0x76addbdb, 0x3bdbe0e0, 0x56643232, 0x4e743a3a, 0x1e140a0a, 0xdb924949, 0x0a0c0606, 0x6c482424, 0xe4b85c5c, 0x5d9fc2c2, 0x6ebdd3d3, 0xef43acac, 0xa6c46262, 0xa8399191, 0xa4319595, 0x37d3e4e4, 0x8bf27979, 0x32d5e7e7, 0x438bc8c8, 0x596e3737, 0xb7da6d6d, 0x8c018d8d, 0x64b1d5d5, 0xd29c4e4e, 0xe049a9a9, 0xb4d86c6c, 0xfaac5656, 0x07f3f4f4, 0x25cfeaea, 0xafca6565, 0x8ef47a7a, 0xe947aeae, 0x18100808, 0xd56fbaba, 0x88f07878, 0x6f4a2525, 0x725c2e2e, 0x24381c1c, 0xf157a6a6, 0xc773b4b4, 0x5197c6c6, 0x23cbe8e8, 0x7ca1dddd, 0x9ce87474, 0x213e1f1f, 0xdd964b4b, 0xdc61bdbd, 0x860d8b8b, 0x850f8a8a, 0x90e07070, 0x427c3e3e, 0xc471b5b5, 0xaacc6666, 0xd8904848, 0x05060303, 0x01f7f6f6, 0x121c0e0e, 0xa3c26161, 0x5f6a3535, 0xf9ae5757, 0xd069b9b9, 0x91178686, 0x5899c1c1, 0x273a1d1d, 0xb9279e9e, 0x38d9e1e1, 0x13ebf8f8, 0xb32b9898, 0x33221111, 0xbbd26969, 0x70a9d9d9, 0x89078e8e, 0xa7339494, 0xb62d9b9b, 0x223c1e1e, 0x92158787, 0x20c9e9e9, 0x4987cece, 0xffaa5555, 0x78502828, 0x7aa5dfdf, 0x8f038c8c, 0xf859a1a1, 0x80098989, 0x171a0d0d, 0xda65bfbf, 0x31d7e6e6, 0xc6844242, 0xb8d06868, 0xc3824141, 0xb0299999, 0x775a2d2d, 0x111e0f0f, 0xcb7bb0b0, 0xfca85454, 0xd66dbbbb, 0x3a2c1616];
        var T3 = [0x63a5c663, 0x7c84f87c, 0x7799ee77, 0x7b8df67b, 0xf20dfff2, 0x6bbdd66b, 0x6fb1de6f, 0xc55491c5, 0x30506030, 0x01030201, 0x67a9ce67, 0x2b7d562b, 0xfe19e7fe, 0xd762b5d7, 0xabe64dab, 0x769aec76, 0xca458fca, 0x829d1f82, 0xc94089c9, 0x7d87fa7d, 0xfa15effa, 0x59ebb259, 0x47c98e47, 0xf00bfbf0, 0xadec41ad, 0xd467b3d4, 0xa2fd5fa2, 0xafea45af, 0x9cbf239c, 0xa4f753a4, 0x7296e472, 0xc05b9bc0, 0xb7c275b7, 0xfd1ce1fd, 0x93ae3d93, 0x266a4c26, 0x365a6c36, 0x3f417e3f, 0xf702f5f7, 0xcc4f83cc, 0x345c6834, 0xa5f451a5, 0xe534d1e5, 0xf108f9f1, 0x7193e271, 0xd873abd8, 0x31536231, 0x153f2a15, 0x040c0804, 0xc75295c7, 0x23654623, 0xc35e9dc3, 0x18283018, 0x96a13796, 0x050f0a05, 0x9ab52f9a, 0x07090e07, 0x12362412, 0x809b1b80, 0xe23ddfe2, 0xeb26cdeb, 0x27694e27, 0xb2cd7fb2, 0x759fea75, 0x091b1209, 0x839e1d83, 0x2c74582c, 0x1a2e341a, 0x1b2d361b, 0x6eb2dc6e, 0x5aeeb45a, 0xa0fb5ba0, 0x52f6a452, 0x3b4d763b, 0xd661b7d6, 0xb3ce7db3, 0x297b5229, 0xe33edde3, 0x2f715e2f, 0x84971384, 0x53f5a653, 0xd168b9d1, 0x00000000, 0xed2cc1ed, 0x20604020, 0xfc1fe3fc, 0xb1c879b1, 0x5bedb65b, 0x6abed46a, 0xcb468dcb, 0xbed967be, 0x394b7239, 0x4ade944a, 0x4cd4984c, 0x58e8b058, 0xcf4a85cf, 0xd06bbbd0, 0xef2ac5ef, 0xaae54faa, 0xfb16edfb, 0x43c58643, 0x4dd79a4d, 0x33556633, 0x85941185, 0x45cf8a45, 0xf910e9f9, 0x02060402, 0x7f81fe7f, 0x50f0a050, 0x3c44783c, 0x9fba259f, 0xa8e34ba8, 0x51f3a251, 0xa3fe5da3, 0x40c08040, 0x8f8a058f, 0x92ad3f92, 0x9dbc219d, 0x38487038, 0xf504f1f5, 0xbcdf63bc, 0xb6c177b6, 0xda75afda, 0x21634221, 0x10302010, 0xff1ae5ff, 0xf30efdf3, 0xd26dbfd2, 0xcd4c81cd, 0x0c14180c, 0x13352613, 0xec2fc3ec, 0x5fe1be5f, 0x97a23597, 0x44cc8844, 0x17392e17, 0xc45793c4, 0xa7f255a7, 0x7e82fc7e, 0x3d477a3d, 0x64acc864, 0x5de7ba5d, 0x192b3219, 0x7395e673, 0x60a0c060, 0x81981981, 0x4fd19e4f, 0xdc7fa3dc, 0x22664422, 0x2a7e542a, 0x90ab3b90, 0x88830b88, 0x46ca8c46, 0xee29c7ee, 0xb8d36bb8, 0x143c2814, 0xde79a7de, 0x5ee2bc5e, 0x0b1d160b, 0xdb76addb, 0xe03bdbe0, 0x32566432, 0x3a4e743a, 0x0a1e140a, 0x49db9249, 0x060a0c06, 0x246c4824, 0x5ce4b85c, 0xc25d9fc2, 0xd36ebdd3, 0xacef43ac, 0x62a6c462, 0x91a83991, 0x95a43195, 0xe437d3e4, 0x798bf279, 0xe732d5e7, 0xc8438bc8, 0x37596e37, 0x6db7da6d, 0x8d8c018d, 0xd564b1d5, 0x4ed29c4e, 0xa9e049a9, 0x6cb4d86c, 0x56faac56, 0xf407f3f4, 0xea25cfea, 0x65afca65, 0x7a8ef47a, 0xaee947ae, 0x08181008, 0xbad56fba, 0x7888f078, 0x256f4a25, 0x2e725c2e, 0x1c24381c, 0xa6f157a6, 0xb4c773b4, 0xc65197c6, 0xe823cbe8, 0xdd7ca1dd, 0x749ce874, 0x1f213e1f, 0x4bdd964b, 0xbddc61bd, 0x8b860d8b, 0x8a850f8a, 0x7090e070, 0x3e427c3e, 0xb5c471b5, 0x66aacc66, 0x48d89048, 0x03050603, 0xf601f7f6, 0x0e121c0e, 0x61a3c261, 0x355f6a35, 0x57f9ae57, 0xb9d069b9, 0x86911786, 0xc15899c1, 0x1d273a1d, 0x9eb9279e, 0xe138d9e1, 0xf813ebf8, 0x98b32b98, 0x11332211, 0x69bbd269, 0xd970a9d9, 0x8e89078e, 0x94a73394, 0x9bb62d9b, 0x1e223c1e, 0x87921587, 0xe920c9e9, 0xce4987ce, 0x55ffaa55, 0x28785028, 0xdf7aa5df, 0x8c8f038c, 0xa1f859a1, 0x89800989, 0x0d171a0d, 0xbfda65bf, 0xe631d7e6, 0x42c68442, 0x68b8d068, 0x41c38241, 0x99b02999, 0x2d775a2d, 0x0f111e0f, 0xb0cb7bb0, 0x54fca854, 0xbbd66dbb, 0x163a2c16];
        var T4 = [0x6363a5c6, 0x7c7c84f8, 0x777799ee, 0x7b7b8df6, 0xf2f20dff, 0x6b6bbdd6, 0x6f6fb1de, 0xc5c55491, 0x30305060, 0x01010302, 0x6767a9ce, 0x2b2b7d56, 0xfefe19e7, 0xd7d762b5, 0xababe64d, 0x76769aec, 0xcaca458f, 0x82829d1f, 0xc9c94089, 0x7d7d87fa, 0xfafa15ef, 0x5959ebb2, 0x4747c98e, 0xf0f00bfb, 0xadadec41, 0xd4d467b3, 0xa2a2fd5f, 0xafafea45, 0x9c9cbf23, 0xa4a4f753, 0x727296e4, 0xc0c05b9b, 0xb7b7c275, 0xfdfd1ce1, 0x9393ae3d, 0x26266a4c, 0x36365a6c, 0x3f3f417e, 0xf7f702f5, 0xcccc4f83, 0x34345c68, 0xa5a5f451, 0xe5e534d1, 0xf1f108f9, 0x717193e2, 0xd8d873ab, 0x31315362, 0x15153f2a, 0x04040c08, 0xc7c75295, 0x23236546, 0xc3c35e9d, 0x18182830, 0x9696a137, 0x05050f0a, 0x9a9ab52f, 0x0707090e, 0x12123624, 0x80809b1b, 0xe2e23ddf, 0xebeb26cd, 0x2727694e, 0xb2b2cd7f, 0x75759fea, 0x09091b12, 0x83839e1d, 0x2c2c7458, 0x1a1a2e34, 0x1b1b2d36, 0x6e6eb2dc, 0x5a5aeeb4, 0xa0a0fb5b, 0x5252f6a4, 0x3b3b4d76, 0xd6d661b7, 0xb3b3ce7d, 0x29297b52, 0xe3e33edd, 0x2f2f715e, 0x84849713, 0x5353f5a6, 0xd1d168b9, 0x00000000, 0xeded2cc1, 0x20206040, 0xfcfc1fe3, 0xb1b1c879, 0x5b5bedb6, 0x6a6abed4, 0xcbcb468d, 0xbebed967, 0x39394b72, 0x4a4ade94, 0x4c4cd498, 0x5858e8b0, 0xcfcf4a85, 0xd0d06bbb, 0xefef2ac5, 0xaaaae54f, 0xfbfb16ed, 0x4343c586, 0x4d4dd79a, 0x33335566, 0x85859411, 0x4545cf8a, 0xf9f910e9, 0x02020604, 0x7f7f81fe, 0x5050f0a0, 0x3c3c4478, 0x9f9fba25, 0xa8a8e34b, 0x5151f3a2, 0xa3a3fe5d, 0x4040c080, 0x8f8f8a05, 0x9292ad3f, 0x9d9dbc21, 0x38384870, 0xf5f504f1, 0xbcbcdf63, 0xb6b6c177, 0xdada75af, 0x21216342, 0x10103020, 0xffff1ae5, 0xf3f30efd, 0xd2d26dbf, 0xcdcd4c81, 0x0c0c1418, 0x13133526, 0xecec2fc3, 0x5f5fe1be, 0x9797a235, 0x4444cc88, 0x1717392e, 0xc4c45793, 0xa7a7f255, 0x7e7e82fc, 0x3d3d477a, 0x6464acc8, 0x5d5de7ba, 0x19192b32, 0x737395e6, 0x6060a0c0, 0x81819819, 0x4f4fd19e, 0xdcdc7fa3, 0x22226644, 0x2a2a7e54, 0x9090ab3b, 0x8888830b, 0x4646ca8c, 0xeeee29c7, 0xb8b8d36b, 0x14143c28, 0xdede79a7, 0x5e5ee2bc, 0x0b0b1d16, 0xdbdb76ad, 0xe0e03bdb, 0x32325664, 0x3a3a4e74, 0x0a0a1e14, 0x4949db92, 0x06060a0c, 0x24246c48, 0x5c5ce4b8, 0xc2c25d9f, 0xd3d36ebd, 0xacacef43, 0x6262a6c4, 0x9191a839, 0x9595a431, 0xe4e437d3, 0x79798bf2, 0xe7e732d5, 0xc8c8438b, 0x3737596e, 0x6d6db7da, 0x8d8d8c01, 0xd5d564b1, 0x4e4ed29c, 0xa9a9e049, 0x6c6cb4d8, 0x5656faac, 0xf4f407f3, 0xeaea25cf, 0x6565afca, 0x7a7a8ef4, 0xaeaee947, 0x08081810, 0xbabad56f, 0x787888f0, 0x25256f4a, 0x2e2e725c, 0x1c1c2438, 0xa6a6f157, 0xb4b4c773, 0xc6c65197, 0xe8e823cb, 0xdddd7ca1, 0x74749ce8, 0x1f1f213e, 0x4b4bdd96, 0xbdbddc61, 0x8b8b860d, 0x8a8a850f, 0x707090e0, 0x3e3e427c, 0xb5b5c471, 0x6666aacc, 0x4848d890, 0x03030506, 0xf6f601f7, 0x0e0e121c, 0x6161a3c2, 0x35355f6a, 0x5757f9ae, 0xb9b9d069, 0x86869117, 0xc1c15899, 0x1d1d273a, 0x9e9eb927, 0xe1e138d9, 0xf8f813eb, 0x9898b32b, 0x11113322, 0x6969bbd2, 0xd9d970a9, 0x8e8e8907, 0x9494a733, 0x9b9bb62d, 0x1e1e223c, 0x87879215, 0xe9e920c9, 0xcece4987, 0x5555ffaa, 0x28287850, 0xdfdf7aa5, 0x8c8c8f03, 0xa1a1f859, 0x89898009, 0x0d0d171a, 0xbfbfda65, 0xe6e631d7, 0x4242c684, 0x6868b8d0, 0x4141c382, 0x9999b029, 0x2d2d775a, 0x0f0f111e, 0xb0b0cb7b, 0x5454fca8, 0xbbbbd66d, 0x16163a2c];

        // Transformations for decryption
        var T5 = [0x51f4a750, 0x7e416553, 0x1a17a4c3, 0x3a275e96, 0x3bab6bcb, 0x1f9d45f1, 0xacfa58ab, 0x4be30393, 0x2030fa55, 0xad766df6, 0x88cc7691, 0xf5024c25, 0x4fe5d7fc, 0xc52acbd7, 0x26354480, 0xb562a38f, 0xdeb15a49, 0x25ba1b67, 0x45ea0e98, 0x5dfec0e1, 0xc32f7502, 0x814cf012, 0x8d4697a3, 0x6bd3f9c6, 0x038f5fe7, 0x15929c95, 0xbf6d7aeb, 0x955259da, 0xd4be832d, 0x587421d3, 0x49e06929, 0x8ec9c844, 0x75c2896a, 0xf48e7978, 0x99583e6b, 0x27b971dd, 0xbee14fb6, 0xf088ad17, 0xc920ac66, 0x7dce3ab4, 0x63df4a18, 0xe51a3182, 0x97513360, 0x62537f45, 0xb16477e0, 0xbb6bae84, 0xfe81a01c, 0xf9082b94, 0x70486858, 0x8f45fd19, 0x94de6c87, 0x527bf8b7, 0xab73d323, 0x724b02e2, 0xe31f8f57, 0x6655ab2a, 0xb2eb2807, 0x2fb5c203, 0x86c57b9a, 0xd33708a5, 0x302887f2, 0x23bfa5b2, 0x02036aba, 0xed16825c, 0x8acf1c2b, 0xa779b492, 0xf307f2f0, 0x4e69e2a1, 0x65daf4cd, 0x0605bed5, 0xd134621f, 0xc4a6fe8a, 0x342e539d, 0xa2f355a0, 0x058ae132, 0xa4f6eb75, 0x0b83ec39, 0x4060efaa, 0x5e719f06, 0xbd6e1051, 0x3e218af9, 0x96dd063d, 0xdd3e05ae, 0x4de6bd46, 0x91548db5, 0x71c45d05, 0x0406d46f, 0x605015ff, 0x1998fb24, 0xd6bde997, 0x894043cc, 0x67d99e77, 0xb0e842bd, 0x07898b88, 0xe7195b38, 0x79c8eedb, 0xa17c0a47, 0x7c420fe9, 0xf8841ec9, 0x00000000, 0x09808683, 0x322bed48, 0x1e1170ac, 0x6c5a724e, 0xfd0efffb, 0x0f853856, 0x3daed51e, 0x362d3927, 0x0a0fd964, 0x685ca621, 0x9b5b54d1, 0x24362e3a, 0x0c0a67b1, 0x9357e70f, 0xb4ee96d2, 0x1b9b919e, 0x80c0c54f, 0x61dc20a2, 0x5a774b69, 0x1c121a16, 0xe293ba0a, 0xc0a02ae5, 0x3c22e043, 0x121b171d, 0x0e090d0b, 0xf28bc7ad, 0x2db6a8b9, 0x141ea9c8, 0x57f11985, 0xaf75074c, 0xee99ddbb, 0xa37f60fd, 0xf701269f, 0x5c72f5bc, 0x44663bc5, 0x5bfb7e34, 0x8b432976, 0xcb23c6dc, 0xb6edfc68, 0xb8e4f163, 0xd731dcca, 0x42638510, 0x13972240, 0x84c61120, 0x854a247d, 0xd2bb3df8, 0xaef93211, 0xc729a16d, 0x1d9e2f4b, 0xdcb230f3, 0x0d8652ec, 0x77c1e3d0, 0x2bb3166c, 0xa970b999, 0x119448fa, 0x47e96422, 0xa8fc8cc4, 0xa0f03f1a, 0x567d2cd8, 0x223390ef, 0x87494ec7, 0xd938d1c1, 0x8ccaa2fe, 0x98d40b36, 0xa6f581cf, 0xa57ade28, 0xdab78e26, 0x3fadbfa4, 0x2c3a9de4, 0x5078920d, 0x6a5fcc9b, 0x547e4662, 0xf68d13c2, 0x90d8b8e8, 0x2e39f75e, 0x82c3aff5, 0x9f5d80be, 0x69d0937c, 0x6fd52da9, 0xcf2512b3, 0xc8ac993b, 0x10187da7, 0xe89c636e, 0xdb3bbb7b, 0xcd267809, 0x6e5918f4, 0xec9ab701, 0x834f9aa8, 0xe6956e65, 0xaaffe67e, 0x21bccf08, 0xef15e8e6, 0xbae79bd9, 0x4a6f36ce, 0xea9f09d4, 0x29b07cd6, 0x31a4b2af, 0x2a3f2331, 0xc6a59430, 0x35a266c0, 0x744ebc37, 0xfc82caa6, 0xe090d0b0, 0x33a7d815, 0xf104984a, 0x41ecdaf7, 0x7fcd500e, 0x1791f62f, 0x764dd68d, 0x43efb04d, 0xccaa4d54, 0xe49604df, 0x9ed1b5e3, 0x4c6a881b, 0xc12c1fb8, 0x4665517f, 0x9d5eea04, 0x018c355d, 0xfa877473, 0xfb0b412e, 0xb3671d5a, 0x92dbd252, 0xe9105633, 0x6dd64713, 0x9ad7618c, 0x37a10c7a, 0x59f8148e, 0xeb133c89, 0xcea927ee, 0xb761c935, 0xe11ce5ed, 0x7a47b13c, 0x9cd2df59, 0x55f2733f, 0x1814ce79, 0x73c737bf, 0x53f7cdea, 0x5ffdaa5b, 0xdf3d6f14, 0x7844db86, 0xcaaff381, 0xb968c43e, 0x3824342c, 0xc2a3405f, 0x161dc372, 0xbce2250c, 0x283c498b, 0xff0d9541, 0x39a80171, 0x080cb3de, 0xd8b4e49c, 0x6456c190, 0x7bcb8461, 0xd532b670, 0x486c5c74, 0xd0b85742];
        var T6 = [0x5051f4a7, 0x537e4165, 0xc31a17a4, 0x963a275e, 0xcb3bab6b, 0xf11f9d45, 0xabacfa58, 0x934be303, 0x552030fa, 0xf6ad766d, 0x9188cc76, 0x25f5024c, 0xfc4fe5d7, 0xd7c52acb, 0x80263544, 0x8fb562a3, 0x49deb15a, 0x6725ba1b, 0x9845ea0e, 0xe15dfec0, 0x02c32f75, 0x12814cf0, 0xa38d4697, 0xc66bd3f9, 0xe7038f5f, 0x9515929c, 0xebbf6d7a, 0xda955259, 0x2dd4be83, 0xd3587421, 0x2949e069, 0x448ec9c8, 0x6a75c289, 0x78f48e79, 0x6b99583e, 0xdd27b971, 0xb6bee14f, 0x17f088ad, 0x66c920ac, 0xb47dce3a, 0x1863df4a, 0x82e51a31, 0x60975133, 0x4562537f, 0xe0b16477, 0x84bb6bae, 0x1cfe81a0, 0x94f9082b, 0x58704868, 0x198f45fd, 0x8794de6c, 0xb7527bf8, 0x23ab73d3, 0xe2724b02, 0x57e31f8f, 0x2a6655ab, 0x07b2eb28, 0x032fb5c2, 0x9a86c57b, 0xa5d33708, 0xf2302887, 0xb223bfa5, 0xba02036a, 0x5ced1682, 0x2b8acf1c, 0x92a779b4, 0xf0f307f2, 0xa14e69e2, 0xcd65daf4, 0xd50605be, 0x1fd13462, 0x8ac4a6fe, 0x9d342e53, 0xa0a2f355, 0x32058ae1, 0x75a4f6eb, 0x390b83ec, 0xaa4060ef, 0x065e719f, 0x51bd6e10, 0xf93e218a, 0x3d96dd06, 0xaedd3e05, 0x464de6bd, 0xb591548d, 0x0571c45d, 0x6f0406d4, 0xff605015, 0x241998fb, 0x97d6bde9, 0xcc894043, 0x7767d99e, 0xbdb0e842, 0x8807898b, 0x38e7195b, 0xdb79c8ee, 0x47a17c0a, 0xe97c420f, 0xc9f8841e, 0x00000000, 0x83098086, 0x48322bed, 0xac1e1170, 0x4e6c5a72, 0xfbfd0eff, 0x560f8538, 0x1e3daed5, 0x27362d39, 0x640a0fd9, 0x21685ca6, 0xd19b5b54, 0x3a24362e, 0xb10c0a67, 0x0f9357e7, 0xd2b4ee96, 0x9e1b9b91, 0x4f80c0c5, 0xa261dc20, 0x695a774b, 0x161c121a, 0x0ae293ba, 0xe5c0a02a, 0x433c22e0, 0x1d121b17, 0x0b0e090d, 0xadf28bc7, 0xb92db6a8, 0xc8141ea9, 0x8557f119, 0x4caf7507, 0xbbee99dd, 0xfda37f60, 0x9ff70126, 0xbc5c72f5, 0xc544663b, 0x345bfb7e, 0x768b4329, 0xdccb23c6, 0x68b6edfc, 0x63b8e4f1, 0xcad731dc, 0x10426385, 0x40139722, 0x2084c611, 0x7d854a24, 0xf8d2bb3d, 0x11aef932, 0x6dc729a1, 0x4b1d9e2f, 0xf3dcb230, 0xec0d8652, 0xd077c1e3, 0x6c2bb316, 0x99a970b9, 0xfa119448, 0x2247e964, 0xc4a8fc8c, 0x1aa0f03f, 0xd8567d2c, 0xef223390, 0xc787494e, 0xc1d938d1, 0xfe8ccaa2, 0x3698d40b, 0xcfa6f581, 0x28a57ade, 0x26dab78e, 0xa43fadbf, 0xe42c3a9d, 0x0d507892, 0x9b6a5fcc, 0x62547e46, 0xc2f68d13, 0xe890d8b8, 0x5e2e39f7, 0xf582c3af, 0xbe9f5d80, 0x7c69d093, 0xa96fd52d, 0xb3cf2512, 0x3bc8ac99, 0xa710187d, 0x6ee89c63, 0x7bdb3bbb, 0x09cd2678, 0xf46e5918, 0x01ec9ab7, 0xa8834f9a, 0x65e6956e, 0x7eaaffe6, 0x0821bccf, 0xe6ef15e8, 0xd9bae79b, 0xce4a6f36, 0xd4ea9f09, 0xd629b07c, 0xaf31a4b2, 0x312a3f23, 0x30c6a594, 0xc035a266, 0x37744ebc, 0xa6fc82ca, 0xb0e090d0, 0x1533a7d8, 0x4af10498, 0xf741ecda, 0x0e7fcd50, 0x2f1791f6, 0x8d764dd6, 0x4d43efb0, 0x54ccaa4d, 0xdfe49604, 0xe39ed1b5, 0x1b4c6a88, 0xb8c12c1f, 0x7f466551, 0x049d5eea, 0x5d018c35, 0x73fa8774, 0x2efb0b41, 0x5ab3671d, 0x5292dbd2, 0x33e91056, 0x136dd647, 0x8c9ad761, 0x7a37a10c, 0x8e59f814, 0x89eb133c, 0xeecea927, 0x35b761c9, 0xede11ce5, 0x3c7a47b1, 0x599cd2df, 0x3f55f273, 0x791814ce, 0xbf73c737, 0xea53f7cd, 0x5b5ffdaa, 0x14df3d6f, 0x867844db, 0x81caaff3, 0x3eb968c4, 0x2c382434, 0x5fc2a340, 0x72161dc3, 0x0cbce225, 0x8b283c49, 0x41ff0d95, 0x7139a801, 0xde080cb3, 0x9cd8b4e4, 0x906456c1, 0x617bcb84, 0x70d532b6, 0x74486c5c, 0x42d0b857];
        var T7 = [0xa75051f4, 0x65537e41, 0xa4c31a17, 0x5e963a27, 0x6bcb3bab, 0x45f11f9d, 0x58abacfa, 0x03934be3, 0xfa552030, 0x6df6ad76, 0x769188cc, 0x4c25f502, 0xd7fc4fe5, 0xcbd7c52a, 0x44802635, 0xa38fb562, 0x5a49deb1, 0x1b6725ba, 0x0e9845ea, 0xc0e15dfe, 0x7502c32f, 0xf012814c, 0x97a38d46, 0xf9c66bd3, 0x5fe7038f, 0x9c951592, 0x7aebbf6d, 0x59da9552, 0x832dd4be, 0x21d35874, 0x692949e0, 0xc8448ec9, 0x896a75c2, 0x7978f48e, 0x3e6b9958, 0x71dd27b9, 0x4fb6bee1, 0xad17f088, 0xac66c920, 0x3ab47dce, 0x4a1863df, 0x3182e51a, 0x33609751, 0x7f456253, 0x77e0b164, 0xae84bb6b, 0xa01cfe81, 0x2b94f908, 0x68587048, 0xfd198f45, 0x6c8794de, 0xf8b7527b, 0xd323ab73, 0x02e2724b, 0x8f57e31f, 0xab2a6655, 0x2807b2eb, 0xc2032fb5, 0x7b9a86c5, 0x08a5d337, 0x87f23028, 0xa5b223bf, 0x6aba0203, 0x825ced16, 0x1c2b8acf, 0xb492a779, 0xf2f0f307, 0xe2a14e69, 0xf4cd65da, 0xbed50605, 0x621fd134, 0xfe8ac4a6, 0x539d342e, 0x55a0a2f3, 0xe132058a, 0xeb75a4f6, 0xec390b83, 0xefaa4060, 0x9f065e71, 0x1051bd6e, 0x8af93e21, 0x063d96dd, 0x05aedd3e, 0xbd464de6, 0x8db59154, 0x5d0571c4, 0xd46f0406, 0x15ff6050, 0xfb241998, 0xe997d6bd, 0x43cc8940, 0x9e7767d9, 0x42bdb0e8, 0x8b880789, 0x5b38e719, 0xeedb79c8, 0x0a47a17c, 0x0fe97c42, 0x1ec9f884, 0x00000000, 0x86830980, 0xed48322b, 0x70ac1e11, 0x724e6c5a, 0xfffbfd0e, 0x38560f85, 0xd51e3dae, 0x3927362d, 0xd9640a0f, 0xa621685c, 0x54d19b5b, 0x2e3a2436, 0x67b10c0a, 0xe70f9357, 0x96d2b4ee, 0x919e1b9b, 0xc54f80c0, 0x20a261dc, 0x4b695a77, 0x1a161c12, 0xba0ae293, 0x2ae5c0a0, 0xe0433c22, 0x171d121b, 0x0d0b0e09, 0xc7adf28b, 0xa8b92db6, 0xa9c8141e, 0x198557f1, 0x074caf75, 0xddbbee99, 0x60fda37f, 0x269ff701, 0xf5bc5c72, 0x3bc54466, 0x7e345bfb, 0x29768b43, 0xc6dccb23, 0xfc68b6ed, 0xf163b8e4, 0xdccad731, 0x85104263, 0x22401397, 0x112084c6, 0x247d854a, 0x3df8d2bb, 0x3211aef9, 0xa16dc729, 0x2f4b1d9e, 0x30f3dcb2, 0x52ec0d86, 0xe3d077c1, 0x166c2bb3, 0xb999a970, 0x48fa1194, 0x642247e9, 0x8cc4a8fc, 0x3f1aa0f0, 0x2cd8567d, 0x90ef2233, 0x4ec78749, 0xd1c1d938, 0xa2fe8cca, 0x0b3698d4, 0x81cfa6f5, 0xde28a57a, 0x8e26dab7, 0xbfa43fad, 0x9de42c3a, 0x920d5078, 0xcc9b6a5f, 0x4662547e, 0x13c2f68d, 0xb8e890d8, 0xf75e2e39, 0xaff582c3, 0x80be9f5d, 0x937c69d0, 0x2da96fd5, 0x12b3cf25, 0x993bc8ac, 0x7da71018, 0x636ee89c, 0xbb7bdb3b, 0x7809cd26, 0x18f46e59, 0xb701ec9a, 0x9aa8834f, 0x6e65e695, 0xe67eaaff, 0xcf0821bc, 0xe8e6ef15, 0x9bd9bae7, 0x36ce4a6f, 0x09d4ea9f, 0x7cd629b0, 0xb2af31a4, 0x23312a3f, 0x9430c6a5, 0x66c035a2, 0xbc37744e, 0xcaa6fc82, 0xd0b0e090, 0xd81533a7, 0x984af104, 0xdaf741ec, 0x500e7fcd, 0xf62f1791, 0xd68d764d, 0xb04d43ef, 0x4d54ccaa, 0x04dfe496, 0xb5e39ed1, 0x881b4c6a, 0x1fb8c12c, 0x517f4665, 0xea049d5e, 0x355d018c, 0x7473fa87, 0x412efb0b, 0x1d5ab367, 0xd25292db, 0x5633e910, 0x47136dd6, 0x618c9ad7, 0x0c7a37a1, 0x148e59f8, 0x3c89eb13, 0x27eecea9, 0xc935b761, 0xe5ede11c, 0xb13c7a47, 0xdf599cd2, 0x733f55f2, 0xce791814, 0x37bf73c7, 0xcdea53f7, 0xaa5b5ffd, 0x6f14df3d, 0xdb867844, 0xf381caaf, 0xc43eb968, 0x342c3824, 0x405fc2a3, 0xc372161d, 0x250cbce2, 0x498b283c, 0x9541ff0d, 0x017139a8, 0xb3de080c, 0xe49cd8b4, 0xc1906456, 0x84617bcb, 0xb670d532, 0x5c74486c, 0x5742d0b8];
        var T8 = [0xf4a75051, 0x4165537e, 0x17a4c31a, 0x275e963a, 0xab6bcb3b, 0x9d45f11f, 0xfa58abac, 0xe303934b, 0x30fa5520, 0x766df6ad, 0xcc769188, 0x024c25f5, 0xe5d7fc4f, 0x2acbd7c5, 0x35448026, 0x62a38fb5, 0xb15a49de, 0xba1b6725, 0xea0e9845, 0xfec0e15d, 0x2f7502c3, 0x4cf01281, 0x4697a38d, 0xd3f9c66b, 0x8f5fe703, 0x929c9515, 0x6d7aebbf, 0x5259da95, 0xbe832dd4, 0x7421d358, 0xe0692949, 0xc9c8448e, 0xc2896a75, 0x8e7978f4, 0x583e6b99, 0xb971dd27, 0xe14fb6be, 0x88ad17f0, 0x20ac66c9, 0xce3ab47d, 0xdf4a1863, 0x1a3182e5, 0x51336097, 0x537f4562, 0x6477e0b1, 0x6bae84bb, 0x81a01cfe, 0x082b94f9, 0x48685870, 0x45fd198f, 0xde6c8794, 0x7bf8b752, 0x73d323ab, 0x4b02e272, 0x1f8f57e3, 0x55ab2a66, 0xeb2807b2, 0xb5c2032f, 0xc57b9a86, 0x3708a5d3, 0x2887f230, 0xbfa5b223, 0x036aba02, 0x16825ced, 0xcf1c2b8a, 0x79b492a7, 0x07f2f0f3, 0x69e2a14e, 0xdaf4cd65, 0x05bed506, 0x34621fd1, 0xa6fe8ac4, 0x2e539d34, 0xf355a0a2, 0x8ae13205, 0xf6eb75a4, 0x83ec390b, 0x60efaa40, 0x719f065e, 0x6e1051bd, 0x218af93e, 0xdd063d96, 0x3e05aedd, 0xe6bd464d, 0x548db591, 0xc45d0571, 0x06d46f04, 0x5015ff60, 0x98fb2419, 0xbde997d6, 0x4043cc89, 0xd99e7767, 0xe842bdb0, 0x898b8807, 0x195b38e7, 0xc8eedb79, 0x7c0a47a1, 0x420fe97c, 0x841ec9f8, 0x00000000, 0x80868309, 0x2bed4832, 0x1170ac1e, 0x5a724e6c, 0x0efffbfd, 0x8538560f, 0xaed51e3d, 0x2d392736, 0x0fd9640a, 0x5ca62168, 0x5b54d19b, 0x362e3a24, 0x0a67b10c, 0x57e70f93, 0xee96d2b4, 0x9b919e1b, 0xc0c54f80, 0xdc20a261, 0x774b695a, 0x121a161c, 0x93ba0ae2, 0xa02ae5c0, 0x22e0433c, 0x1b171d12, 0x090d0b0e, 0x8bc7adf2, 0xb6a8b92d, 0x1ea9c814, 0xf1198557, 0x75074caf, 0x99ddbbee, 0x7f60fda3, 0x01269ff7, 0x72f5bc5c, 0x663bc544, 0xfb7e345b, 0x4329768b, 0x23c6dccb, 0xedfc68b6, 0xe4f163b8, 0x31dccad7, 0x63851042, 0x97224013, 0xc6112084, 0x4a247d85, 0xbb3df8d2, 0xf93211ae, 0x29a16dc7, 0x9e2f4b1d, 0xb230f3dc, 0x8652ec0d, 0xc1e3d077, 0xb3166c2b, 0x70b999a9, 0x9448fa11, 0xe9642247, 0xfc8cc4a8, 0xf03f1aa0, 0x7d2cd856, 0x3390ef22, 0x494ec787, 0x38d1c1d9, 0xcaa2fe8c, 0xd40b3698, 0xf581cfa6, 0x7ade28a5, 0xb78e26da, 0xadbfa43f, 0x3a9de42c, 0x78920d50, 0x5fcc9b6a, 0x7e466254, 0x8d13c2f6, 0xd8b8e890, 0x39f75e2e, 0xc3aff582, 0x5d80be9f, 0xd0937c69, 0xd52da96f, 0x2512b3cf, 0xac993bc8, 0x187da710, 0x9c636ee8, 0x3bbb7bdb, 0x267809cd, 0x5918f46e, 0x9ab701ec, 0x4f9aa883, 0x956e65e6, 0xffe67eaa, 0xbccf0821, 0x15e8e6ef, 0xe79bd9ba, 0x6f36ce4a, 0x9f09d4ea, 0xb07cd629, 0xa4b2af31, 0x3f23312a, 0xa59430c6, 0xa266c035, 0x4ebc3774, 0x82caa6fc, 0x90d0b0e0, 0xa7d81533, 0x04984af1, 0xecdaf741, 0xcd500e7f, 0x91f62f17, 0x4dd68d76, 0xefb04d43, 0xaa4d54cc, 0x9604dfe4, 0xd1b5e39e, 0x6a881b4c, 0x2c1fb8c1, 0x65517f46, 0x5eea049d, 0x8c355d01, 0x877473fa, 0x0b412efb, 0x671d5ab3, 0xdbd25292, 0x105633e9, 0xd647136d, 0xd7618c9a, 0xa10c7a37, 0xf8148e59, 0x133c89eb, 0xa927eece, 0x61c935b7, 0x1ce5ede1, 0x47b13c7a, 0xd2df599c, 0xf2733f55, 0x14ce7918, 0xc737bf73, 0xf7cdea53, 0xfdaa5b5f, 0x3d6f14df, 0x44db8678, 0xaff381ca, 0x68c43eb9, 0x24342c38, 0xa3405fc2, 0x1dc37216, 0xe2250cbc, 0x3c498b28, 0x0d9541ff, 0xa8017139, 0x0cb3de08, 0xb4e49cd8, 0x56c19064, 0xcb84617b, 0x32b670d5, 0x6c5c7448, 0xb85742d0];

        // Transformations for decryption key expansion
        var U1 = [0x00000000, 0x0e090d0b, 0x1c121a16, 0x121b171d, 0x3824342c, 0x362d3927, 0x24362e3a, 0x2a3f2331, 0x70486858, 0x7e416553, 0x6c5a724e, 0x62537f45, 0x486c5c74, 0x4665517f, 0x547e4662, 0x5a774b69, 0xe090d0b0, 0xee99ddbb, 0xfc82caa6, 0xf28bc7ad, 0xd8b4e49c, 0xd6bde997, 0xc4a6fe8a, 0xcaaff381, 0x90d8b8e8, 0x9ed1b5e3, 0x8ccaa2fe, 0x82c3aff5, 0xa8fc8cc4, 0xa6f581cf, 0xb4ee96d2, 0xbae79bd9, 0xdb3bbb7b, 0xd532b670, 0xc729a16d, 0xc920ac66, 0xe31f8f57, 0xed16825c, 0xff0d9541, 0xf104984a, 0xab73d323, 0xa57ade28, 0xb761c935, 0xb968c43e, 0x9357e70f, 0x9d5eea04, 0x8f45fd19, 0x814cf012, 0x3bab6bcb, 0x35a266c0, 0x27b971dd, 0x29b07cd6, 0x038f5fe7, 0x0d8652ec, 0x1f9d45f1, 0x119448fa, 0x4be30393, 0x45ea0e98, 0x57f11985, 0x59f8148e, 0x73c737bf, 0x7dce3ab4, 0x6fd52da9, 0x61dc20a2, 0xad766df6, 0xa37f60fd, 0xb16477e0, 0xbf6d7aeb, 0x955259da, 0x9b5b54d1, 0x894043cc, 0x87494ec7, 0xdd3e05ae, 0xd33708a5, 0xc12c1fb8, 0xcf2512b3, 0xe51a3182, 0xeb133c89, 0xf9082b94, 0xf701269f, 0x4de6bd46, 0x43efb04d, 0x51f4a750, 0x5ffdaa5b, 0x75c2896a, 0x7bcb8461, 0x69d0937c, 0x67d99e77, 0x3daed51e, 0x33a7d815, 0x21bccf08, 0x2fb5c203, 0x058ae132, 0x0b83ec39, 0x1998fb24, 0x1791f62f, 0x764dd68d, 0x7844db86, 0x6a5fcc9b, 0x6456c190, 0x4e69e2a1, 0x4060efaa, 0x527bf8b7, 0x5c72f5bc, 0x0605bed5, 0x080cb3de, 0x1a17a4c3, 0x141ea9c8, 0x3e218af9, 0x302887f2, 0x223390ef, 0x2c3a9de4, 0x96dd063d, 0x98d40b36, 0x8acf1c2b, 0x84c61120, 0xaef93211, 0xa0f03f1a, 0xb2eb2807, 0xbce2250c, 0xe6956e65, 0xe89c636e, 0xfa877473, 0xf48e7978, 0xdeb15a49, 0xd0b85742, 0xc2a3405f, 0xccaa4d54, 0x41ecdaf7, 0x4fe5d7fc, 0x5dfec0e1, 0x53f7cdea, 0x79c8eedb, 0x77c1e3d0, 0x65daf4cd, 0x6bd3f9c6, 0x31a4b2af, 0x3fadbfa4, 0x2db6a8b9, 0x23bfa5b2, 0x09808683, 0x07898b88, 0x15929c95, 0x1b9b919e, 0xa17c0a47, 0xaf75074c, 0xbd6e1051, 0xb3671d5a, 0x99583e6b, 0x97513360, 0x854a247d, 0x8b432976, 0xd134621f, 0xdf3d6f14, 0xcd267809, 0xc32f7502, 0xe9105633, 0xe7195b38, 0xf5024c25, 0xfb0b412e, 0x9ad7618c, 0x94de6c87, 0x86c57b9a, 0x88cc7691, 0xa2f355a0, 0xacfa58ab, 0xbee14fb6, 0xb0e842bd, 0xea9f09d4, 0xe49604df, 0xf68d13c2, 0xf8841ec9, 0xd2bb3df8, 0xdcb230f3, 0xcea927ee, 0xc0a02ae5, 0x7a47b13c, 0x744ebc37, 0x6655ab2a, 0x685ca621, 0x42638510, 0x4c6a881b, 0x5e719f06, 0x5078920d, 0x0a0fd964, 0x0406d46f, 0x161dc372, 0x1814ce79, 0x322bed48, 0x3c22e043, 0x2e39f75e, 0x2030fa55, 0xec9ab701, 0xe293ba0a, 0xf088ad17, 0xfe81a01c, 0xd4be832d, 0xdab78e26, 0xc8ac993b, 0xc6a59430, 0x9cd2df59, 0x92dbd252, 0x80c0c54f, 0x8ec9c844, 0xa4f6eb75, 0xaaffe67e, 0xb8e4f163, 0xb6edfc68, 0x0c0a67b1, 0x02036aba, 0x10187da7, 0x1e1170ac, 0x342e539d, 0x3a275e96, 0x283c498b, 0x26354480, 0x7c420fe9, 0x724b02e2, 0x605015ff, 0x6e5918f4, 0x44663bc5, 0x4a6f36ce, 0x587421d3, 0x567d2cd8, 0x37a10c7a, 0x39a80171, 0x2bb3166c, 0x25ba1b67, 0x0f853856, 0x018c355d, 0x13972240, 0x1d9e2f4b, 0x47e96422, 0x49e06929, 0x5bfb7e34, 0x55f2733f, 0x7fcd500e, 0x71c45d05, 0x63df4a18, 0x6dd64713, 0xd731dcca, 0xd938d1c1, 0xcb23c6dc, 0xc52acbd7, 0xef15e8e6, 0xe11ce5ed, 0xf307f2f0, 0xfd0efffb, 0xa779b492, 0xa970b999, 0xbb6bae84, 0xb562a38f, 0x9f5d80be, 0x91548db5, 0x834f9aa8, 0x8d4697a3];
        var U2 = [0x00000000, 0x0b0e090d, 0x161c121a, 0x1d121b17, 0x2c382434, 0x27362d39, 0x3a24362e, 0x312a3f23, 0x58704868, 0x537e4165, 0x4e6c5a72, 0x4562537f, 0x74486c5c, 0x7f466551, 0x62547e46, 0x695a774b, 0xb0e090d0, 0xbbee99dd, 0xa6fc82ca, 0xadf28bc7, 0x9cd8b4e4, 0x97d6bde9, 0x8ac4a6fe, 0x81caaff3, 0xe890d8b8, 0xe39ed1b5, 0xfe8ccaa2, 0xf582c3af, 0xc4a8fc8c, 0xcfa6f581, 0xd2b4ee96, 0xd9bae79b, 0x7bdb3bbb, 0x70d532b6, 0x6dc729a1, 0x66c920ac, 0x57e31f8f, 0x5ced1682, 0x41ff0d95, 0x4af10498, 0x23ab73d3, 0x28a57ade, 0x35b761c9, 0x3eb968c4, 0x0f9357e7, 0x049d5eea, 0x198f45fd, 0x12814cf0, 0xcb3bab6b, 0xc035a266, 0xdd27b971, 0xd629b07c, 0xe7038f5f, 0xec0d8652, 0xf11f9d45, 0xfa119448, 0x934be303, 0x9845ea0e, 0x8557f119, 0x8e59f814, 0xbf73c737, 0xb47dce3a, 0xa96fd52d, 0xa261dc20, 0xf6ad766d, 0xfda37f60, 0xe0b16477, 0xebbf6d7a, 0xda955259, 0xd19b5b54, 0xcc894043, 0xc787494e, 0xaedd3e05, 0xa5d33708, 0xb8c12c1f, 0xb3cf2512, 0x82e51a31, 0x89eb133c, 0x94f9082b, 0x9ff70126, 0x464de6bd, 0x4d43efb0, 0x5051f4a7, 0x5b5ffdaa, 0x6a75c289, 0x617bcb84, 0x7c69d093, 0x7767d99e, 0x1e3daed5, 0x1533a7d8, 0x0821bccf, 0x032fb5c2, 0x32058ae1, 0x390b83ec, 0x241998fb, 0x2f1791f6, 0x8d764dd6, 0x867844db, 0x9b6a5fcc, 0x906456c1, 0xa14e69e2, 0xaa4060ef, 0xb7527bf8, 0xbc5c72f5, 0xd50605be, 0xde080cb3, 0xc31a17a4, 0xc8141ea9, 0xf93e218a, 0xf2302887, 0xef223390, 0xe42c3a9d, 0x3d96dd06, 0x3698d40b, 0x2b8acf1c, 0x2084c611, 0x11aef932, 0x1aa0f03f, 0x07b2eb28, 0x0cbce225, 0x65e6956e, 0x6ee89c63, 0x73fa8774, 0x78f48e79, 0x49deb15a, 0x42d0b857, 0x5fc2a340, 0x54ccaa4d, 0xf741ecda, 0xfc4fe5d7, 0xe15dfec0, 0xea53f7cd, 0xdb79c8ee, 0xd077c1e3, 0xcd65daf4, 0xc66bd3f9, 0xaf31a4b2, 0xa43fadbf, 0xb92db6a8, 0xb223bfa5, 0x83098086, 0x8807898b, 0x9515929c, 0x9e1b9b91, 0x47a17c0a, 0x4caf7507, 0x51bd6e10, 0x5ab3671d, 0x6b99583e, 0x60975133, 0x7d854a24, 0x768b4329, 0x1fd13462, 0x14df3d6f, 0x09cd2678, 0x02c32f75, 0x33e91056, 0x38e7195b, 0x25f5024c, 0x2efb0b41, 0x8c9ad761, 0x8794de6c, 0x9a86c57b, 0x9188cc76, 0xa0a2f355, 0xabacfa58, 0xb6bee14f, 0xbdb0e842, 0xd4ea9f09, 0xdfe49604, 0xc2f68d13, 0xc9f8841e, 0xf8d2bb3d, 0xf3dcb230, 0xeecea927, 0xe5c0a02a, 0x3c7a47b1, 0x37744ebc, 0x2a6655ab, 0x21685ca6, 0x10426385, 0x1b4c6a88, 0x065e719f, 0x0d507892, 0x640a0fd9, 0x6f0406d4, 0x72161dc3, 0x791814ce, 0x48322bed, 0x433c22e0, 0x5e2e39f7, 0x552030fa, 0x01ec9ab7, 0x0ae293ba, 0x17f088ad, 0x1cfe81a0, 0x2dd4be83, 0x26dab78e, 0x3bc8ac99, 0x30c6a594, 0x599cd2df, 0x5292dbd2, 0x4f80c0c5, 0x448ec9c8, 0x75a4f6eb, 0x7eaaffe6, 0x63b8e4f1, 0x68b6edfc, 0xb10c0a67, 0xba02036a, 0xa710187d, 0xac1e1170, 0x9d342e53, 0x963a275e, 0x8b283c49, 0x80263544, 0xe97c420f, 0xe2724b02, 0xff605015, 0xf46e5918, 0xc544663b, 0xce4a6f36, 0xd3587421, 0xd8567d2c, 0x7a37a10c, 0x7139a801, 0x6c2bb316, 0x6725ba1b, 0x560f8538, 0x5d018c35, 0x40139722, 0x4b1d9e2f, 0x2247e964, 0x2949e069, 0x345bfb7e, 0x3f55f273, 0x0e7fcd50, 0x0571c45d, 0x1863df4a, 0x136dd647, 0xcad731dc, 0xc1d938d1, 0xdccb23c6, 0xd7c52acb, 0xe6ef15e8, 0xede11ce5, 0xf0f307f2, 0xfbfd0eff, 0x92a779b4, 0x99a970b9, 0x84bb6bae, 0x8fb562a3, 0xbe9f5d80, 0xb591548d, 0xa8834f9a, 0xa38d4697];
        var U3 = [0x00000000, 0x0d0b0e09, 0x1a161c12, 0x171d121b, 0x342c3824, 0x3927362d, 0x2e3a2436, 0x23312a3f, 0x68587048, 0x65537e41, 0x724e6c5a, 0x7f456253, 0x5c74486c, 0x517f4665, 0x4662547e, 0x4b695a77, 0xd0b0e090, 0xddbbee99, 0xcaa6fc82, 0xc7adf28b, 0xe49cd8b4, 0xe997d6bd, 0xfe8ac4a6, 0xf381caaf, 0xb8e890d8, 0xb5e39ed1, 0xa2fe8cca, 0xaff582c3, 0x8cc4a8fc, 0x81cfa6f5, 0x96d2b4ee, 0x9bd9bae7, 0xbb7bdb3b, 0xb670d532, 0xa16dc729, 0xac66c920, 0x8f57e31f, 0x825ced16, 0x9541ff0d, 0x984af104, 0xd323ab73, 0xde28a57a, 0xc935b761, 0xc43eb968, 0xe70f9357, 0xea049d5e, 0xfd198f45, 0xf012814c, 0x6bcb3bab, 0x66c035a2, 0x71dd27b9, 0x7cd629b0, 0x5fe7038f, 0x52ec0d86, 0x45f11f9d, 0x48fa1194, 0x03934be3, 0x0e9845ea, 0x198557f1, 0x148e59f8, 0x37bf73c7, 0x3ab47dce, 0x2da96fd5, 0x20a261dc, 0x6df6ad76, 0x60fda37f, 0x77e0b164, 0x7aebbf6d, 0x59da9552, 0x54d19b5b, 0x43cc8940, 0x4ec78749, 0x05aedd3e, 0x08a5d337, 0x1fb8c12c, 0x12b3cf25, 0x3182e51a, 0x3c89eb13, 0x2b94f908, 0x269ff701, 0xbd464de6, 0xb04d43ef, 0xa75051f4, 0xaa5b5ffd, 0x896a75c2, 0x84617bcb, 0x937c69d0, 0x9e7767d9, 0xd51e3dae, 0xd81533a7, 0xcf0821bc, 0xc2032fb5, 0xe132058a, 0xec390b83, 0xfb241998, 0xf62f1791, 0xd68d764d, 0xdb867844, 0xcc9b6a5f, 0xc1906456, 0xe2a14e69, 0xefaa4060, 0xf8b7527b, 0xf5bc5c72, 0xbed50605, 0xb3de080c, 0xa4c31a17, 0xa9c8141e, 0x8af93e21, 0x87f23028, 0x90ef2233, 0x9de42c3a, 0x063d96dd, 0x0b3698d4, 0x1c2b8acf, 0x112084c6, 0x3211aef9, 0x3f1aa0f0, 0x2807b2eb, 0x250cbce2, 0x6e65e695, 0x636ee89c, 0x7473fa87, 0x7978f48e, 0x5a49deb1, 0x5742d0b8, 0x405fc2a3, 0x4d54ccaa, 0xdaf741ec, 0xd7fc4fe5, 0xc0e15dfe, 0xcdea53f7, 0xeedb79c8, 0xe3d077c1, 0xf4cd65da, 0xf9c66bd3, 0xb2af31a4, 0xbfa43fad, 0xa8b92db6, 0xa5b223bf, 0x86830980, 0x8b880789, 0x9c951592, 0x919e1b9b, 0x0a47a17c, 0x074caf75, 0x1051bd6e, 0x1d5ab367, 0x3e6b9958, 0x33609751, 0x247d854a, 0x29768b43, 0x621fd134, 0x6f14df3d, 0x7809cd26, 0x7502c32f, 0x5633e910, 0x5b38e719, 0x4c25f502, 0x412efb0b, 0x618c9ad7, 0x6c8794de, 0x7b9a86c5, 0x769188cc, 0x55a0a2f3, 0x58abacfa, 0x4fb6bee1, 0x42bdb0e8, 0x09d4ea9f, 0x04dfe496, 0x13c2f68d, 0x1ec9f884, 0x3df8d2bb, 0x30f3dcb2, 0x27eecea9, 0x2ae5c0a0, 0xb13c7a47, 0xbc37744e, 0xab2a6655, 0xa621685c, 0x85104263, 0x881b4c6a, 0x9f065e71, 0x920d5078, 0xd9640a0f, 0xd46f0406, 0xc372161d, 0xce791814, 0xed48322b, 0xe0433c22, 0xf75e2e39, 0xfa552030, 0xb701ec9a, 0xba0ae293, 0xad17f088, 0xa01cfe81, 0x832dd4be, 0x8e26dab7, 0x993bc8ac, 0x9430c6a5, 0xdf599cd2, 0xd25292db, 0xc54f80c0, 0xc8448ec9, 0xeb75a4f6, 0xe67eaaff, 0xf163b8e4, 0xfc68b6ed, 0x67b10c0a, 0x6aba0203, 0x7da71018, 0x70ac1e11, 0x539d342e, 0x5e963a27, 0x498b283c, 0x44802635, 0x0fe97c42, 0x02e2724b, 0x15ff6050, 0x18f46e59, 0x3bc54466, 0x36ce4a6f, 0x21d35874, 0x2cd8567d, 0x0c7a37a1, 0x017139a8, 0x166c2bb3, 0x1b6725ba, 0x38560f85, 0x355d018c, 0x22401397, 0x2f4b1d9e, 0x642247e9, 0x692949e0, 0x7e345bfb, 0x733f55f2, 0x500e7fcd, 0x5d0571c4, 0x4a1863df, 0x47136dd6, 0xdccad731, 0xd1c1d938, 0xc6dccb23, 0xcbd7c52a, 0xe8e6ef15, 0xe5ede11c, 0xf2f0f307, 0xfffbfd0e, 0xb492a779, 0xb999a970, 0xae84bb6b, 0xa38fb562, 0x80be9f5d, 0x8db59154, 0x9aa8834f, 0x97a38d46];
        var U4 = [0x00000000, 0x090d0b0e, 0x121a161c, 0x1b171d12, 0x24342c38, 0x2d392736, 0x362e3a24, 0x3f23312a, 0x48685870, 0x4165537e, 0x5a724e6c, 0x537f4562, 0x6c5c7448, 0x65517f46, 0x7e466254, 0x774b695a, 0x90d0b0e0, 0x99ddbbee, 0x82caa6fc, 0x8bc7adf2, 0xb4e49cd8, 0xbde997d6, 0xa6fe8ac4, 0xaff381ca, 0xd8b8e890, 0xd1b5e39e, 0xcaa2fe8c, 0xc3aff582, 0xfc8cc4a8, 0xf581cfa6, 0xee96d2b4, 0xe79bd9ba, 0x3bbb7bdb, 0x32b670d5, 0x29a16dc7, 0x20ac66c9, 0x1f8f57e3, 0x16825ced, 0x0d9541ff, 0x04984af1, 0x73d323ab, 0x7ade28a5, 0x61c935b7, 0x68c43eb9, 0x57e70f93, 0x5eea049d, 0x45fd198f, 0x4cf01281, 0xab6bcb3b, 0xa266c035, 0xb971dd27, 0xb07cd629, 0x8f5fe703, 0x8652ec0d, 0x9d45f11f, 0x9448fa11, 0xe303934b, 0xea0e9845, 0xf1198557, 0xf8148e59, 0xc737bf73, 0xce3ab47d, 0xd52da96f, 0xdc20a261, 0x766df6ad, 0x7f60fda3, 0x6477e0b1, 0x6d7aebbf, 0x5259da95, 0x5b54d19b, 0x4043cc89, 0x494ec787, 0x3e05aedd, 0x3708a5d3, 0x2c1fb8c1, 0x2512b3cf, 0x1a3182e5, 0x133c89eb, 0x082b94f9, 0x01269ff7, 0xe6bd464d, 0xefb04d43, 0xf4a75051, 0xfdaa5b5f, 0xc2896a75, 0xcb84617b, 0xd0937c69, 0xd99e7767, 0xaed51e3d, 0xa7d81533, 0xbccf0821, 0xb5c2032f, 0x8ae13205, 0x83ec390b, 0x98fb2419, 0x91f62f17, 0x4dd68d76, 0x44db8678, 0x5fcc9b6a, 0x56c19064, 0x69e2a14e, 0x60efaa40, 0x7bf8b752, 0x72f5bc5c, 0x05bed506, 0x0cb3de08, 0x17a4c31a, 0x1ea9c814, 0x218af93e, 0x2887f230, 0x3390ef22, 0x3a9de42c, 0xdd063d96, 0xd40b3698, 0xcf1c2b8a, 0xc6112084, 0xf93211ae, 0xf03f1aa0, 0xeb2807b2, 0xe2250cbc, 0x956e65e6, 0x9c636ee8, 0x877473fa, 0x8e7978f4, 0xb15a49de, 0xb85742d0, 0xa3405fc2, 0xaa4d54cc, 0xecdaf741, 0xe5d7fc4f, 0xfec0e15d, 0xf7cdea53, 0xc8eedb79, 0xc1e3d077, 0xdaf4cd65, 0xd3f9c66b, 0xa4b2af31, 0xadbfa43f, 0xb6a8b92d, 0xbfa5b223, 0x80868309, 0x898b8807, 0x929c9515, 0x9b919e1b, 0x7c0a47a1, 0x75074caf, 0x6e1051bd, 0x671d5ab3, 0x583e6b99, 0x51336097, 0x4a247d85, 0x4329768b, 0x34621fd1, 0x3d6f14df, 0x267809cd, 0x2f7502c3, 0x105633e9, 0x195b38e7, 0x024c25f5, 0x0b412efb, 0xd7618c9a, 0xde6c8794, 0xc57b9a86, 0xcc769188, 0xf355a0a2, 0xfa58abac, 0xe14fb6be, 0xe842bdb0, 0x9f09d4ea, 0x9604dfe4, 0x8d13c2f6, 0x841ec9f8, 0xbb3df8d2, 0xb230f3dc, 0xa927eece, 0xa02ae5c0, 0x47b13c7a, 0x4ebc3774, 0x55ab2a66, 0x5ca62168, 0x63851042, 0x6a881b4c, 0x719f065e, 0x78920d50, 0x0fd9640a, 0x06d46f04, 0x1dc37216, 0x14ce7918, 0x2bed4832, 0x22e0433c, 0x39f75e2e, 0x30fa5520, 0x9ab701ec, 0x93ba0ae2, 0x88ad17f0, 0x81a01cfe, 0xbe832dd4, 0xb78e26da, 0xac993bc8, 0xa59430c6, 0xd2df599c, 0xdbd25292, 0xc0c54f80, 0xc9c8448e, 0xf6eb75a4, 0xffe67eaa, 0xe4f163b8, 0xedfc68b6, 0x0a67b10c, 0x036aba02, 0x187da710, 0x1170ac1e, 0x2e539d34, 0x275e963a, 0x3c498b28, 0x35448026, 0x420fe97c, 0x4b02e272, 0x5015ff60, 0x5918f46e, 0x663bc544, 0x6f36ce4a, 0x7421d358, 0x7d2cd856, 0xa10c7a37, 0xa8017139, 0xb3166c2b, 0xba1b6725, 0x8538560f, 0x8c355d01, 0x97224013, 0x9e2f4b1d, 0xe9642247, 0xe0692949, 0xfb7e345b, 0xf2733f55, 0xcd500e7f, 0xc45d0571, 0xdf4a1863, 0xd647136d, 0x31dccad7, 0x38d1c1d9, 0x23c6dccb, 0x2acbd7c5, 0x15e8e6ef, 0x1ce5ede1, 0x07f2f0f3, 0x0efffbfd, 0x79b492a7, 0x70b999a9, 0x6bae84bb, 0x62a38fb5, 0x5d80be9f, 0x548db591, 0x4f9aa883, 0x4697a38d];

        function convertToInt32(bytes) {
            var result = [];
            for (var i = 0; i < bytes.length; i += 4) {
                result.push(
                    (bytes[i    ] << 24) |
                    (bytes[i + 1] << 16) |
                    (bytes[i + 2] <<  8) |
                     bytes[i + 3]
                );
            }
            return result;
        }

        var AES = function(key) {
            if (!(this instanceof AES)) {
                throw Error('AES must be instanitated with `new`');
            }

            Object.defineProperty(this, 'key', {
                value: coerceArray(key, true)
            });

            this._prepare();
        }


        AES.prototype._prepare = function() {

            var rounds = numberOfRounds[this.key.length];
            if (rounds == null) {
                throw new Error('invalid key size (must be 16, 24 or 32 bytes)');
            }

            // encryption round keys
            this._Ke = [];

            // decryption round keys
            this._Kd = [];

            for (var i = 0; i <= rounds; i++) {
                this._Ke.push([0, 0, 0, 0]);
                this._Kd.push([0, 0, 0, 0]);
            }

            var roundKeyCount = (rounds + 1) * 4;
            var KC = this.key.length / 4;

            // convert the key into ints
            var tk = convertToInt32(this.key);

            // copy values into round key arrays
            var index;
            for (var i = 0; i < KC; i++) {
                index = i >> 2;
                this._Ke[index][i % 4] = tk[i];
                this._Kd[rounds - index][i % 4] = tk[i];
            }

            // key expansion (fips-197 section 5.2)
            var rconpointer = 0;
            var t = KC, tt;
            while (t < roundKeyCount) {
                tt = tk[KC - 1];
                tk[0] ^= ((S[(tt >> 16) & 0xFF] << 24) ^
                          (S[(tt >>  8) & 0xFF] << 16) ^
                          (S[ tt        & 0xFF] <<  8) ^
                           S[(tt >> 24) & 0xFF]        ^
                          (rcon[rconpointer] << 24));
                rconpointer += 1;

                // key expansion (for non-256 bit)
                if (KC != 8) {
                    for (var i = 1; i < KC; i++) {
                        tk[i] ^= tk[i - 1];
                    }

                // key expansion for 256-bit keys is "slightly different" (fips-197)
                } else {
                    for (var i = 1; i < (KC / 2); i++) {
                        tk[i] ^= tk[i - 1];
                    }
                    tt = tk[(KC / 2) - 1];

                    tk[KC / 2] ^= (S[ tt        & 0xFF]        ^
                                  (S[(tt >>  8) & 0xFF] <<  8) ^
                                  (S[(tt >> 16) & 0xFF] << 16) ^
                                  (S[(tt >> 24) & 0xFF] << 24));

                    for (var i = (KC / 2) + 1; i < KC; i++) {
                        tk[i] ^= tk[i - 1];
                    }
                }

                // copy values into round key arrays
                var i = 0, r, c;
                while (i < KC && t < roundKeyCount) {
                    r = t >> 2;
                    c = t % 4;
                    this._Ke[r][c] = tk[i];
                    this._Kd[rounds - r][c] = tk[i++];
                    t++;
                }
            }

            // inverse-cipher-ify the decryption round key (fips-197 section 5.3)
            for (var r = 1; r < rounds; r++) {
                for (var c = 0; c < 4; c++) {
                    tt = this._Kd[r][c];
                    this._Kd[r][c] = (U1[(tt >> 24) & 0xFF] ^
                                      U2[(tt >> 16) & 0xFF] ^
                                      U3[(tt >>  8) & 0xFF] ^
                                      U4[ tt        & 0xFF]);
                }
            }
        }

        AES.prototype.encrypt = function(plaintext) {
            if (plaintext.length != 16) {
                throw new Error('invalid plaintext size (must be 16 bytes)');
            }

            var rounds = this._Ke.length - 1;
            var a = [0, 0, 0, 0];

            // convert plaintext to (ints ^ key)
            var t = convertToInt32(plaintext);
            for (var i = 0; i < 4; i++) {
                t[i] ^= this._Ke[0][i];
            }

            // apply round transforms
            for (var r = 1; r < rounds; r++) {
                for (var i = 0; i < 4; i++) {
                    a[i] = (T1[(t[ i         ] >> 24) & 0xff] ^
                            T2[(t[(i + 1) % 4] >> 16) & 0xff] ^
                            T3[(t[(i + 2) % 4] >>  8) & 0xff] ^
                            T4[ t[(i + 3) % 4]        & 0xff] ^
                            this._Ke[r][i]);
                }
                t = a.slice();
            }

            // the last round is special
            var result = createArray(16), tt;
            for (var i = 0; i < 4; i++) {
                tt = this._Ke[rounds][i];
                result[4 * i    ] = (S[(t[ i         ] >> 24) & 0xff] ^ (tt >> 24)) & 0xff;
                result[4 * i + 1] = (S[(t[(i + 1) % 4] >> 16) & 0xff] ^ (tt >> 16)) & 0xff;
                result[4 * i + 2] = (S[(t[(i + 2) % 4] >>  8) & 0xff] ^ (tt >>  8)) & 0xff;
                result[4 * i + 3] = (S[ t[(i + 3) % 4]        & 0xff] ^  tt       ) & 0xff;
            }

            return result;
        }

        AES.prototype.decrypt = function(ciphertext) {
            if (ciphertext.length != 16) {
                throw new Error('invalid ciphertext size (must be 16 bytes)');
            }

            var rounds = this._Kd.length - 1;
            var a = [0, 0, 0, 0];

            // convert plaintext to (ints ^ key)
            var t = convertToInt32(ciphertext);
            for (var i = 0; i < 4; i++) {
                t[i] ^= this._Kd[0][i];
            }

            // apply round transforms
            for (var r = 1; r < rounds; r++) {
                for (var i = 0; i < 4; i++) {
                    a[i] = (T5[(t[ i          ] >> 24) & 0xff] ^
                            T6[(t[(i + 3) % 4] >> 16) & 0xff] ^
                            T7[(t[(i + 2) % 4] >>  8) & 0xff] ^
                            T8[ t[(i + 1) % 4]        & 0xff] ^
                            this._Kd[r][i]);
                }
                t = a.slice();
            }

            // the last round is special
            var result = createArray(16), tt;
            for (var i = 0; i < 4; i++) {
                tt = this._Kd[rounds][i];
                result[4 * i    ] = (Si[(t[ i         ] >> 24) & 0xff] ^ (tt >> 24)) & 0xff;
                result[4 * i + 1] = (Si[(t[(i + 3) % 4] >> 16) & 0xff] ^ (tt >> 16)) & 0xff;
                result[4 * i + 2] = (Si[(t[(i + 2) % 4] >>  8) & 0xff] ^ (tt >>  8)) & 0xff;
                result[4 * i + 3] = (Si[ t[(i + 1) % 4]        & 0xff] ^  tt       ) & 0xff;
            }

            return result;
        }


        /**
         *  Mode Of Operation - Electonic Codebook (ECB)
         */
        var ModeOfOperationECB = function(key) {
            if (!(this instanceof ModeOfOperationECB)) {
                throw Error('AES must be instanitated with `new`');
            }

            this.description = "Electronic Code Block";
            this.name = "ecb";

            this._aes = new AES(key);
        }

        ModeOfOperationECB.prototype.encrypt = function(plaintext) {
            plaintext = coerceArray(plaintext);

            if ((plaintext.length % 16) !== 0) {
                throw new Error('invalid plaintext size (must be multiple of 16 bytes)');
            }

            var ciphertext = createArray(plaintext.length);
            var block = createArray(16);

            for (var i = 0; i < plaintext.length; i += 16) {
                copyArray(plaintext, block, 0, i, i + 16);
                block = this._aes.encrypt(block);
                copyArray(block, ciphertext, i);
            }

            return ciphertext;
        }

        ModeOfOperationECB.prototype.decrypt = function(ciphertext) {
            ciphertext = coerceArray(ciphertext);

            if ((ciphertext.length % 16) !== 0) {
                throw new Error('invalid ciphertext size (must be multiple of 16 bytes)');
            }

            var plaintext = createArray(ciphertext.length);
            var block = createArray(16);

            for (var i = 0; i < ciphertext.length; i += 16) {
                copyArray(ciphertext, block, 0, i, i + 16);
                block = this._aes.decrypt(block);
                copyArray(block, plaintext, i);
            }

            return plaintext;
        }


        /**
         *  Mode Of Operation - Cipher Block Chaining (CBC)
         */
        var ModeOfOperationCBC = function(key, iv) {
            if (!(this instanceof ModeOfOperationCBC)) {
                throw Error('AES must be instanitated with `new`');
            }

            this.description = "Cipher Block Chaining";
            this.name = "cbc";

            if (!iv) {
                iv = createArray(16);

            } else if (iv.length != 16) {
                throw new Error('invalid initialation vector size (must be 16 bytes)');
            }

            this._lastCipherblock = coerceArray(iv, true);

            this._aes = new AES(key);
        }

        ModeOfOperationCBC.prototype.encrypt = function(plaintext) {
            plaintext = coerceArray(plaintext);

            if ((plaintext.length % 16) !== 0) {
                throw new Error('invalid plaintext size (must be multiple of 16 bytes)');
            }

            var ciphertext = createArray(plaintext.length);
            var block = createArray(16);

            for (var i = 0; i < plaintext.length; i += 16) {
                copyArray(plaintext, block, 0, i, i + 16);

                for (var j = 0; j < 16; j++) {
                    block[j] ^= this._lastCipherblock[j];
                }

                this._lastCipherblock = this._aes.encrypt(block);
                copyArray(this._lastCipherblock, ciphertext, i);
            }

            return ciphertext;
        }

        ModeOfOperationCBC.prototype.decrypt = function(ciphertext) {
            ciphertext = coerceArray(ciphertext);

            if ((ciphertext.length % 16) !== 0) {
                throw new Error('invalid ciphertext size (must be multiple of 16 bytes)');
            }

            var plaintext = createArray(ciphertext.length);
            var block = createArray(16);

            for (var i = 0; i < ciphertext.length; i += 16) {
                copyArray(ciphertext, block, 0, i, i + 16);
                block = this._aes.decrypt(block);

                for (var j = 0; j < 16; j++) {
                    plaintext[i + j] = block[j] ^ this._lastCipherblock[j];
                }

                copyArray(ciphertext, this._lastCipherblock, 0, i, i + 16);
            }

            return plaintext;
        }


        /**
         *  Mode Of Operation - Cipher Feedback (CFB)
         */
        var ModeOfOperationCFB = function(key, iv, segmentSize) {
            if (!(this instanceof ModeOfOperationCFB)) {
                throw Error('AES must be instanitated with `new`');
            }

            this.description = "Cipher Feedback";
            this.name = "cfb";

            if (!iv) {
                iv = createArray(16);

            } else if (iv.length != 16) {
                throw new Error('invalid initialation vector size (must be 16 size)');
            }

            if (!segmentSize) { segmentSize = 1; }

            this.segmentSize = segmentSize;

            this._shiftRegister = coerceArray(iv, true);

            this._aes = new AES(key);
        }

        ModeOfOperationCFB.prototype.encrypt = function(plaintext) {
            if ((plaintext.length % this.segmentSize) != 0) {
                throw new Error('invalid plaintext size (must be segmentSize bytes)');
            }

            var encrypted = coerceArray(plaintext, true);

            var xorSegment;
            for (var i = 0; i < encrypted.length; i += this.segmentSize) {
                xorSegment = this._aes.encrypt(this._shiftRegister);
                for (var j = 0; j < this.segmentSize; j++) {
                    encrypted[i + j] ^= xorSegment[j];
                }

                // Shift the register
                copyArray(this._shiftRegister, this._shiftRegister, 0, this.segmentSize);
                copyArray(encrypted, this._shiftRegister, 16 - this.segmentSize, i, i + this.segmentSize);
            }

            return encrypted;
        }

        ModeOfOperationCFB.prototype.decrypt = function(ciphertext) {
            if ((ciphertext.length % this.segmentSize) != 0) {
                throw new Error('invalid ciphertext size (must be segmentSize bytes)');
            }

            var plaintext = coerceArray(ciphertext, true);

            var xorSegment;
            for (var i = 0; i < plaintext.length; i += this.segmentSize) {
                xorSegment = this._aes.encrypt(this._shiftRegister);

                for (var j = 0; j < this.segmentSize; j++) {
                    plaintext[i + j] ^= xorSegment[j];
                }

                // Shift the register
                copyArray(this._shiftRegister, this._shiftRegister, 0, this.segmentSize);
                copyArray(ciphertext, this._shiftRegister, 16 - this.segmentSize, i, i + this.segmentSize);
            }

            return plaintext;
        }

        /**
         *  Mode Of Operation - Output Feedback (OFB)
         */
        var ModeOfOperationOFB = function(key, iv) {
            if (!(this instanceof ModeOfOperationOFB)) {
                throw Error('AES must be instanitated with `new`');
            }

            this.description = "Output Feedback";
            this.name = "ofb";

            if (!iv) {
                iv = createArray(16);

            } else if (iv.length != 16) {
                throw new Error('invalid initialation vector size (must be 16 bytes)');
            }

            this._lastPrecipher = coerceArray(iv, true);
            this._lastPrecipherIndex = 16;

            this._aes = new AES(key);
        }

        ModeOfOperationOFB.prototype.encrypt = function(plaintext) {
            var encrypted = coerceArray(plaintext, true);

            for (var i = 0; i < encrypted.length; i++) {
                if (this._lastPrecipherIndex === 16) {
                    this._lastPrecipher = this._aes.encrypt(this._lastPrecipher);
                    this._lastPrecipherIndex = 0;
                }
                encrypted[i] ^= this._lastPrecipher[this._lastPrecipherIndex++];
            }

            return encrypted;
        }

        // Decryption is symetric
        ModeOfOperationOFB.prototype.decrypt = ModeOfOperationOFB.prototype.encrypt;


        /**
         *  Counter object for CTR common mode of operation
         */
        var Counter = function(initialValue) {
            if (!(this instanceof Counter)) {
                throw Error('Counter must be instanitated with `new`');
            }

            // We allow 0, but anything false-ish uses the default 1
            if (initialValue !== 0 && !initialValue) { initialValue = 1; }

            if (typeof(initialValue) === 'number') {
                this._counter = createArray(16);
                this.setValue(initialValue);

            } else {
                this.setBytes(initialValue);
            }
        }

        Counter.prototype.setValue = function(value) {
            if (typeof(value) !== 'number' || parseInt(value) != value) {
                throw new Error('invalid counter value (must be an integer)');
            }

            for (var index = 15; index >= 0; --index) {
                this._counter[index] = value % 256;
                value = value >> 8;
            }
        }

        Counter.prototype.setBytes = function(bytes) {
            bytes = coerceArray(bytes, true);

            if (bytes.length != 16) {
                throw new Error('invalid counter bytes size (must be 16 bytes)');
            }

            this._counter = bytes;
        };

        Counter.prototype.increment = function() {
            for (var i = 15; i >= 0; i--) {
                if (this._counter[i] === 255) {
                    this._counter[i] = 0;
                } else {
                    this._counter[i]++;
                    break;
                }
            }
        }


        /**
         *  Mode Of Operation - Counter (CTR)
         */
        var ModeOfOperationCTR = function(key, counter) {
            if (!(this instanceof ModeOfOperationCTR)) {
                throw Error('AES must be instanitated with `new`');
            }

            this.description = "Counter";
            this.name = "ctr";

            if (!(counter instanceof Counter)) {
                counter = new Counter(counter)
            }

            this._counter = counter;

            this._remainingCounter = null;
            this._remainingCounterIndex = 16;

            this._aes = new AES(key);
        }

        ModeOfOperationCTR.prototype.encrypt = function(plaintext) {
            var encrypted = coerceArray(plaintext, true);

            for (var i = 0; i < encrypted.length; i++) {
                if (this._remainingCounterIndex === 16) {
                    this._remainingCounter = this._aes.encrypt(this._counter._counter);
                    this._remainingCounterIndex = 0;
                    this._counter.increment();
                }
                encrypted[i] ^= this._remainingCounter[this._remainingCounterIndex++];
            }

            return encrypted;
        }

        // Decryption is symetric
        ModeOfOperationCTR.prototype.decrypt = ModeOfOperationCTR.prototype.encrypt;


        ///////////////////////
        // Padding

        // See:https://tools.ietf.org/html/rfc2315
        function pkcs7pad(data) {
            data = coerceArray(data, true);
            var padder = 16 - (data.length % 16);
            var result = createArray(data.length + padder);
            copyArray(data, result);
            for (var i = data.length; i < result.length; i++) {
                result[i] = padder;
            }
            return result;
        }

        function pkcs7strip(data) {
            data = coerceArray(data, true);
            if (data.length < 16) { throw new Error('PKCS#7 invalid length'); }

            var padder = data[data.length - 1];
            if (padder > 16) { throw new Error('PKCS#7 padding byte out of range'); }

            var length = data.length - padder;
            for (var i = 0; i < padder; i++) {
                if (data[length + i] !== padder) {
                    throw new Error('PKCS#7 invalid padding byte');
                }
            }

            var result = createArray(length);
            copyArray(data, result, 0, 0, length);
            return result;
        }

        ///////////////////////
        // Exporting


        // The block cipher
        var aesjs = {
            AES: AES,
            Counter: Counter,

            ModeOfOperation: {
                ecb: ModeOfOperationECB,
                cbc: ModeOfOperationCBC,
                cfb: ModeOfOperationCFB,
                ofb: ModeOfOperationOFB,
                ctr: ModeOfOperationCTR
            },

            utils: {
                hex: convertHex,
                utf8: convertUtf8
            },

            padding: {
                pkcs7: {
                    pad: pkcs7pad,
                    strip: pkcs7strip
                }
            },

            _arrayTest: {
                coerceArray: coerceArray,
                createArray: createArray,
                copyArray: copyArray,
            }
        };



        // If there was an existing library at "aesjs" make sure it's still available
        if (root.aesjs) {
            aesjs._aesjs = root.aesjs;
        }

        root.aesjs = aesjs;


    })(root);


    /* CBOR code from https://github.com/paroga/cbor-js/blob/65dc49611107db83aff8308a6b381f4d7933824b/cbor.js */
    /*
    * The MIT License (MIT)
    *
    * Copyright (c) 2014-2016 Patrick Gansterer <paroga@paroga.com>
    *
    * Permission is hereby granted, free of charge, to any person obtaining a copy
    * of this software and associated documentation files (the "Software"), to deal
    * in the Software without restriction, including without limitation the rights
    * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    * copies of the Software, and to permit persons to whom the Software is
    * furnished to do so, subject to the following conditions:
    *
    * The above copyright notice and this permission notice shall be included in all
    * copies or substantial portions of the Software.
    *
    * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    * SOFTWARE.
    */

    (function(global, undefined) {
        "use strict";
        var POW_2_24 = 5.960464477539063e-8,
        POW_2_32 = 4294967296,
        POW_2_53 = 9007199254740992;

        function encode(value) {
            var data = new ArrayBuffer(256);
            var dataView = new DataView(data);
            var lastLength;
            var offset = 0;

            function prepareWrite(length) {
                var newByteLength = data.byteLength;
                var requiredLength = offset + length;
                while (newByteLength < requiredLength)
                    newByteLength <<= 1;
                if (newByteLength !== data.byteLength) {
                    var oldDataView = dataView;
                    data = new ArrayBuffer(newByteLength);
                    dataView = new DataView(data);
                    var uint32count = (offset + 3) >> 2;
                    for (var i = 0; i < uint32count; ++i)
                        dataView.setUint32(i << 2, oldDataView.getUint32(i << 2));
                }

                lastLength = length;
                return dataView;
            }

            function commitWrite() {
                offset += lastLength;
            }

            function writeFloat64(value) {
                commitWrite(prepareWrite(8).setFloat64(offset, value));
            }

            function writeUint8(value) {
                commitWrite(prepareWrite(1).setUint8(offset, value));
            }

            function writeUint8Array(value) {
                var dataView = prepareWrite(value.length);
                for (var i = 0; i < value.length; ++i)
                    dataView.setUint8(offset + i, value[i]);
                commitWrite();
            }

            function writeUint16(value) {
                commitWrite(prepareWrite(2).setUint16(offset, value));
            }

            function writeUint32(value) {
                commitWrite(prepareWrite(4).setUint32(offset, value));
            }

            function writeUint64(value) {
                var low = value % POW_2_32;
                var high = (value - low) / POW_2_32;
                var dataView = prepareWrite(8);
                dataView.setUint32(offset, high);
                dataView.setUint32(offset + 4, low);
                commitWrite();
            }

            function writeTypeAndLength(type, length) {
                if (length < 24) {
                    writeUint8(type << 5 | length);
                } else if (length < 0x100) {
                    writeUint8(type << 5 | 24);
                    writeUint8(length);
                } else if (length < 0x10000) {
                    writeUint8(type << 5 | 25);
                    writeUint16(length);
                } else if (length < 0x100000000) {
                    writeUint8(type << 5 | 26);
                    writeUint32(length);
                } else {
                    writeUint8(type << 5 | 27);
                    writeUint64(length);
                }
            }

            function encodeItem(value) {
                var i;

                if (value === false)
                    return writeUint8(0xf4);
                if (value === true)
                    return writeUint8(0xf5);
                if (value === null)
                    return writeUint8(0xf6);
                if (value === undefined)
                    return writeUint8(0xf7);

                switch (typeof value) {
                    case "number":
                    if (Math.floor(value) === value) {
                        if (0 <= value && value <= POW_2_53)
                            return writeTypeAndLength(0, value);
                        if (-POW_2_53 <= value && value < 0)
                            return writeTypeAndLength(1, -(value + 1));
                    }
                    writeUint8(0xfb);
                    return writeFloat64(value);

                    case "string":
                    var utf8data = [];
                    for (i = 0; i < value.length; ++i) {
                        var charCode = value.charCodeAt(i);
                        if (charCode < 0x80) {
                            utf8data.push(charCode);
                        } else if (charCode < 0x800) {
                            utf8data.push(0xc0 | charCode >> 6);
                            utf8data.push(0x80 | charCode & 0x3f);
                        } else if (charCode < 0xd800) {
                            utf8data.push(0xe0 | charCode >> 12);
                            utf8data.push(0x80 | (charCode >> 6) & 0x3f);
                            utf8data.push(0x80 | charCode & 0x3f);
                        } else {
                            charCode = (charCode & 0x3ff) << 10;
                            charCode |= value.charCodeAt(++i) & 0x3ff;
                            charCode += 0x10000;

                            utf8data.push(0xf0 | charCode >> 18);
                            utf8data.push(0x80 | (charCode >> 12) & 0x3f);
                            utf8data.push(0x80 | (charCode >> 6) & 0x3f);
                            utf8data.push(0x80 | charCode & 0x3f);
                        }
                    }

                    writeTypeAndLength(3, utf8data.length);
                    return writeUint8Array(utf8data);

                    default:
                    var length;
                    if (Array.isArray(value)) {
                        length = value.length;
                        writeTypeAndLength(4, length);
                        for (i = 0; i < length; ++i)
                            encodeItem(value[i]);
                    } else if (value instanceof Uint8Array) {
                        writeTypeAndLength(2, value.length);
                        writeUint8Array(value);
                    } else {
                        var keys = Object.keys(value);
                        length = keys.length;
                        writeTypeAndLength(5, length);
                        for (i = 0; i < length; ++i) {
                            var key = keys[i];
                            encodeItem(key);
                            encodeItem(value[key]);
                        }
                    }
                }
            }

            encodeItem(value);

            if ("slice" in data)
                return data.slice(0, offset);

            var ret = new ArrayBuffer(offset);
            var retView = new DataView(ret);
            for (var i = 0; i < offset; ++i)
                retView.setUint8(i, dataView.getUint8(i));
            return ret;
        }

        function decode(data, tagger, simpleValue) {
            var dataView = new DataView(data);
            var offset = 0;

            if (typeof tagger !== "function")
                tagger = function(value) { return value; };
            if (typeof simpleValue !== "function")
                simpleValue = function() { return undefined; };

            function commitRead(length, value) {
                offset += length;
                return value;
            }

            function readArrayBuffer(length) {
                return commitRead(length, new Uint8Array(data, offset, length));
            }

            function readFloat16() {
                var tempArrayBuffer = new ArrayBuffer(4);
                var tempDataView = new DataView(tempArrayBuffer);
                var value = readUint16();

                var sign = value & 0x8000;
                var exponent = value & 0x7c00;
                var fraction = value & 0x03ff;

                if (exponent === 0x7c00)
                    exponent = 0xff << 10;
                else if (exponent !== 0)
                    exponent += (127 - 15) << 10;
                else if (fraction !== 0)
                    return (sign ? -1 : 1) * fraction * POW_2_24;

                tempDataView.setUint32(0, sign << 16 | exponent << 13 | fraction << 13);
                return tempDataView.getFloat32(0);
            }

            function readFloat32() {
                return commitRead(4, dataView.getFloat32(offset));
            }

            function readFloat64() {
                return commitRead(8, dataView.getFloat64(offset));
            }

            function readUint8() {
                return commitRead(1, dataView.getUint8(offset));
            }

            function readUint16() {
                return commitRead(2, dataView.getUint16(offset));
            }

            function readUint32() {
                return commitRead(4, dataView.getUint32(offset));
            }

            function readUint64() {
                return readUint32() * POW_2_32 + readUint32();
            }

            function readBreak() {
                if (dataView.getUint8(offset) !== 0xff)
                    return false;
                offset += 1;
                return true;
            }

            function readLength(additionalInformation) {
                if (additionalInformation < 24)
                    return additionalInformation;
                if (additionalInformation === 24)
                    return readUint8();
                if (additionalInformation === 25)
                    return readUint16();
                if (additionalInformation === 26)
                    return readUint32();
                if (additionalInformation === 27)
                    return readUint64();
                if (additionalInformation === 31)
                    return -1;
                throw "Invalid length encoding";
            }

            function readIndefiniteStringLength(majorType) {
                var initialByte = readUint8();
                if (initialByte === 0xff)
                    return -1;
                var length = readLength(initialByte & 0x1f);
                if (length < 0 || (initialByte >> 5) !== majorType)
                    throw "Invalid indefinite length element";
                return length;
            }

            function appendUtf16Data(utf16data, length) {
                for (var i = 0; i < length; ++i) {
                    var value = readUint8();
                    if (value & 0x80) {
                        if (value < 0xe0) {
                            value = (value & 0x1f) << 6 |
                            (readUint8() & 0x3f);
                            length -= 1;
                        } else if (value < 0xf0) {
                            value = (value & 0x0f) << 12 |
                            (readUint8() & 0x3f) << 6 |
                            (readUint8() & 0x3f);
                            length -= 2;
                        } else {
                            value = (value & 0x0f) << 18 |
                            (readUint8() & 0x3f) << 12 |
                            (readUint8() & 0x3f) << 6 |
                            (readUint8() & 0x3f);
                            length -= 3;
                        }
                    }

                    if (value < 0x10000) {
                        utf16data.push(value);
                } else {
                    value -= 0x10000;
                    utf16data.push(0xd800 | (value >> 10));
                    utf16data.push(0xdc00 | (value & 0x3ff));
                }
            }
        }

        function decodeItem() {
            var initialByte = readUint8();
            var majorType = initialByte >> 5;
            var additionalInformation = initialByte & 0x1f;
            var i;
            var length;

            if (majorType === 7) {
                switch (additionalInformation) {
                    case 25:
                        return readFloat16();
                    case 26:
                        return readFloat32();
                    case 27:
                        return readFloat64();
                }
            }

            length = readLength(additionalInformation);
            if (length < 0 && (majorType < 2 || 6 < majorType))
                throw "Invalid length";

            switch (majorType) {
                case 0:
                    return length;
                case 1:
                    return -1 - length;
                case 2:
                    if (length < 0) {
                        var elements = [];
                        var fullArrayLength = 0;
                        while ((length = readIndefiniteStringLength(majorType)) >= 0) {
                            fullArrayLength += length;
                            elements.push(readArrayBuffer(length));
                        }
                        var fullArray = new Uint8Array(fullArrayLength);
                        var fullArrayOffset = 0;
                        for (i = 0; i < elements.length; ++i) {
                            fullArray.set(elements[i], fullArrayOffset);
                            fullArrayOffset += elements[i].length;
                        }
                        return fullArray;
                    }
                    return readArrayBuffer(length);
                case 3:
                    var utf16data = [];
                    if (length < 0) {
                        while ((length = readIndefiniteStringLength(majorType)) >= 0)
                            appendUtf16Data(utf16data, length);
                    } else
                        appendUtf16Data(utf16data, length);
                    return String.fromCharCode.apply(null, utf16data);
                case 4:
                    var retArray;
                    if (length < 0) {
                        retArray = [];
                        while (!readBreak())
                            retArray.push(decodeItem());
                    } else {
                        retArray = new Array(length);
                        for (i = 0; i < length; ++i)
                            retArray[i] = decodeItem();
                    }
                    return retArray;
                case 5:
                    var retObject = {};
                    for (i = 0; i < length || length < 0 && !readBreak(); ++i) {
                        var key = decodeItem();
                        retObject[key] = decodeItem();
                    }
                    return retObject;
                case 6:
                    return tagger(decodeItem(), length);
                case 7:
                    switch (length) {
                        case 20:
                            return false;
                        case 21:
                            return true;
                        case 22:
                            return null;
                        case 23:
                            return undefined;
                        default:
                            return simpleValue(length);
                    }
            }
        }

        var ret = decodeItem();
        if (offset !== data.byteLength)
            throw "Remaining bytes";
        return ret;
    }

    var obj = { encode: encode, decode: decode };

    if (typeof define === "function" && define.amd)
        define("cbor/cbor", obj);
    else if (typeof module !== "undefined" && module.exports)
        module.exports = obj;
    else if (!global.CBOR)
        global.CBOR = obj;

    })(root);

    /* Thingsquare thsq.js code starts here */
    function VirtualDevice(options) {
        EventEmitter.call(this);

        this.options = options;
        if (this.options === undefined) {
            this.options = {};
        }
        this.dvars = {};
        this.svars = {};
        this.ivars = {};

        this.closed = false;
        if (this.options.eui) {
            this.eui = this.options.eui;
        } else {
            this.eui = Math.random().toString(36).substr(2, 10);
        }
        this.scrt = this.options.scrt;
        this.xid = this.options.xid;
        this.mode = this.options.mode;
        this.networks = this.options.networks;

        this.proptime = this.options.proptime || 0;

        if (this.mode === 'feather' || this.mode === 'deadleaf') {
            this.ping();
        } else {
            this.connect();
        }

        var _this = this;
        _this.heartbeat = setTimeout(heartbeatfunction, 8000 * Math.random());
        function heartbeatfunction() {
            if (_this.options.heartbeat !== false) {
                if (_this.auth && _this.options.nearby !== false) {
                    addNearbyDevice(thsq, _this.auth.toString());
                }
            }
            if (_this.mode === 'feather' || _this.mode === 'deadleaf') {
                if (_this.nextping && Date.now() > _this.nextping) {
                    delete _this.nextping;
                    _this.sendping();
                }
            } else {
                _this.connect();
            }
            _this.heartbeat = setTimeout(heartbeatfunction, 8000 * Math.random());
        }

        if (this.xid && this.eui) {
            if (this.networks === undefined) {
                this.networks = {};
            }
            if (this.networks[this.xid] === undefined) {
                this.networks[this.xid] = {};
            }
            this.networks[this.xid][this.eui] = this;
            if (this.mode === 'master') {
                this.networks[this.xid].master = this;
            } else {
                this.parenteui = this.networks[this.xid].master.eui;
            }
        }
    }

    VirtualDevice.prototype = new EventEmitter();

    VirtualDevice.prototype.setServerVariable = function (key, val, important) {
        if (important) {
            this.ivars[key] = val;
        } else {
            this.svars[key] = val;
        }

    };

    function senddata(d, type, key, val) {
        var array, i;

        if (!val) {
            val = '';
        }
        array = new Uint8Array(1 + key.length + 1 + 2 + val.length);
        array[0] = type.charCodeAt(0);
        for (i = 0; i < key.length; i += 1) {
            array[1 + i] = key.charCodeAt(i);
        }
        array[1 + key.length] = ','.charCodeAt(0);
        array[1 + key.length + 1] = val.length / 256;
        array[1 + key.length + 2] = val.length & 255;
        for (i = 0; i < val.length; i += 1) {
            if (typeof Buffer !== 'undefined' && Buffer.isBuffer(val)) {
                array[1 + key.length + 3 + i] = val[i];
            } else if ((val.constructor !== String)) {
                array[1 + key.length + 3 + i] = val[i];
            } else {
                array[1 + key.length + 3 + i] = val.charCodeAt(i);
            }
        }

        if (d.connection) {
            if (type === 'c') {
                return false;
            }
            if (nodejs) {
                d.connection.sendBytes(Buffer.from(array));
            } else if (browser) {
                if (d.connection.readyState === WebSocket.OPEN) {
                    d.connection.send(array.buffer);
                    return true;
                }
            }
        } else if (d.mode === 'feather' || d.mode === 'deadleaf') {
            if (!d.scrt && type !== 'p') {
                /* Only send pings via feather before we have a scrt */
                return false;
            }

            if (d.eui !== undefined && d.networks !== undefined &&
                    d.networks[d.xid] && d.networks[d.xid].master) {
                if (nodejs) {
                    forwardFeatherMessage(d.networks[d.xid].master, d.eui, Buffer.from(array));
                } else {
                    forwardFeatherMessage(d.networks[d.xid].master, d.eui, array);
                }
                return true;
            }
        } else {
            // not connected
            d.connect();
            return false;
        }
    }
    VirtualDevice.prototype.pushVariable = function (type, key, val) {
        return senddata(this, type, key, val);
    };

    VirtualDevice.prototype.pushServerVariables = function () {
        var i, k;
        var variables;

        if (!this.scrt && !this.connected) {
            /* Only send pings via feather before we have a scrt */
            return false;
        }

        /* Send the important s variables first */
        variables = Object.keys(this.ivars);

        variables = variables.sort(function (a, b) {
            if (a < b) {
                return -1;
            }
            if (a > b) {
                return 1;
            }
            return 0;
        });
        for (i = 0; i < variables.length; i += 1) {
            k = variables[i];
            if (this.ivars[k]) {
                if (this.pushVariable('i', k, this.ivars[k])) {
                    delete this.ivars[k];
                }
            }
        }

        /* Check if we need to retransmit something */
        if (Object.keys(this.ivars).length > 0) {
            var _this = this;
            if (this.rexmittimeout === undefined) {
                this.rexmittimeout = setTimeout(function () {
                    _this.pushServerVariables();
                    delete this.rexmittimeout;
                }, 1000);
            }
        }

        /* Then the regular s variables */
        variables = Object.keys(this.svars);

        variables = variables.sort(function (a, b) {
            if (a < b) {
                return -1;
            }
            if (a > b) {
                return 1;
            }
            return 0;
        });
        for (i = 0; i < variables.length; i += 1) {
            k = variables[i];
            if (this.svars[k]) {
                this.pushVariable('s', k, this.svars[k]);
            }
        }
        this.svars = {};
    };

    VirtualDevice.prototype.pushDeviceVariables = function () {
        var msg, k;

        msg = { d: this.dvars };
        for (k in this.dvars) {
            if (this.dvars.hasOwnProperty(k) && this.dvars[k]) {
                this.pushVariable('d', k, this.dvars[k]);
            }
        }
    };

    VirtualDevice.prototype.getVariable = function (key) {
        return this.dvars[key];
    };

    VirtualDevice.prototype.setVariable = function (key, val) {
        var oldval = this.dvars[key];
        this.dvars[key] = val;
        this.emit('variable', this, key, val, oldval);
        if ((this.connection && this.connected) || this.needsdone === false) {
            this.pushDeviceVariables();
        }
    };


    VirtualDevice.prototype.ping = function (msg) {
        this.nextping = Date.now() + Math.random() * 10000;
        this.nextpingmsg = msg;
        this.pings = 0;
        this.needsdone = true;
        this.sendping();
    };
    VirtualDevice.prototype.sendping = function () {
        if (this.needsdone === false) {
            return;
        }

        var msg = this.nextpingmsg;
        if (msg === undefined) {
            msg = String(Date.now());
        }
        var pingtext = msg + ',0,0';

        if (this.scrt) {
            var hash;
            if (nodejs) {
                hash = crypto.createHmac('sha1', this.scrt).update(pingtext).digest('hex');
                pingtext += ',' + hash;
            } else {
                var sha1 = createMethod('js-sha1');
                sha1(this.scrt);
                hash = sha1.create();
                hash.update(pingtext);
                pingtext += ',' + hash.hex();
            }
        }
        var _this = this;
        senddata(_this, 'p', pingtext, '');
        this.lastping = Date.now();
        this.pingsent = true;
        var doublings = this.pings;
        if (doublings > 3) {
            doublings = 3;
        }
        this.nextping = Date.now() + (10000 << doublings) + Math.random() * (10000 << doublings);
        this.pings += 1;
    };

    function forwardFeatherMessage(dev, eui, msgbuffer) {
        setTimeout(function () {
            dev.pushVariable('f', eui, msgbuffer);
        }, dev.proptime);
    }

    function domd5(value) {
        var valuemd5;
        if (crypto && typeof crypto === 'object' && typeof crypto.createHash === 'function') {
            if (value) {
                valuemd5 = crypto.createHash('md5').update(value).digest();
            } else {
                valuemd5 = crypto.createHash('md5').digest();
            }
        } else {
            valuemd5 = md5(value, null, true);
        }
        return valuemd5;
    }

    function receiveMsg(dev, message) {

        var handle_command = function (dev, command, data, avoidack) {
            var key, value, valuemd5, i, length, eui, oldval;

            if (command !== 'fthr' && command !== 'push' &&
                command !== 'sync' && command !== 'fwup' &&
                command !== 'zwup' && command !== 'zwin' &&
                command !== 'pang' && command !== 'pong' &&
                avoidack !== true) {
                senddata(dev, 'c', command.toString(), domd5(message.binaryData));
            }

            if (command === 'bond') {
                var newxid = data.toString();
                if (dev.networks) {
                    if (dev.networks[dev.xid]) {
                        delete dev.networks[dev.xid][dev.eui];
                    }
                    if (dev.networks[newxid] === undefined) {
                        dev.networks[newxid] = {};
                    }
                    if (dev.mode === 'master') {
                        dev.networks[newxid].master = dev;
                    } else {
                        dev.networks[newxid][dev.eui] = dev;
                    }
                }
                if (dev.mode !== 'master') {
                    dev.ping('bond');
                }
                dev.pushVariable('s', '_xid', newxid);
                dev.xid = newxid;

            } else if (command === 'scrt') {
                dev.scrt = data.toString();
            } else if (command === 'auth') {
                dev.auth = data;
                if (dev.options.nearby !== false) {
                    addNearbyDevice(thsq, data.toString());
                }
            } else if (command === 'sync') {
                key = data.toString().split(',')[0];
                value = data.toString().substring(key.length + 1);
                oldval = dev.dvars[key];
                dev.dvars[key] = value;

                senddata(dev, 'm', key, domd5(value));
                dev.emit('variable', dev, key, value, oldval);

                if (key === 'peui') {
                    dev.parenteui = value;
                    if (dev.options.mode === 'feather') {
                        dev.sendStats();
                    }
                }
            } else if (command === 'iack') {
                key = data.toString().split(',')[0];
                delete dev.ivars[key];
            } else if (command === 'push') {
                key = data.toString().split(',')[0];
                value = data.toString().substring(key.length + 1);
                oldval = dev.dvars[key];
                dev.dvars[key] = value;

                valuemd5 = md5(value, null, true);
                senddata(dev, 'A', key, domd5(value));
                dev.emit('variable', dev, key, value, oldval);

                if (key === 'peui') {
                    dev.parenteui = value;
                    if (dev.options.mode === 'feather') {
                        dev.sendStats();
                    }
                }
            } else if (command === 'zraw' && dev.mode === 'master') {
                if (dev.networks && dev.networks[dev.xid]) {
                    for (eui in dev.networks[dev.xid]) {
                        receiveMsg(dev.networks[dev.xid][eui], { type: 'binary', binaryData: data });
                    }
                }
            } else if (command === 'fthr' && dev.mode === 'master') {
                var fmessage;
                eui = data.toString().split(',')[0];
                fmessage = data.slice(eui.length + 1);
                if (dev.networks && dev.networks[dev.xid] &&
                    dev.networks[dev.xid][eui]) {
                    setTimeout(function () {
                        if (nodejs) {
                            receiveMsg(dev.networks[dev.xid][eui], { type: 'binary', binaryData: Buffer.from(fmessage) });
                        } else {
                            receiveMsg(dev.networks[dev.xid][eui], { type: 'binary', binaryData: fmessage });
                        }
                    }, dev.proptime);
                }
            } else if (command === 'xcmd') {

                var parsexcmd = function (xcmd) {
                    var i;
                    var xcmdcommands = [];
                    i = 4;
                    while (i < xcmd.length) {
                        var cmdlen = xcmd[i];
                        if (nodejs) {
                            xcmdcommands.push(Buffer.from(xcmd.subarray(i + 3, i + 3 + cmdlen)));
                        } else {
                            xcmdcommands.push(Uint8Array.from(xcmd.subarray(i + 3, i + 3 + cmdlen)));
                        }
                        i += cmdlen + 3;
                    }
                    return xcmdcommands;
                };


                var prettyprintcommand = function (c) {
                    if (c.toString().substring(0, 4) === 'xcmd') {
                        return 'xcmd (' + parsexcmd(c).map(prettyprintcommand).join(', ') + ')';
                    } else if (c.toString().substring(0, 4) === 'iack') {
                        return c.toString().split(',')[0];
                    } else {
                        return c.toString().substring(0, 4);
                    }
                };

                var xcmdcommands;
                if (nodejs) {
                    xcmdcommands = parsexcmd(Buffer.concat([ Buffer.from('xcmd'), Buffer.from(data) ]));
                } else {
                    var concat = function (a, b) {
                        var c = new a.constructor(a.length + b.length);
                        c.set(a);
                        c.set(b, a.length);
                        return c;
                    };
                    xcmdcommands = parsexcmd(concat(Uint8Array.from('xcmd'), Uint8Array.from(data)));
                }

                for (i = 0; i < xcmdcommands.length; i += 1) {
                    handle_command(dev, xcmdcommands[i].toString().substring(0, 4), xcmdcommands[i].slice(4), true);
                }

            } else if (command === 'stat') {
                dev.sendStats(data);
            } else if (command === 'pang') {
                dev.connect();
            } else if (command === 'ping') {
                dev.ping();
            } else if (command === 'pong') {
                dev.pingsent = false;
                key = data.toString().split(',')[0];
                senddata(dev, 'P', key, '');
                dev.pushServerVariables();

            } else if (command === 'done') {
                dev.needsdone = false;
                dev.pushServerVariables();
                dev.pushDeviceVariables();
            }
            dev.emit('command', dev, command, data);
        };

        var command, data;
        if (message.type === 'utf8') {

            command = message.utf8Data.substring(0, 4);
            data = message.utf8Data.substring(4);

        } else if (message.type === 'binary') {
            command = message.binaryData.toString('utf8').slice(0, 4);
            data = message.binaryData.slice(4);
        }
        handle_command(dev, command, data);

    }


    VirtualDevice.prototype.getCookie = function () {
        var host, cookie;
        if (this.options.cookie) {
            cookie = this.options.cookie;
        } else {
            host = thsq.getServer();
            if (jar) {
                cookie = jar.getCookieString(host);
            }
        }
        return cookie;
    };

    VirtualDevice.prototype.sendStats = function (variable) {
        this.setServerVariable('p', this.parenteui + ',-10,256,512,0,-10', true);
        this.pushServerVariables();
    };


    VirtualDevice.prototype.getEUI = function () {
        return this.eui;
    };

    VirtualDevice.prototype.connect = function () {
        var _this, timeout;
        _this = this;

        if (this.connected === true || this.connecting) {
            return;
        }

        this.connecting = true;

        timeout = 0;

        function onmessage(message) {
            if (message.data) {
                // For when running in browser
                var msg = { type: 'binary',
                            binaryData: String.fromCharCode.apply(null, new Uint8Array(message.data)) };
                message = msg;
            }

            receiveMsg(_this, message);
        }

        function doclose(type, event) {
            var wasconnected = false;
            if (_this.connected === true) {
                wasconnected = true;
            }
            _this.connected = false;
            _this.connecting = false;
            _this.lasterror = type + ' ' + JSON.stringify(event);

            if (_this.socket) {
                if (typeof _this.socket.removeAllListeners === 'function') {
                    _this.socket.removeAllListeners();
                }
                delete _this.socket;
            }
            if (_this.connection) {
                if (typeof _this.connection.removeAllListeners === 'function') {
                    _this.connection.removeAllListeners();
                }
                delete _this.connection;
            }

            if (wasconnected) {
                _this.emit('server-disconnected', type, event);
            }

            if (_this.mode === 'master' || _this.mode === 'connected') {
                _this.connect();
            }
        }

        function onclose(event) {
            doclose('close', event);
        }
        function onerror(event) {
            doclose('error', event);
        }

        function onopen() {
            var sid;
            _this.connected = true;
            sid = '';
            sid += 'eui:' + _this.eui + ',';
            if (_this.options.mode) {
                sid += 'default-mode:' + _this.options.mode + ',';
            }
            if (_this.options.platform) {
                sid += 'platform:' + _this.options.platform + ',';
            } else {
                sid += 'platform:thsq-virtual,';
            }
            if (npm_pkg_version === undefined) {
                sid += 'fwid:4.7.0,';
            } else {
                sid += 'fwid:4.7.' + npm_pkg_version + ',';
            }
            _this.setServerVariable('id', sid, true);

            if (_this.options.location) {
                _this.setServerVariable('location', _this.options.location);
            }
            if (_this.options.name) {
                _this.setServerVariable('name', _this.options.name);
            }

            if (_this.lasterror) {
                _this.setServerVariable('lasterror', _this.lasterror);
                delete _this.lasterror;
            }
            if (_this.options.xid) {
                _this.setServerVariable('_xid', _this.options.xid, true);
                _this.pushServerVariables();
            }

            if (_this.options.mode === 'master') {
                _this.setVariable('mode', _this.options.mode);
            }
            if (_this.options.mode === 'feather') {
                _this.sendStats();
            }
            _this.pushDeviceVariables();
            _this.pushServerVariables();
            _this.pushVariable('g', '', '');

            _this.lastping = _this.lastpong = Date.now();

            if (_this.pingtimer === undefined) {
                _this.pingtimer = setInterval(function () {

                    if (nodejs) {
                        /* Kill the connection if we haven't received a pong in 60 seconds. */
                        if (Date.now() - _this.lastpong > 60000) {
                            _this.close('because pong time was ' + (Date.now() - _this.lastpong));
                        }

                        if (_this.connected) {
                            if (_this.connection) {
                                _this.lastping = Date.now();
                                _this.connection.ping();
                            }
                        }
                    }
                }, 30000);
            }

            _this.emit('server-connected');
        }

        function onpong(data) {
            let rtt = Date.now() - _this.lastping;
            _this.lastpong = Date.now();
        }

        setTimeout(function () {
            var protocol, host;

            if (nodejs) {
                if (_this.socket && typeof _this.socket.removeAllListeners === 'function') {
                    _this.socket.removeAllListeners();
                }
                _this.socket = new WebSocketClient();
                _this.socket.on('connectFailed', function(error) {
                    _this.emit('server-disconnected', 'error', error);
                    doclose('error', error);
                });

                _this.socket.on('connect', function(connection) {
                    _this.connection = connection;
                    connection.on('error', onerror);
                    connection.on('close', onclose);
                    connection.on('message', onmessage);
                    connection.on('pong', onpong);
                    onopen();
                });

                host = thsq.getServer();
                protocol = host.split('//')[0];
                var wsprotocol;
                if (protocol === 'https:') {
                    wsprotocol = 'wss://';
                } else if (protocol === 'http:') {
                    wsprotocol = 'ws://';
                }

                var cookie = '', endpoint;
                if (_this.options.deterministicscrt) {

                    if (_this.options.frontend && _this.eui) {
                        endpoint = '/device/' + _this.eui + '/' + _this.options.frontend;
                    } else {
                        throw new Error('Cannot create a device with deterministic scrt without frontend and eui');
                    }
                } else if (_this.options.virtualdevice === false) {
                    endpoint = '/device';
                    if (_this.scrt) {
                        cookie = 'thsqtoken=' + _this.scrt;
                    }
                } else {
                    /* Normal virtual device behavior */
                    cookie = _this.getCookie();
                    _this.options.cookie = cookie;
                    endpoint = '/virtual-device/' + _this.eui;
                }
                _this.socket.connect(wsprotocol + host.split('//')[1] + endpoint, 'thsq', null,
                    { Cookie: cookie });

            } else if (browser) {
                var server = thsq.getServer();
                if (server && server.split('//')[1]) {

                    if (server.split('//')[0] === 'http:') {
                        protocol = 'ws://';
                    } else {
                        protocol = 'wss://';
                    }

                    host = server.split('//')[1];
                } else if (location.protocol === 'file:') {
                    protocol = 'wss://';
                    host = 'developer.thingsquare.com';
                } else {
                    protocol = 'ws://';
                    if (typeof location === 'undefined' || location.protocol === 'https:') {
                        protocol = 'wss://';
                    }
                    if (typeof location !== 'undefined' ) {
                        host = location.hostname;
                    }
                    if (!host.match(/.+\.thingsquare\.com/) && !host.match(/.+\.thsq\.io/) &&
                        !host.match(/.+\.thsq\.local/) &&
                        !location.hostname.match(/(\d+)\.(\d+)\.(\d+)\.(\d+).*/)) {
                        protocol = 'wss://';
                        host = 'developer.thingsquare.com';
                    }
                }

                _this.connection = new WebSocket(protocol + host + '/virtual-device/' + _this.eui, 'thsq');
                _this.connection.binaryType = 'arraybuffer';

                _this.connection.onopen = onopen;
                _this.connection.onerror = onerror;
                _this.connection.onclose = onclose;
                _this.connection.onmessage = onmessage;
            }

        }, timeout);
    };

    VirtualDevice.prototype.close = function () {

        let wasconnected = false;
        if (this.connected === true) {
            wasconnected = true;
        }

        clearTimeout(this.heartbeat);
        clearInterval(this.pingtimer);
        this.closed = true;
        this.connected = false;
        this.connecting = false;
        if (this.socket) {
            if (typeof this.socket.removeAllListeners === 'function') {
                this.socket.removeAllListeners();
            }
            delete this.socket;
        }
        if (this.connection) {
            this.connection.close();
            if (typeof this.connection.removeAllListeners === 'function') {
                this.connection.removeAllListeners();
            }
            delete this.connection;
        }

        if (wasconnected) {
            this.emit('server-disconnected', 'closed');
        }
    };


    function shallowclone(obj) {
        if (obj !== undefined) {
            return {
                d: Object.assign({}, obj.d),
                s: Object.assign({}, obj.s),
                meta: Object.assign({}, obj.meta)
            };
        }
    }


    function clone(obj) {
        if (obj !== undefined) {
            return JSON.parse(JSON.stringify(obj));
        }
    }


    function clonedata(data) {
        if (data) {
            return Object.assign({}, data);
        }
    }

    function clonedevice(dev) {
        var d = { d: {} , s: {}, meta: {} };
        for (var t in d) {
            for (var n in dev[t]) {
                d[t][n] = clonedata(dev[t][n]);
            }
        }
        d.id = dev.id;
        return d;
    }


    function getDeviceFromCache(thsq, unique) {
        if (thsq.devicecache[unique] && thsq.devicecache[unique].device) {
            return thsq.devicecache[unique].device;
        }
    }

    function deviceunique(device) {
        return thsq.deviceunique(device);
    }

    function Thsq() {
        EventEmitter.call(this);

        this.inited = false;
        this.initcallbacks = [];

        this.options = {};

        this.setServer();
        this.app = this.getApp();

        this.fetchqueues = {};
        this.fetching = {};
        this.fetchingnow = {};
        this.fetchingtimeout = {};

        this.usercache = {};

        this.devicecache = {};
        this.devicesfetching = {};

        this.claimeduniqueids = {};
        this.deviceuniqueids = {};
        this.uniquefromids = {};

        this.dauthcache = {};
        this.nearbydevices = {};
        this.nearbyerrordevices = {};
        this.nearbydevicesrssi = {};
    }

    Thsq.prototype = new EventEmitter();

    /* Convenience functions */

    Thsq.prototype.isacked = function (v) {
        function valuesareequal(value1, value2) {
            if (value1 == null && value2 == null) {
                return true;
            }
            return thsq.valuestring(value1) === thsq.valuestring(value2);
        }
        if (v === undefined || v === null) {
            return true;
        }
        if (v.acked && v.syncacked) {
            if (v.time > v.acked && v.time > v.syncacked) {
                return false;
            }
        } else if (v.acked) {
            if (v.time > v.acked) {
                return false;
            }
        } else if (v.syncacked) {
            if (v.time > v.syncacked) {
                return false;
            }
        }
        if (valuesareequal(v.value, v.ackedvalue)) {
            return true;
        }
        return false;
    };

    Thsq.prototype.valuestring = function (value) {
        if (value !== undefined) {
            if (typeof value === 'string') {
                return value;
            } else if (typeof value === 'object' && value !== null && value.data) {
                return String.fromCharCode.apply(null, value.data);
            } else if (Array.isArray(value)) {
                return String.fromCharCode.apply(null, value);
            }
            return String(value);
        }
    };

    Thsq.prototype.valuebuffer = function (val) {
        if (val === undefined) {
            return {
                type: 'Buffer',
                data: [ ]
            };
        }
        if (val && typeof val === 'object' && val.type === 'Buffer') {
            return val;
        }
        if (val && typeof val === 'string') {
            return {
                type: 'Buffer',
                data: Array.from(val).map(function(c) {
                    return c.charCodeAt(0);
                })
            };
        }
        return {
            type: 'Buffer',
            data: [ Number(val) ]
        };
    };

    Thsq.prototype.valuecbor = function (buffervalue) {
        return CBOR.decode(Uint8Array.from(buffervalue.data).buffer);
    };


    Thsq.prototype.devicename = function (device) {
        function raw() {
            if (device && device.s && device.s.name && device.s.name.value) {
                return thsq.valuestring(device.s.name.value);
            }
            if (device && device.s && thsq.devicePlatform(device)) {
                var eui = '';
                if (thsq.deviceEUI(device) && thsq.deviceEUI(device).substring(12, 16)) {
                    eui = ' (' + thsq.deviceEUI(device).substring(12, 16) + ')';
                }
                return thsq.devicePlatform(device) + eui;
            }
            return 'Unknown';
        }

        return raw().replace(/<|>/g, '');
    };


    Thsq.prototype.deviceidstring = function (devicedata, str) {
        var r;
        if (devicedata && devicedata.s && devicedata.s.id && typeof devicedata.s.id.value === 'string') {
            r = devicedata.s.id.value.split(',').filter(function (e) { return e.split(':')[0] === str; });
            if (r.length > 0) {
                return r[0].split(':')[1];
            }
        }
    };

    Thsq.prototype.devicefwver = function (devicedata) {
        var fwver;
        fwver = thsq.deviceidstring(devicedata, 'fwver');
        if (!fwver) {
            if (devicedata && devicedata.s && devicedata.s.fwver) {
                fwver = devicedata.s.fwver.value;
            }
        }
        return Number(fwver);
    };


    Thsq.prototype.devicefwid = function (devicedata) {
        var fwid;
        fwid = thsq.deviceidstring(devicedata, 'fwid');
        if (!fwid) {
            if (devicedata && devicedata.s && devicedata.s.fwid) {
                fwid = devicedata.s.fwid.value;
            }
        }
        return fwid;
    };



    Thsq.prototype.devicefreq = function (devicedata) {
        return thsq.deviceidstring(devicedata, 'freq');
    };

    Thsq.prototype.devicepower = function (devicedata) {
        var power;
        power = thsq.deviceidstring(devicedata, 'power');
        if (!power) {
            if (devicedata && devicedata.s && devicedata.s.power) {
                power = devicedata.s.power.value;
            }
        }
        return power;
    };

    Thsq.prototype.variablevalue = function (device, type, name, fallback) {
        if (device && device[type] && device[type][name] && device[type][name].value !== undefined) {
            return device[type][name].value;
        }
        return fallback;
    };

    Thsq.prototype.variableage = function (device, type, name, fallback) {
        if (device && device[type] && device[type][name] && device[type][name].time !== undefined) {
            return Date.now() - device[type][name].time;
        }
        return fallback;
    };


    Thsq.prototype.variabletime = function (device, type, name, fallback) {
        if (device && device[type] && device[type][name] && device[type][name].time !== undefined) {
            return device[type][name].time;
        }
        return fallback;
    };

    Thsq.prototype.ackedvariablevalue = function (device, type, name, fallback) {
        if (device && device[type] && device[type][name] && device[type][name].ackedvalue !== undefined && device[type][name].ackedvalue !== null) {
            return device[type][name].ackedvalue;
        }
        return fallback;
    };

    Thsq.prototype.variablevaluebuffer = function (device, type, name, fallback) {
        var val = thsq.variablevalue(device, type, name, fallback);
        return thsq.valuebuffer(val);
    };

    Thsq.prototype.ackedvariablevaluebuffer = function (device, type, name, fallback) {
        var val = thsq.ackedvariablevalue(device, type, name, fallback);
        return thsq.valuebuffer(val);
    };

    Thsq.prototype.ackedvariablevaluestring = function (device, type, name, fallback) {
        var val = thsq.ackedvariablevalue(device, type, name, fallback);
        if (val && typeof val === 'object' && val.type === 'Buffer') {
            return String.fromCharCode.apply(null, val.data);
        }
        if (val !== undefined) {
            return val;
        }
        return fallback;
    };

    Thsq.prototype.variablevaluenum = function (device, type, name, fallback) {
        var val = thsq.variablevaluestring(device, type, name, fallback);
        var num = Number(val);
        if (Number.isNaN(num)) {
            return fallback;
        }
        return num;
    };

    Thsq.prototype.variablevaluestring = function (device, type, name, fallback) {
        var val = thsq.variablevalue(device, type, name, fallback);
        if (val && typeof val === 'object' && val.type === 'Buffer') {
            return String.fromCharCode.apply(null, val.data);
        }
        if (val !== undefined) {
            return String(val);
        }
        return fallback;
    };

    Thsq.prototype.variablevaluestringnum = function (device, type, name, num, fallback) {
        var val = thsq.variablevaluestring(device, type, name, '-').split(',')[num];
        if (val === undefined) {
            return fallback;
        }
        return val;
    };


    Thsq.prototype.variablevaluenumbernum = function (device, type, name, num, fallback) {
        var n = Number(thsq.variablevaluestring(device, type, name, '-').split(',')[num]);
        if (Number.isNaN(n)) {
            return fallback;
        }
        return n;
    };

    Thsq.prototype.variablevaluecbor = function (device, type, name, fallback) {
        if (thsq.variablevalue(device, type, name) === undefined) {
            return fallback;
        }
        return CBOR.decode(Uint8Array.from(thsq.variablevaluebuffer(device, type, name, []).data).buffer);
    };

    Thsq.prototype.variablevaluejson = function (device, type, name, fallback) {
        try {
            var o = JSON.parse(thsq.variablevaluestring(device, type, name));
            return o;
        } catch (ignored) {
            return fallback;
        }
    };

    Thsq.prototype.deviceunique = function (device) {
        return thsq.variablevaluestring(device, 'meta', 'unique');
    };


    Thsq.prototype.currentmode = function (device) {

        if (thsq.variablevaluestring(device, 'meta', 'current-mode')) {
            return thsq.variablevaluestring(device, 'meta', 'current-mode').split(',');
        }
        if (thsq.variablevaluestring(device, 'd', 'deadleaf')) {
            return ['unknown', device.d.deadleaf.value];
        }
        return ['unknown', 60];
    };

    Thsq.prototype.reachable = function (device) {
        var now, interval, time, wakeup;

        if (device.meta.wakeup) {
            wakeup = Number(device.meta.wakeup.value);
        } else if (thsq.deviceEUI(device)) {
            wakeup = Number.parseInt(thsq.deviceEUI(device).slice(-8), 16);
        }
        if (thsq.currentmode(device)[0] !== 'deadleaf') {
            return 0;
        }
        if (thsq.currentmode(device)[1]) {
            interval = thsq.currentmode(device)[1];
        } else {
            interval = 60;
        }

        now = Date.now() / 1000;

        time = (interval + (wakeup % interval) - (now % interval)) % interval;

        if (time === 0) {
            time = interval;
        }
        return (now + time) * 1000;
    };



    Thsq.prototype.createVirtualDevice = function (options) {
        return new VirtualDevice(options);
    };

    function updateUserCache(thsq, callback) {
        thsq.getUser(function (data) {
            var wasloggedin;
            thsq.usercache = data;

            thsq.nearbyerrordevices = {};

            wasloggedin = thsq.loggedin;
            thsq.loggedin = thsq.usercache !== undefined && Object.keys(thsq.usercache).length !== 0 && thsq.usercache.login !== '';

            if (wasloggedin === undefined) {
                if (thsq.loggedin === true) {
                    thsq.emit('logged-in');
                } else if (thsq.loggedin === false) {
                    thsq.emit('logged-out');
                    if (thsq.options.browsercache) {
                        delete localStorage['devicecache'];
                        delete localStorage['devicecache-lastseen'];
                    }
                }
            } else if (wasloggedin === true && thsq.loggedin === false) {
                thsq.emit('logged-out');
                if (thsq.options.browsercache) {
                    delete localStorage['devicecache'];
                    delete localStorage['devicecache-lastseen'];
                }
            } else if (wasloggedin === false && thsq.loggedin === true) {
                thsq.emit('logged-in');
            }
            fetchUpdateDeviceCache(thsq, undefined, function () {
                callback(data);
            });
        });
    }

    function fetch(thsq, queue, fetchoptions) {
        var _this;

        _this = thsq;

        if (thsq.fetchqueues[queue] === undefined) {
            thsq.fetchqueues[queue] = [];
        }
        thsq.fetchqueues[queue].push(fetchoptions);

        function queuelen() {
            var q, num;
            num = 0;
            for (q in thsq.fetchqueues) {
                num += thsq.fetchqueues[q].length;
            }

            num += Object.values(_this.fetching).reduce(function (q, a) { if (q) { return a + 1; } else { return a; }}, 0);
            return num;
        }
        function fetchnext() {
            var options, calloptions, method, data;
            if (_this.fetching[queue] !== true) {
                if (_this.fetchqueues[queue]) {
                    options = _this.fetchqueues[queue].shift();
                }
                if (!options) {
                    var listeners = 0;
                    var ev;
                    for (ev in _this.eventNames()) {
                        listeners += _this.listenerCount(_this.eventNames()[ev]);
                    }
                    if (listeners === 0) {
                        if (typeof XMLHttpRequest === 'undefined') {
                            var q;
                            var num = queuelen();
                            for (q in thsq.fetchingnow) {
                                if (thsq.fetchingnow[q] !== undefined) {
                                    num += 1;
                                }
                            }
                            if (num === 0 && _this.options.exit === true) {
                                _this.exit();
                            }
                        }
                    }
                    return;
                }
                if (options) {

                    var timeout = 40000;
                    if (thsq.options.timeout) {
                        timeout = thsq.options.timeout;
                    }

                    _this.fetchingnow[queue] = options;
                    _this.fetching[queue] = true;
                    if (_this.fetchingtimeout[queue]) {
                        clearTimeout(_this.fetchingtimeout[queue]);
                    }
                    _this.fetchingtimeout[queue] = setTimeout(function () {
                        while (_this.fetchqueues[queue] && _this.fetchqueues[queue].length > 0) {
                            options = _this.fetchqueues[queue].shift();
                            if (typeof options.error === 'function') {
                                options.error('timeout');
                            }
                        }
                        _this.fetching[queue] = false;
                        delete _this.fetchingnow[queue];
                        fetchnext();
                    }, timeout + 5000);
                    method = 'GET';
                    if (options.method) {
                        method = options.method;
                    }
                    data = {};
                    if (options.data) {
                        data = options.data;
                    }
                    data.app = _this.app;

                    var t0 = Date.now();

                    calloptions = {
                        url: options.url,
                        xhrFields: { withCredentials: true },
                        formData: options.formData,
                        data: data,
                        method: method,
                        timeout: timeout,
                        success: function (e, status, xhr) {
                            if (options === _this.fetchingnow[queue]) {
                                options.success(e, status, xhr);
                            }
                        },
                        error: function (xhr, status) {
                            if (options === _this.fetchingnow[queue]) {
                                options.error(status, xhr);
                            }
                        },
                        complete: function (xhr, status) {
                            if (options === _this.fetchingnow[queue]) {
                                _this.fetching[queue] = false;
                                delete _this.fetchingnow[queue];
                                clearTimeout(_this.fetchingtimeout[queue]);
                                delete _this.fetchingtimeout[queue];
                                _this.emit('server-loading-done', options.url, method,
                                            { time: Date.now() - t0, status: status, queuelen: queuelen() });
                                fetchnext();
                            }
                        }
                    };
                    _this.emit('server-loading', options.url, method);
                    ajax(calloptions);
                } else {
                    if (_this.fetchingtimeout[queue]) {
                        clearTimeout(_this.fetchingtimeout[queue]);
                    }
                }
            }
        }

        fetchnext();
    }

    Thsq.prototype.setDeviceHistoryFilter = function () {
        throw new Error('obsolete thsq.setDeviceHistoryFilter(), see thsq.addHistoryListener()');
    };

    Thsq.prototype.setServer = function (host, app) {
        if (host) {
            this.host = host;
        } else {
            if (this.options.server) {
                if (this.options.server.indexOf('http') === 0) {
                    this.host = this.options.server;
                } else {
                    this.host = 'https://' + this.options.server;
                }
            } else {
                if (typeof location !== 'undefined' && (location.protocol === 'https:' || location.protocol === 'http:')) {
                    if (!location.hostname.match(/.+\.thingsquare\.com/) && !location.hostname.match(/.+\.thsq\.io/) &&
                        !location.hostname.match(/.+\.thsq\.local/) &&
                        location.hostname.indexOf('.') !== -1 &&
                        !location.hostname.match(/(\d+)\.(\d+)\.(\d+)\.(\d+).*/)) {
                        this.host = 'https://developer.thingsquare.com';
                    } else {
                        this.host = '';
                    }
                } else {
                    this.host = 'https://developer.thingsquare.com';
                }
            }
        }
        if (app) {
            this.app = app;
        } else {
            if (this.options.frontend) {
                this.app = this.options.frontend;
            } else {
                var name;
                this.app = '0ac48bf3-9fab-4bad-8455-e394808eda6b';
                if (typeof location !== 'undefined') {
                    name = location.hostname.split('.')[0];
                    if (name && name.match(/\w+-\w+-\d+/)) {
                        this.app = name;
                    }
                }
            }
        }
    };

    Thsq.prototype.getServer = function () {
        if (this.host) {
            return this.host;
        }

        var port = '';
        if (typeof location !== 'undefined' && location.port && location.port !== 443 && location.port !== 80) {
            port = ':' + location.port;
        }

        return location.protocol + '//' + location.hostname + port;
    };

    Thsq.prototype.getApp = function () {
        var name;
        if (this.app) {
            return this.app;
        }
        name = location.hostname.split('.')[0];
        if (name && name.match(/\w+-\w+-\d+/)) {
            return name;
        }
        return 'b72af87d-42d7-4f09-bf8c-b9f721a3e6ef';
    };

    Thsq.prototype.deviceStateName = function (state) {
        if (state === 0) {
            return 'Not connected';
        }
        if (state === 1) {
            return 'Waiting for channel';
        }
        if (state === 2) {
            return 'Waiting for mesh';
        }
        if (state === 3) {
            return 'Waiting for gateway';
        }
        if (state === 4) {
            return 'Waiting for server';
        }
        if (state === 5) {
            return 'Waiting for Internet';
        }
        if (state === 6) {
            return 'Waiting for connection';
        }
        if (state === 7) {
            return 'Waiting for authentication';
        }
        if (state === 9) {
            return 'Network cable unplugged';
        }
        if (state === 10) {
            return 'Waiting for network';
        }
    };

    function addNearbyDeviceActual(_this, key, state, platform, rssi, data) {
        if (!key) {
            return;
        }
        if (_this.nearbyerrordevices[key]) {
            if ((_this.nearbyerrordevices[key].time - Date.now() + 60000) > 0) {
                return;
            } else {
                delete _this.nearbyerrordevices[key];
            }
        }
        if (platform !== undefined && state !== undefined) {
            _this.emit('device-unknown-nearby-seen', key, platform, state);
        } else if (!fetchingDeviceToCache(_this, key)) {
            var unique;
            if (_this.dauthcache[key]) {
                unique = _this.dauthcache[key];
            }
            if (unique && _this.nearbydevices[unique] && _this.nearbydevices[unique].device &&
                    _this.nearbydevices[unique].time > Date.now() - 4000) {
                if (rssi !== undefined) {
                    _this.nearbydevicesrssi[unique] = rssi;
                }
                _this.emit('device-nearby-seen', _this.nearbydevices[unique].device, unique, state);
            } else {
                fetchDeviceToCache(_this, key, function (device, err) {
                    if (device) {
                        unique = deviceunique(device);
                        _this.dauthcache[key] = unique;
                        if (rssi !== undefined) {
                            _this.nearbydevicesrssi[unique] = rssi;
                        }
                        if (_this.nearbydevices[unique]) {
                            _this.emit('device-nearby-seen', device, unique, state);
                        } else {
                            _this.emit('device-nearby-first-seen', device, unique, state);
                            _this.emit('device-nearby-seen', device, unique, state);
                        }
                        _this.nearbydevices[unique] = {
                            time: Date.now(),
                            id: key,
                            device: device
                        };
                    } else {
                        if (!_this.nearbyerrordevices[key]) {
                            _this.nearbyerrordevices[key] = {
                                time: Date.now(),
                                id: key
                            };
                        }
                    }
                });
            }
        }
    }
    function addNearbyDevice(_this, key, state, platform, rssi, data) {
        setTimeout(function () {
            addNearbyDeviceActual(_this, key, state, platform, rssi, data);
        }, 0);
    }


    Thsq.prototype.setFcmNotification = function (cb) {
        this.fcmCallback = cb;
    };

    Thsq.prototype.fcmFilterOnUsername = function (username) {
        if (typeof AndroidFCMService !== 'undefined') {
            if (username) {
                AndroidFCMService.setCurrentUsername(username);
            } else {
                AndroidFCMService.setCurrentUsername('loggedout');
            }
        }
    };

    Thsq.prototype.getCachedFcmToken = function () {
        return this.fcmToken;
    };

    Thsq.prototype.removeDevice = function (unique, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (id) {
            fetch(_this, 'user', {
                url: _this.host + '/0/devices/' + id,
                method: 'DELETE',
                success: function (e) {
                    if (typeof callback === 'function') {
                        callback(e);
                    }
                },
                error: function () {
                    if (typeof callback === 'function') {
                        callback();
                    }
                }
            });
        });
    };

    Thsq.prototype.claimDevice = function (unique, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (key) {
            fetch(_this, 'user', {
                url: _this.host + '/0/devices/',
                method: 'POST',
                data: { auth: key },
                success: function (e) {
                    if (typeof callback === 'function') {
                        callback(e);
                    }
                },
                error: function (e) {
                    if (typeof callback === 'function') {
                        callback(e);
                    }
                }
            });
        });
    };

    Thsq.prototype.replaceDevice = function (oldunique, newunique, callback) {
        var _this = this;
        convertToIdFromUnique(_this, oldunique, function (key) {
            convertToIdFromUnique(_this, newunique, function (replace) {
                fetch(_this, 'user', {
                    url: _this.host + '/0/devices/' + key,
                    method: 'POST',
                    data: { replace: replace },
                    success: function (e) {
                        fetchUpdateDeviceCache(thsq, undefined, function () {
                            if (typeof callback === 'function') {
                                callback(e);
                            }
                        });
                    },
                    error: function (e) {
                        if (typeof callback === 'function') {
                            callback(e);
                        }
                    }
                });
            });
        });
    };

    Thsq.prototype.shareDevice = function (unique, username, callback) {
        var _this = this;

        /* Check if the device we are going to share is already shared with that user. If so, we don't need to go throgh the full REST call, but we just call the callback immediately. */
        if (_this.devicecache[unique]) {
            var device = _this.devicecache[unique].device;
            if (JSON.parse(thsq.variablevaluestring(device, 'meta', 'users', '[]')).indexOf(username) !== -1) {
                /* Device is already shared with the user */
                setTimeout(function () {
                    callback('add-user-ok');
                });
                return;
            }
        }


        convertToIdFromUnique(this, unique, function (id) {
            fetch(_this, 'user', {
                url: _this.host + '/0/devices/',
                method: 'POST',
                data: { auth: id, user: username },
                success: function (e) {
                    if (typeof callback === 'function') {
                        callback(e);
                    }
                },
                error: function (e) {
                    if (typeof callback === 'function') {
                        callback(e);
                    }
                }
            });
        });
    };

    Thsq.prototype.sendPing = function () {
        if (this.dontPing) {
            return;
        }
        if (this.connected && this.connection) {
            this.connection.send("ping");
        }
    };

    Thsq.prototype.session = function (callback) {
        var _this = this;

        fetch(_this, 'user', {
            url: _this.host + '/0/session/',
            method: 'POST',
            data: { token: _this.options.token },
            success: function (data, status) {
                _this.disconnect();
                updateUserCache(_this, function () {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                });
            },
            error: function (data) {
                _this.disconnect();
                updateUserCache(_this, function () {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                });
            }
        });
    };

    Thsq.prototype.logout = function (callback) {
        var _this = this;
        delete _this.options.token;
        _this.devicecache = {};
        _this.claimeduniqueids = {};
        _this.deviceuniqueids = {};
        _this.lastseen = undefined;
        if (_this.options.browsercache) {
            delete localStorage['devicecache'];
            delete localStorage['devicecache-lastseen'];
        }
        fetch(_this, 'user', {
            url: _this.host + '/0/session/logout',
            method: 'POST',
            success: function (data) {
                /* Create new anon session */
                _this.session(callback);
            },
            error: function () {
                /* Create new anon session */
                updateUserCache(_this, function () {
                    if (typeof callback === 'function') {
                        callback();
                    }
                });
            }
        });
    };

    Thsq.prototype.login = function (username, password, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/session/login',
            method: 'POST',
            data: {
                login: username,
                password: password,
            },
            success: function (data) {
                if (_this.options.browsercache) {
                    delete localStorage['devicecache'];
                    delete localStorage['devicecache-lastseen'];
                }
                _this.disconnect();
                _this.reconnect();
                updateUserCache(_this, function () {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                });
            },
            error: function (data) {
                if (_this.options.browsercache) {
                    delete localStorage['devicecache'];
                    delete localStorage['devicecache-lastseen'];
                }
                _this.disconnect();
                _this.reconnect();
                updateUserCache(_this, function () {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                });
            }
        });
    };

    Thsq.prototype.getFirmwarelist = function (callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/firmware/',
            success: function (e) {
                callback(JSON.parse(e));
            },
            error: callback
        });
    };

    Thsq.prototype.stopFirmwareUpdate = function (unique, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (id) {
            fetch(_this, 'user', {
                url: _this.host + '/0/devices/' + id + '/firmware/',
                method: 'DELETE',
                success: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                },
                error: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                }
            });
        });
    };

    Thsq.prototype.startFirmwareUpdate = function (unique, firmwarename, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (id) {
            fetch(_this, 'device', {
                url: _this.host + '/0/devices/' + id + '/firmware/' + firmwarename,
                method: 'POST',
                success: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                },
                error: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                }
            });
        });
    };

    Thsq.prototype.startFirmwareUpdateWithFile = function (unique, filedata, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (id) {
            fetch(_this, 'device', {
                url: _this.host + '/0/devices/' + id + '/fwupfile/',
                method: 'POST',
                formData: {
                    data: {
                        value: filedata,
                        options: {
                            filename: 'data',
                            contentType: 'application/binary'
                        }
                    }
                },
                success: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                },
                error: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                }
            });
        });
    };

    Thsq.prototype.startNetworkUpdate = function (unique, firmwarename, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (id) {
            fetch(_this, 'device', {
                url: _this.host + '/0/devices/' + id + '/zwup/' + firmwarename,
                method: 'POST',
                success: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                },
                error: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                }
            });
        });
    };

    Thsq.prototype.getDevicelist = function (callback) {
        var devicelist, _this;
        _this = this;
        function populatedevicelist() {
            var unique;
            devicelist = {};
            for (unique in _this.claimeduniqueids) {
                if (_this.devicecache[unique]) {
                    devicelist[unique] = clonedevice(_this.devicecache[unique].device);
                } else {
                    delete devicelist[unique];
                }
            }
            callback(devicelist);
        }
        if (Object.keys(this.claimeduniqueids).length === 0) {
            fetchUpdateDeviceCache(this, undefined, populatedevicelist);
        } else {
            populatedevicelist();
        }
    };

    Thsq.prototype.getDevicelistWithFilter = function (filterfunction, callback) {
        var devicelist, _this;
        _this = this;
        function populatedevicelist() {
            var unique;
            devicelist = {};
            for (unique in _this.claimeduniqueids) {
                if (_this.devicecache[unique]) {
                    if (filterfunction(_this.devicecache[unique].device)) {
                        devicelist[unique] = clonedevice(_this.devicecache[unique].device);
                    }
                } else {
                    delete devicelist[unique];
                }
            }
            callback(devicelist);
        }
        if (Object.keys(this.claimeduniqueids).length === 0) {
            fetchUpdateDeviceCache(this, undefined, populatedevicelist);
        } else {
            populatedevicelist();
        }
    };

    function callinitcallbacks(thsq) {

        var callbacks = thsq.initcallbacks;
        thsq.initcallbacks = [];
        var i;
        for (i = 0; i < callbacks.length; i += 1) {
            callbacks[i]();
        }
    }
    function convertToUniqueFromId(thsq, id, callback) {

        function getUniqueFromId(thsq, id) {
            var unique;

            if (thsq.uniquefromids[id]) {
                return thsq.uniquefromids[id];
            }

            for (unique in thsq.claimeduniqueids) {
                if (thsq.claimeduniqueids.hasOwnProperty(unique)) {
                    if (thsq.claimeduniqueids[unique] === id) {
                        thsq.uniquefromids[id] = unique;
                        return unique;
                    }
                }
            }
            for (unique in thsq.deviceuniqueids) {
                if (thsq.deviceuniqueids.hasOwnProperty(unique)) {
                    if (thsq.deviceuniqueids[unique] === id) {
                        thsq.uniquefromids[id] = unique;
                        return unique;
                    }
                }
            }
            return id;
        }
        if (thsq.inited) {
            callback(getUniqueFromId(thsq, id));
        } else {
            thsq.initcallbacks.push(function () {
                callback(getUniqueFromId(thsq, id));
            });
        }
    }

    function convertToIdFromUnique(thsq, unique, callback) {
        function getIdFromUnique(thsq, unique) {
            var id;
            id = thsq.claimeduniqueids[unique];
            if (!id) {
                id = thsq.deviceuniqueids[unique];
            }
            if (id) {
                return id;
            } else {
                return unique;
            }
        }

        if (thsq.inited) {
            callback(getIdFromUnique(thsq, unique));
        } else {
            thsq.initcallbacks.push(function () {
                callback(getIdFromUnique(thsq, unique));
            });
        }
    }


    function fetchingDeviceToCache(thsq, id) {
        if (id && thsq.devicesfetching[id] === true) {
            return true;
        }
        if (id === undefined && Object.keys(thsq.devicesfetching).length > 0) {
            return true;
        }
        return false;
    }

    function fetchDeviceToCache(thsq, id, callback) {
        var unique;
        if (fetchingDeviceToCache(thsq, id)) {
            callback();
            return;
        }
        thsq.devicesfetching[id] = true;

        var isempty = function (obj) {
            for (var e in obj) {
                return false;
            }
            return true;
        };

        fetch(thsq, 'device-cache', {
            url: thsq.host + '/0/devices/' + id,
            success: function (e) {
                var device;
                delete thsq.devicesfetching[id];
                if (e) {
                    try {
                        device = JSON.parse(e);
                    } catch (ignored) { }
                    if (deviceunique(device)) {
                        unique = deviceunique(device);
                        updateDeviceUnique(thsq, unique, id);
                        if (device.d === undefined) {
                            device.d = {};
                        }
                        if (device.s === undefined) {
                            device.s = {};
                        }
                        if (device.meta === undefined) {
                            device.meta = {};
                        }

                        if (thsq.claimeduniqueids[unique]) {
                            var diff = updateDeviceCache(thsq, unique, device);
                            if (!isempty(diff.d) || !isempty(diff.s) || !isempty(diff.meta)) {
                                thsq.emit('device-updated', device, unique, diff);
                            }
                        }

                        callback(device);
                    } else {
                        //log('error ' + err);
                        callback();
                    }
                } else {
                    callback();
                }
            },
            error: function (e) {
                delete thsq.devicesfetching[id];
                convertToUniqueFromId(thsq, id, function (unique) {
                    if (unique && thsq.devicecache[unique]) {
                        if (e === 'timeout') {
                            callback(thsq.devicecache[unique].device, e);
                        } else {
                            delete thsq.devicecache[unique];
                            callback(undefined, e);
                        }
                    } else {
                        callback(undefined, e);
                    }
                });
            }
        });
    }

    function updateDeviceCache(thsq, unique, device) {
        var diff, v, t;

        diff = { meta: {}, d: {}, s: {} };
        if (device) {
            if (thsq.devicecache[unique]) {
                var cached = thsq.devicecache[unique].device;
                for (t in diff) {
                    for (v in device[t]) {
                        if (cached[t] === undefined) {
                            cached[t] = {};
                        }
                        if (device[t][v]) {

                            /* Make sure to check if .seq and .ackseq is undefined, since '1 > undefined' is false, but in this case should be interpreted as `1 > 0`, which is true */
                            if (cached[t][v] === undefined ||
                                ((cached[t][v].seq === undefined && device[t][v].seq !== undefined) || device[t][v].seq > cached[t][v].seq) ||
                                ((cached[t][v].ackseq === undefined && device[t][v].ackseq !== undefined) || device[t][v].ackseq > cached[t][v].ackseq)) {
                                diff[t][v] = clonedata(device[t][v]);
                            }
                        }
                    }
                }
            } else {
                diff = clonedevice(device);
            }
            if (device.d === undefined) {
                device.d = {};
            }
            if (device.s === undefined) {
                device.s = {};
            }
            if (device.meta === undefined) {
                device.meta = {};
            }
            thsq.devicecache[unique] = { device: clonedevice(device) };
        } else {
            if (thsq.claimeduniqueids[unique] === undefined) {
                delete thsq.devicecache[unique];
            }
        }
        return diff;
    }

    function addDeltaToDeviceCache(thsq, unique, delta) {
        var device, newdevice, diff, t, v;

        diff = { meta: {}, d: {}, s: {} };
        if (thsq.devicecache[unique]) {
            device = thsq.devicecache[unique].device;
            newdevice = clonedevice(device);
            for (t in diff) {
                for (v in delta[t]) {
                    if (device[t] === undefined) {
                        device[t] = {};
                    }
                    if (newdevice[t] === undefined) {
                        newdevice[t] = {};
                    }

                    if (device[t][v] === undefined) {
                        diff[t][v] = clonedata(delta[t][v]);
                    } else {
                        /* Make sure to check if .seq and .ackseq is undefined, since '1 > undefined' is false, but in this case should be interpreted as `1 > 0`, which is true */
                        if (delta[t][v] === undefined ||
                            ((device[t][v].seq === undefined && delta[t][v].seq !== undefined) ||
                                delta[t][v].seq > device[t][v].seq) ||
                            ((device[t][v].ackseq === undefined && delta[t][v].ackseq !== undefined) ||
                                delta[t][v].ackseq > device[t][v].ackseq)) {
                            if (diff[t][v] === undefined) {
                                diff[t][v] = {};
                            }
                            diff[t][v] = clonedata(delta[t][v]);
                        }
                    }
                    newdevice[t][v] = clonedata(delta[t][v]);
                }
            }
        } else {
            newdevice = clonedevice(delta);
        }
        if (newdevice) {
            if (newdevice.d === undefined) {
                newdevice.d = {};
            }
            if (newdevice.s === undefined) {
                newdevice.s = {};
            }
        }

        thsq.devicecache[unique] = { device: newdevice };
        return diff;
    }

    var fetchDevicelistqueue = [];
    function fetchDevicelist(thsq, callback) {
        var callcallbacks = function (list) {
            var i;
            var callbacks = fetchDevicelistqueue;
            fetchDevicelistqueue = [];
            for (i = 0; i < callbacks.length; i += 1) {
                callbacks[i](list);
            }
        };

        fetchDevicelistqueue.push(callback);

        if (fetchDevicelistqueue.length > 1) {
            return;
        }
        fetch(thsq, 'device', {
            url: thsq.host + '/0/devices/',
            data: {
                devicedatafilter: thsq.options.devicedatafilter,
                lastseen: thsq.lastseen,  /* Load only the changes since we last saw the devices */
            },
            success: function (e) {
                var list;
                try {
                    list = JSON.parse(e);
                } catch (ignored) {
                }
                if (list) {
                    callcallbacks(list);
                } else {
                    callcallbacks();
                }
            },
            error: function () {
                callcallbacks();
            }
        });
    }

    function updateDeviceUnique(thsq, unique, id) {
        thsq.deviceuniqueids[unique] = id;
    }
    function fetchUpdateDeviceCache(thsq, unique, callback) {
        var device;
        if (unique === undefined) {
            fetchDevicelist(thsq, function (devices) {
                var d;
                for (d in devices) {
                    if (devices.hasOwnProperty(d)) {
                        if (devices[d] === null) {
                            /* Device was unclaimed */
                            thsq.emit('device-removed', d);
                        } else {
                            unique = deviceunique(devices[d]);
                            thsq.uniquefromids[d] = unique;
                            devices[d].id = d;
                            updateDeviceCache(thsq, unique, devices[d]);
                            thsq.claimeduniqueids[unique] = d;
                        }
                    }
                }
                callback();
            });
            return;
        }
        convertToIdFromUnique(thsq, unique, function (id) {
            fetch(thsq, 'device-cache', {
                url: thsq.host + '/0/devices/' + id,
                success: function (e) {
                    if (e) {
                        try {
                            device = JSON.parse(e);
                        } catch (ignored) { }
                        if (device && e !== '{}') {
                            device.id = id;
                            unique = deviceunique(device);
                            updateDeviceUnique(thsq, unique, id);
                            updateDeviceCache(thsq, unique, device);
                            callback(device, unique);
                        } else {
                            updateDeviceCache(thsq, unique);
                            callback();
                        }
                    } else {
                        updateDeviceCache(thsq, unique);
                        callback();
                    }
                },
                error: function () {
                    updateDeviceCache(thsq, unique);
                    callback();
                }
            });
        });
    }

    Thsq.prototype.getDevice = function (id, callback) {
        var _this = this;
        convertToUniqueFromId(this, id, function (unique) {
            if (unique === undefined || unique === '') {
                callback();
            } else if (_this.claimeduniqueids[unique] && _this.devicecache[unique] && _this.devicecache[unique].device) {
                callback(_this.devicecache[unique].device, unique);
            } else {
                fetchUpdateDeviceCache(_this, unique, callback);
            }
        });
    };

    function getDeviceDataHistory(thsq, unique, type, name, num, from, startid, callback) {
        convertToIdFromUnique(thsq, unique, function (id) {
            fetch(thsq, 'device', {
                url: thsq.host + '/0/devices/' + id + '/' + type + '/' + name,
                data: {
                    latest: num,
                    from: from,
                    startid: startid,
                    received: true
                },
                success: function (e) {
                    var list, i;
                    if (e) {
                        list = JSON.parse(e);
                        // For backwards compatibility
                        for (i = 0; i < list.length; i += 1) {
                            if (list[i].timestamp && !list[i].time) {
                                list[i].time = list[i].timestamp;
                            }
                        }
                        callback(list);
                    } else {
                        callback([]);
                    }
                },
                error: function () {
                    callback();
                }
            });
        });
    }

    function getDeviceDataHistoryForward(thsq, unique, type, name, num, startid, callback) {
        /* Fetch history starting at given id, going forward. Sort by id.
         * TODO Add support for num > chunksize */
        getDeviceDataHistory(thsq, unique, type, name, num, undefined, startid, function (data) {
            if (data !== undefined && Array.isArray(data)) {
                data = data.sort(function (a, b) {
                    return b.id - a.id;
                });
                callback(data);
            } else {
                callback();
            }
        });
    }

    function getDeviceDataHistoryBackwards(thsq, unique, type, name, chunksize, until, callback) {
        /* fetches more and more data while calling until() repeatedly.
         * when no more data is needed, until() returns true. */
        var data = [];
        if (typeof until !== 'function') {
            callback();
            return;
        }

        var getmore = function (id) {
            if (until(data)) {
                data = data.sort(function (a, b) {
                    return b.id - a.id;
                });
                callback(data);
                return;
            }

            getDeviceDataHistory(thsq, unique, type, name, chunksize, id, undefined, function (datachunk) {
                if (datachunk === undefined || !Array.isArray(datachunk)) {
                    callback();
                    return;
                }

                data = data.concat(datachunk);
                data = data.sort(function (a, b) {
                    return b.id - a.id;
                });
                if (datachunk.length === 0) {
                    /* No more data to fetch */
                    callback(data);
                } else {
                    getmore(datachunk[datachunk.length - 1].id);
                }
            });
        };
        getmore();
    }

    Thsq.prototype.getVariableHistory = function (id, type, name, options, callback) {
        var _this = this;
        convertToUniqueFromId(this, id, function (unique) {

            /* fetch from most recent value going backwards, options: num, chunksize */
            /* fetch from given id going forward, options: startid, num */
            /* fetch elements in given time interval, options: starttime, endtime, chunksize */

            if (typeof options === 'function') {
                callback = options;
                options = {};
            }
            if (!options) {
                options = {};
            }
            if (options.num === undefined) {
                options.num = 1000;
            }
            if (!options.chunksize) {
                options.chunksize = 1000;
            }

            if (options.startid !== undefined) {
                /* Fetch history elements starting at given options.startid and going forward in time. */
                getDeviceDataHistoryForward(thsq, unique, type, name, options.num, options.startid, callback);
                return;
            }

            if (options.starttime !== undefined || options.endtime !== undefined) {
                /* Fetch history elements in given time interval */
                if (options.endtime === undefined) {
                    options.endtime = Date.now();
                }
                if (options.endtime < options.starttime) {
                    /* Error: sanity-check time interval options */
                    callback();
                    return;
                }
                var gotenoughdates = function(data) {
                    if (data.length > 0) {
                        if (typeof options.progress === 'function') {
                            options.progress(data.length, unique, data[data.length - 1]);
                        }
                        return (data[0].time < options.starttime ||
                            data[data.length - 1].time < options.starttime);
                    }
                };
                getDeviceDataHistoryBackwards(thsq, unique, type, name, options.chunksize, gotenoughdates, function(data) {
                    if (!data || !Array.isArray(data)) {
                        callback();
                        return;
                    }
                    data = data.sort(function (a, b) {
                        return b.id - a.id;
                    });
                    data = data.filter(function (d) {
                        return (d.time >= options.starttime && d.time <= options.endtime);
                    });
                    callback(data);
                });
            } else {
                /* Fetch history starting at given time, going backwards. Sort by id. */
                var gotenoughelements = function(data) {
                    if (typeof options.progress === 'function') {
                        options.progress(data.length, unique, data[data.length - 1]);
                    }
                    return data.length >= options.num;
                };
                if (options.num && options.chunksize && options.num < options.chunksize) {
                    options.chunksize = options.num;
                }
                getDeviceDataHistoryBackwards(thsq, unique, type, name, options.chunksize, gotenoughelements, function(data) {
                    if (!data || !Array.isArray(data)) {
                        callback();
                        return;
                    }
                    data = data.sort(function (a, b) {
                        return b.id - a.id;
                    });
                    if (data.length > options.num) {
                        data.splice(options.num);
                    }
                    callback(data);
                });
            }
        });
    };

    Thsq.prototype.deleteDeviceHistory = function (unique, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (deviceid) {
            fetch(_this, 'device', {
                url: _this.host + '/0/devices/' + deviceid + '/fuglyhistory',
                method: 'DELETE',
                success: function (e) {
                    if (typeof callback === 'function') {
                        callback(e);
                    }
                },
                error: function () {
                    if (typeof callback === 'function') {
                        callback();
                    }
                }
            });
        });
    };

    Thsq.prototype.deleteDeviceCommands = function (unique, command, callback) {
        var _this = this;
        if (command === '') {
            /* We need to check for an empty command, otherwise we will unclaim the device with the REST call below */
            callback();
            return
        }
        convertToIdFromUnique(this, unique, function (deviceid) {
            fetch(_this, 'device', {
                url: _this.host + '/0/devices/' + deviceid + '/' + command,
                method: 'DELETE',
                success: function (e) {
                    if (typeof callback === 'function') {
                        callback(e);
                    }
                },
                error: function () {
                    if (typeof callback === 'function') {
                        callback();
                    }
                }
            });
        });
    };

    Thsq.prototype.getNotifications = function (callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/user/event/',
            success: function (e) {
                if (e) {
                    callback(JSON.parse(e));
                } else {
                    callback();
                }
            },
            error: function () {
                callback();
            }
        });
    };

    Thsq.prototype.addNotification = function (unique, name, message, trigger, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (deviceid) {
            fetch(_this, 'user', {
                url: _this.host + '/0/user/event/',
                method: 'POST',
                data: {
                    name: name,
                    message: message,
                    trigger: trigger,
                    deviceid: deviceid
                },
                success: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                },
                error: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                }
            });
        });
    };

    Thsq.prototype.updateNotification = function (id, unique, name, message, trigger, callback) {
        var _this = this;
        convertToIdFromUnique(this, unique, function (id) {
            fetch(_this, 'user', {
                url: _this.host + '/0/user/event/' + id,
                method: 'POST',
                data: {
                    name: name,
                    message: message,
                    trigger: trigger,
                    deviceid: id
                },
                success: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                },
                error: function (data) {
                    if (typeof callback === 'function') {
                        callback(data);
                    }
                }
            });
        });
    };

    Thsq.prototype.getNotification = function (id, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/user/event/' + id,
            method: 'GET',
            success: function (data) {
                callback(JSON.parse(data));
            },
            error: function (data) {
                callback(data);
            }
        });
    };


    Thsq.prototype.testNotification = function (id, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/user/event/' + id,
            method: 'POST',
            success: function (data) {
                if (typeof callback === 'function') {
                    callback(data);
                }
            },
            error: function (data) {
                if (typeof callback === 'function') {
                    callback(data);
                }
            }
        });
    };

    Thsq.prototype.removeNotification = function (id, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/user/event/' + id,
            method: 'DELETE',
            success: function (data) {
                if (typeof callback === 'function') {
                    callback(data);
                }
            },
            error: function (data) {
                if (typeof callback === 'function') {
                    callback(data);
                }
            }
        });
    };

    Thsq.prototype.getEvents = function (num, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/user/events/',
            data: {
                latest: num,
            },
            success: function (e) {
                if (e) {
                    callback(JSON.parse(e));
                } else {
                    callback();
                }
            },
            error: function () {
                callback();
            }
        });
    };


    Thsq.prototype.userSignup = function (username, password, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/session/signup',
            method: 'POST',
            data: {
                login: username,
                password: password,
            },
            success: function (data) {
                _this.disconnect();
                _this.reconnect();
                if (typeof callback === 'function') {
                    callback(data);
                }
            },
            error: function () {
                _this.disconnect();
                _this.reconnect();
                if (typeof callback === 'function') {
                    callback();
                }
            }
        });
    };

    Thsq.prototype.userResendConfirmationEmail = function (callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/session/resend',
            method: 'POST',
            success: function (e) {
                if (typeof callback === 'function') {
                    callback(e);
                }
            },
            error: function () {
                if (typeof callback === 'function') {
                    callback();
                }
            }
        });
    };

    Thsq.prototype.userSendPasswordRecoveryEmail = function (username, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/session/recover',
            method: 'POST',
            data: { login: username },
            success: function (e) {
                if (typeof callback === 'function') {
                    callback(e);
                }
            },
            error: function () {
                if (typeof callback === 'function') {
                    callback();
                }
            }
        });
    };

    Thsq.prototype.createAccessToken = function (callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/user/accesstoken',
            method: 'POST',
            success: function (e) {
                if (typeof callback === 'function') {
                    callback(e);
                }
            },
            error: function () {
                if (typeof callback === 'function') {
                    callback();
                }
            }
        });
    };

    Thsq.prototype.deleteAccessToken = function (token, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/user/accesstoken/' + token,
            method: 'DELETE',
            success: function (e) {
                if (typeof callback === 'function') {
                    callback(e);
                }
            },
            error: function () {
                if (typeof callback === 'function') {
                    callback();
                }
            }
        });
    };

    Thsq.prototype.recoverNewPassword = function (username, token, password, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/session/recover',
            method: 'POST',
            data: { login: username, token: token, password: password },
            success: function (e) {
                if (typeof callback === 'function') {
                    callback(e);
                }
            },
            error: function () {
                if (typeof callback === 'function') {
                    callback();
                }
            }
        });
    };

    Thsq.prototype.getVariable = function (id, type, variable, callback) {
        var _this = this;
        convertToUniqueFromId(this, id, function (unique) {
            if (type !== 's' && type !== 'd' && type !== 'meta') {
                if (typeof callback === 'function') {
                    callback();
                }
                return;
            }
            if (_this.devicecache[unique] === undefined || _this.devicecache[unique].device === undefined) {
                if (typeof callback === 'function') {
                    callback();
                }
                return;
            }
            if (_this.devicecache[unique].device[type] === undefined) {
                if (typeof callback === 'function') {
                    callback();
                }
                return;
            }
            if (typeof callback === 'function') {
                callback(_this.devicecache[unique].device[type][variable]);
            }
            return _this.devicecache[unique].device[type][variable];
        });
    };

    Thsq.prototype.getVariableValue = function (id, type, variable, fallback, callback) {
        this.getVariable(id, type, variable, function(value) {
            if (value) {
                if (value.value !== undefined) {
                    if (typeof Buffer !== 'undefined' && value.value.data && value.value.type === 'Buffer') { // node.js
                        callback(Buffer.from(value.value.data), value);
                    } else {
                        callback(value.value, value);
                    }
                } else {
                    callback(fallback, value);
                }
            } else {
                callback(fallback);
            }
        });
    };


    Thsq.prototype.getVariableStringValue = function (id, type, variable, fallback, callback) {
        function tostring(val) {
            if (typeof val === 'string') {
                return val;
            }
            if (typeof Buffer !== 'undefined') { // node.js
                return Buffer.from(val).toString();
            }
            if (val.data) { // browser
                return String.fromCharCode.apply(null, val.data);
            }
            return '';
        }
        thsq.getVariableValue(id, type, variable, fallback, function (value, variable) {
            callback(tostring(value), variable);
        });
    };


    Thsq.prototype.getVariableBufferValue = function (id, type, variable, fallback, callback) {
        thsq.getVariableValue(id, type, variable, fallback, function (value, variable) {
            callback(Buffer.from(value), variable);
        });
    };


    Thsq.prototype.getVariableNumberValue = function (id, type, variable, fallback, callback) {
        thsq.getVariableValue(id, type, variable, fallback, function (value, variable) {
            if (Number.isNaN(Number(value))) {
                callback(Number(fallback), variable);
            } else {
                callback(Number(value), variable);
            }
        });
    };

    Thsq.prototype.deleteVariable = function (unique, type, variable, callback) {
        var _this;

        if (type !== 's' && type !== 'd') {
            callback();
            return;
        }

        _this = this;

        // Update cache directly
        var diff;
        diff = {};
        diff[type] = {};
        diff[type][variable] = { value: undefined, time: Date.now() };
        var device = getDeviceFromCache(_this, unique);
        if (device) {
            if (device[type][variable]) {
                if (device[type][variable].value === undefined) {
                    /* We have previously deleted this value, so we should delete the entire variable info now */
                    diff[type][variable] = undefined;
                } else {
                    /* Value is not deleted, so we should only delete the value */
                }
            }
        }
        addDeltaToDeviceCache(_this, unique, diff);


        convertToIdFromUnique(this, unique, function (id) {
            fetch(_this, 'device', {
                url: _this.host + '/0/devices/' + id + '/' + type + '/' + variable,
                method: 'DELETE',
                success: function (e) {
                    if (e === 'delete-ok') {
                        if (typeof callback === 'function') {
                            callback(e);
                        }
                    } else {
                        if (typeof callback === 'function') {
                            callback(e);
                        }
                    }
                },
                error: function (e) {
                    if (typeof callback === 'function') {
                        callback(e);
                    }
                }
            });
        });
    };

    function isAscii(str) {
        return /^[\x20-\x7f]*$/.test(str);
    }

    Thsq.prototype.getDeviceRssi = function (unique) {
        if (unique !== undefined) {
            return this.nearbydevicesrssi[unique];
        }
    };

    Thsq.prototype.setVariable = function (unique, type, variable, val, options, callback) {
        var _this, value, value64, postdata;

        if (options === undefined) {
            options = {};
        }
        if (typeof options === 'function') {
            callback = options;
            options = {};
        }

        _this = this;

        if (type !== 's' && type !== 'd') {
            if (typeof callback === 'function') {
                callback();
            }
            return;
        }

        convertToIdFromUnique(this, unique, function (id) {
            value = val;

            postdata = { value: value };

            var time = Date.now();
            if (options.timestamp !== undefined) {
                time = postdata.timestamp = options.timestamp;
            }

            // Update cache directly
            if (value !== undefined) {
                var diff;
                diff = {};
                diff[type] = {};
                diff[type][variable] = { value: value, time: time };
                addDeltaToDeviceCache(_this, unique, diff);
            }

            fetch(_this, 'device', {
                url: _this.host + '/0/devices/' + id + '/' + type +'/' + variable,
                method: 'POST',
                data: postdata,
                success: function (e) {
                    if (e === 'device-ok') {
                        if (typeof callback === 'function') {
                            callback(e);
                        }
                    } else {
                        if (typeof callback === 'function') {
                            callback(e);
                        }
                    }
                },
                error: function () {
                    if (typeof callback === 'function') {
                        callback();
                    }
                }
            });
        });
    };

    Thsq.prototype.clearHistoryListener = function (token) {
        var _this = this;

        Object.keys(this.historylisteners).map(function (key) {
            var arr = _this.historylisteners[key];
            if (_this.historylisteners[key]) {
                _this.historylisteners[key] = _this.historylisteners[key].filter(function (l) {
                     return l.token !== token;
                });
            }
        });
    };

    Thsq.prototype.clearAllHistoryListeners = function () {
        this.historylisteners = {};
    };

    Thsq.prototype.addHistoryListener = function (unique, type, name, options, listener) {
        var _this = this;

        var l = {
            unique: unique,
            type: type,
            name: name,
            options: options,
            listener: listener,
            token: Math.random().toString(36),
            key: unique + '-' + type + '-' + name,
            count: 0,
        };
        if (!l.options) {
            l.options = {};
        }
        if (l.options.num === undefined) {
            l.options.num = 1000;
        }

        if (!_this.historylisteners[l.key]) {
            _this.historylisteners[l.key] = [];
        }
        _this.historylisteners[l.key].push(l);

        thsq.getDevice(unique, function (device) {
            if (device && device[type] && device[type][name]) {
                _this.historylisteners[l.key].seq = device[type][name].seq;
            }
        });

        if (_this.exiting) {
            /* We are currently exiting, so we do not fetch the history */
        } else if (options.num > 0) {
            fetchpartialhistory(thsq, l);
        }

        return l.token;
    };


    Thsq.prototype.sendCommand = function (unique, command, options, callback) {
        var _this, data;
        var commandoptions = {};
        _this = this;

        if (typeof options === 'function') {
            callback = options;
            options = {};
        }

        if (options === undefined) {
            options = {};
        }

        if (typeof Buffer === 'function' && Buffer.isBuffer(command)) {
            /* This if a Buffer. If so, we should send its data as an array to the API. */
            command = command.toJSON().data;
        }

        var sendcommand = function (callback) {
            convertToIdFromUnique(_this, unique, function (id) {

                data = { command: command };

                Object.keys(commandoptions).map(function (v, i) {
                    data[v] = commandoptions[v];
                });

                fetch(_this, 'device', {
                    url: _this.host + '/0/devices/' + id + '/',
                    data: data,
                    method: 'POST',
                    success: function (e) {
                        if (typeof callback === 'function') {
                            callback(e);
                        }
                    },
                    error: function () {
                        if (typeof callback === 'function') {
                            callback();
                        }
                    }
                });
            });
        };

        var commandid;

        if (options.sendcallback === undefined) {
            options.sendcallback = function () {
            };
        }
        if (options.ackedcallback === undefined) {
            options.ackedcallback = function () {
            };
        }
        if (options.sendtimeoutcallback === undefined &&
            options.ackedtimeoutcallback === undefined) {
            options.sendtimeout = 0;
            options.ackedtimeout = 0;
        }

        if (options.sendtimeoutcallback === undefined) {
            options.sendtimeoutcallback = function () {
            };
        }
        if (options.ackedtimeoutcallback === undefined) {
            options.ackedtimeoutcallback = function () {
            };
        }
        if (options.ackedtimeout === undefined) {
            options.ackedtimeout = 10000;
        }
        if (options.sendtimeout === undefined) {
            options.sendtimeout = 60000;
        }

        commandoptions.priority = options.priority;
        commandoptions.proptime = options.proptime;
        commandoptions.lifetime = options.lifetime;
        commandoptions.broadcast = options.broadcast;

        var cleanup = function () {
            thsq.off('device-updated', deviceupdated);
            clearTimeout(ackedtimeoutid);
            clearTimeout(sendtimeoutid);
        };

        var ackedtimeoutid;
        var sendtimeoutid;

        sendtimeoutid = setTimeout(function () {
            options.sendtimeoutcallback();
            options.ackedtimeoutcallback();
            cleanup();
        }, options.sendtimeout);


        var wassentcalled = false;
        var wasackedcalled = false;

        var wassent = function () {
            if (wassentcalled === false) {
                wassentcalled = true;
                options.sendcallback();
                clearTimeout(sendtimeoutid);
                ackedtimeoutid = setTimeout(function () {
                    options.ackedtimeoutcallback();
                    cleanup();
                }, options.ackedtimeout);
            }
        };

        var wasacked = function () {
            if (wasackedcalled === false) {
                wasackedcalled = true;
                options.ackedcallback();
                cleanup();
            }
        };

        var deviceupdated = function (device, u, update) {
            if (u === unique) {

                var commands, i;

                commands = JSON.parse(thsq.variablevaluestring(device, 'meta', 'command-sent-recent', '[]'));
                for (i = 0; i < commands.length; i += 1) {
                    if (commands[i].cid.indexOf(commandid) !== -1) {
                        wassent();
                    }
                }
                commands = JSON.parse(thsq.variablevaluestring(device, 'meta', 'command-acked-recent', '[]'));
                for (i = 0; i < commands.length; i += 1) {
                    if (commands[i].cid.indexOf(commandid) !== -1) {
                        wassent();
                        wasacked();
                    }
                }
            }
        };
        sendcommand(function (id) {
            commandid = id;
            thsq.on('device-updated', deviceupdated);
            if (typeof callback === 'function') {
                if (id) {
                    callback('device-ok');
                } else {
                    callback();
                }
            }
        });
        return { cancel: cleanup };
    };
    Thsq.prototype.storeUserData = function (data, callback) {
        var _this = this;
        fetch(_this, 'user', {
            url: _this.host + '/0/user/',
            method: 'POST',
            data: data,
            success: function () {
                if (typeof callback === 'function') {
                    callback();
                }
            },
            error: function () {
                if (typeof callback === 'function') {
                    callback();
                }
            }
        });
    };


    Thsq.prototype.sendBLE = function (data) {
        var message;

        if (Array.isArray(data) || data instanceof Uint8Array) {
            const tohex = bytes => bytes.reduce((str, byte) => str + byte.toString(16).padStart(2, '0'), '');
            data = tohex(data);
        }


        /* iOS */
        if (typeof window !== 'undefined' &&
                window.webkit &&
                window.webkit.messageHandlers &&
                window.webkit.messageHandlers.javascriptBridge &&
                window.webkit.messageHandlers.javascriptBridge.postMessage &&
                typeof window.webkit.messageHandlers.javascriptBridge.postMessage === 'function') {
                message = { BLEData: data };
            window.webkit.messageHandlers.javascriptBridge.postMessage(message);
            return true;
        } else if (typeof AndroidBLESender !== 'undefined') {
            var devinfo = [ 0x02, 0x01, 0x06, 0x02, 0x09, 0x54 ]; /* devinfo + "T" */
            var manufacturerinfo = 'FF' + data;
            manufacturerinfo = tohex([manufacturerinfo.length / 2]) + manufacturerinfo;
            var pkt = (tohex(devinfo) + manufacturerinfo).toLowerCase();
            AndroidBLESender.send(pkt);
            return true;
        } else {

            /* If we have someone listening for the 'ble-beacon' event, we use that to send our beacons via */
            if (thsq.listeners('ble-beacon') && thsq.listeners('ble-beacon').length > 0) {
                thsq.emit('ble-beacon', data);
                return true;
            }

            if (typeof location !== 'undefined' && location.protocol === 'file:') {
                /* Emit a local event with the data as hex, for testing and development */
                setTimeout(function () {
                    thsq.emit('ble-received', data);
                }, 500);
                return true;
            }
        }
        return false;
    };


    Thsq.prototype.scanBLE = function (milliseconds) {
        var message, host, port;

        /* Android */
        if (typeof AndroidBLEListener !== 'undefined') {
            AndroidBLEListener.scanLeDevice(milliseconds);
        }

        /* iOS */
        if (typeof window !== 'undefined' &&
                window.webkit &&
                window.webkit.messageHandlers &&
                window.webkit.messageHandlers.javascriptBridge &&
                window.webkit.messageHandlers.javascriptBridge.postMessage &&
                typeof window.webkit.messageHandlers.javascriptBridge.postMessage === 'function') {
                message = { 'BLEProximityUUID': '5d813fd4-fe87-450b-9473-da262973d4c4', 'BLEIdentifier': 'THSQ', 'BLEListenPeriod': milliseconds};
            window.webkit.messageHandlers.javascriptBridge.postMessage(message);
        }
    };

    if (typeof window !== 'undefined') {
        var appWindow, appOrigin;

        window.addEventListener('message', function (e) {
            appWindow = e.source;
            appOrigin = e.origin;
        });
    }

    function sendudp(ipaddr, port, msg) {
        var message;
        /* iOS */
        if (typeof window !== 'undefined' &&
                window.webkit &&
                window.webkit.messageHandlers &&
                window.webkit.messageHandlers.javascriptBridge &&
                window.webkit.messageHandlers.javascriptBridge.postMessage &&
                typeof window.webkit.messageHandlers.javascriptBridge.postMessage === 'function') {
                message = { 'udpHost': ipaddr, 'udpPort': port, 'udpMsg': msg };
            window.webkit.messageHandlers.javascriptBridge.postMessage(message);
        }
        if (typeof AndroidUDPSender !== 'undefined') {
            AndroidUDPSender.send(ipaddr, port, msg);
        }
        if (appWindow && appOrigin) {
            message = {
                'udpHost' : ipaddr,
                'udpPort' : port,
                'data' : msg
            };
            appWindow.postMessage(JSON.stringify(message), appOrigin);
        }
    }
    var sendudpinterval;

    function scanForNetworkDevice (addr, authcode) {
        sendudp(addr, 590, authcode);
    }

    Thsq.prototype.scanForNetworkDevice = function (addr, authcode) {
        scanForNetworkDevice(addr, authcode);
    };

    var getuserqueue = [];
    Thsq.prototype.getUser = function (callback) {
        var _this = this;
        getuserqueue.push(callback);
        if (getuserqueue.length > 1) {
            return;
        }
        fetch(_this, 'user', {
            url: _this.host + '/0/user/',
            success: function (jsondata, status, xhr) {
                var data;
                if (!jsondata || jsondata === "undefined") {
                    jsondata = "{}";
                }

                data = JSON.parse(jsondata);
                if (data.authcode) {
                    scanForNetworkDevice('224.0.0.1', data.authcode);
                    scanForNetworkDevice('broadcast', data.authcode);
                    if (sendudpinterval) {
                        clearInterval(sendudpinterval);
                    }
                    if (_this.sendudpinterval) {
                        clearInterval(_this.sendudpinterval);
                    }
                    _this.sendudpinterval = setInterval(function () {
                        scanForNetworkDevice('224.0.0.1', data.authcode);
                        scanForNetworkDevice('broadcast', data.authcode);
                    }, 5000);
                }
                var i;
                var callbacks = getuserqueue;
                getuserqueue = [];
                for (i = 0; i < callbacks.length; i += 1) {
                    callbacks[i](data);
                }
            },
            error: function (xhr, status) {
                var callbacks = getuserqueue;
                var i;
                getuserqueue = [];
                for (i = 0; i < callbacks.length; i += 1) {
                    callbacks[i]();
                }
            }
        });
    };

    Thsq.prototype.getNetworkDevices = function (callback) {
        var _this = this;
        fetch(_this, 'device', {
            url: _this.host + '/0/network/',
            success: function (jsondata) {
                var d, unique;
                var devices = JSON.parse(jsondata);
                for (d in devices) {
                    if (deviceunique(devices[d])) {
                        unique = deviceunique(devices[d]);
                        updateDeviceUnique(_this, unique, d);
                    }
                    if (devices[d].d === undefined) {
                        devices[d].d = {};
                    }
                    if (devices[d].s === undefined) {
                        devices[d].s = {};
                    }
                }
                callback(devices);
            },
            error: function () {
                callback();
            }
        });
    };

    function deviceidstring(devicedata, str) {
        var r;
        if (devicedata && devicedata.s && devicedata.s.id) {
            r = devicedata.s.id.value.split(',').filter(function (e) { return e.split(':')[0] === str; });
            if (r.length > 0) {
                return r[0].split(':')[1];
            }
        }
    }

    function deviceidvalue(devicedata, val) {
        var eui;
        eui = deviceidstring(devicedata, val);
        if (!eui) {
            if (devicedata && devicedata.s && devicedata.s[val]) {
                eui = devicedata.s[val].value;
            }
        }
        return eui;
    }

    Thsq.prototype.deviceEUI = function (devicedata) {
        var eui;
        eui = deviceidstring(devicedata, 'eui');
        if (!eui) {
            if (devicedata.s.eui) {
                eui = devicedata.s.eui.value;
            }
        }
        return eui;
    };

    Thsq.prototype.devicePlatform = function (devicedata) {
        var platform;
        platform = deviceidstring(devicedata, 'platform');
        if (!platform) {
            if (devicedata && devicedata.s && devicedata.s.platform) {
                platform = devicedata.s.platform.value;
            }
        }
        return platform;
    };

    Thsq.prototype.deviceId = function (unique, callback) {
        convertToIdFromUnique(this, unique, callback);
    };

    Thsq.prototype.deviceposition = function (device) {
        var pos = thsq.variablevaluejson(device, 's', 'position');
        if (pos) {
            if (typeof pos.lat === 'number' && !Number.isNaN(pos.lat) &&
                typeof pos.lng === 'number' && !Number.isNaN(pos.lng)) {
                return pos;
            }
            /* We don't have a valid position, so we return undefined */
            return;
        }

        if (thsq.variablevaluenum(device, 's', 'lat') && thsq.variablevaluenum(device, 's', 'lng')) {
            return { lat: thsq.variablevaluenum(device, 's', 'lat'), lng: thsq.variablevaluenum(device, 's', 'lng') };
        }
    };

    Thsq.prototype.devicepositionlocked = function (device) {
        var pos = thsq.deviceposition(device);
        if (pos && pos.locked === true) {
            return true;
        }
        return false;
    };

    Thsq.prototype.setdeviceposition = function (device, position, callback) {
        var pos = thsq.deviceposition(device);

        if (pos === undefined) {
            pos = {};
        }
        if (pos.locked === true) {
            if (typeof callback === 'function') {
                callback();
            }
            return false;
        }
        pos.lat = position.lat;
        pos.lng = position.lng;
        thsq.setVariable(thsq.deviceunique(device), 's', 'position', JSON.stringify(pos), callback);
        return true;
    };

    Thsq.prototype.lockdeviceposition = function (device, callback) {
        var pos = thsq.deviceposition(device);
        if (pos) {
            pos.locked = true;
            thsq.setVariable(thsq.deviceunique(device), 's', 'position', JSON.stringify(pos), callback);
        } else {
            if (typeof callback === 'function') {
                callback();
            }
        }
    };


    Thsq.prototype.unlockdeviceposition = function (device, callback) {
        var pos = thsq.deviceposition(device);
        pos.locked = false;
        thsq.setVariable(thsq.deviceunique(device), 's', 'position', JSON.stringify(pos));
        if (typeof callback === 'function') {
            callback();
        }
    };

    Thsq.prototype.parseBinarySchedule = function (binaryschedule) {
        var events = [];
        var event;
        var i, d;
        var days, hour, minute, dimlevels;

        if (binaryschedule) {
            for (i = 0; i < binaryschedule.length; i += 7) {
                if (binaryschedule[i + 0] === 119) {
                    days = [];
                    for (d = 0; d < 7; d += 1) {
                        if ((binaryschedule[i + 1] & (1 << d)) !== 0) {
                            days.push(d);
                        }
                    }
                    event = {
                        days: days,
                        time: { hour: binaryschedule[i + 2],
                            minute: binaryschedule[i + 3] },
                            intensity: [ binaryschedule[i + 4],
                            binaryschedule[i + 5],
                            binaryschedule[i + 6] ]
                    };
                }
                events.push(event);
            }
            events = events.sort(function (a, b) {
                return a.time.hour * 60 + a.time.minute - (b.time.hour * 60 + b.time.minute);
            });
        }
        return events;
    };

    Thsq.prototype.compileBinarySchedule = function (events) {
        var i, j, compiledlist, compiled, binary;


        function compile(event) {
            var schedule, k, daybits;

            daybits = 0;

            for (k = 0; k < event.days.length; k += 1) {
                daybits |= (1 << event.days[k]);
            }

            schedule = new Uint8Array(7);
            schedule[0] = 119;
            schedule[1] = daybits;
            schedule[2] = event.time.hour;
            schedule[3] = event.time.minute;
            schedule[4] = event.intensity[0];
            schedule[5] = event.intensity[1];
            schedule[6] = event.intensity[2];

            return schedule;

        }
        if (events.length > 8) {
            throw new Error('Too long schedule (must be maximum 8)');
        }

        compiledlist = [];
        for (i = 0; i < events.length; i += 1) {
            compiledlist.push(compile(events[i]));
        }

        compiled = new Uint8Array(7 * compiledlist.length);
        for (i = 0; i < compiledlist.length; i += 1) {
            for (j = 0; j < 7; j += 1) {
                compiled[i * 7 + j] = compiledlist[i][j];
            }
        }
        return compiled;
    };

    function fetchpartialhistory(thsq, l) {
        var _this = thsq;

        if (_this.exiting) {
            /* We are currently exiting, so we do not fetch the history */
            return;
        }
        if (l.fetching) {
            /* Already fetching history, ignore request */
            return;
        }
        l.fetching = true;

        function maybefetchagain(reason) {
            /* Check if there's more data to be fetched, when compared to
             * the latest timestamp provided by the device data. This can be
             * motivated by a very large amount of data added to the history
             * recently that has to be chunked, or that the history was not
             * ready to be fetched previously. */

            if (_this.exiting) {
                /* We are currently exiting, so we do not fetch the history */
                return;
            }

            if (reason === 'fetchfailed') {
                /* Previous fetch request failed. Perhaps we cannot reach the server.
                 * Instead of trying again, we give up and await a device list or device
                 * event to trigger a new fetch of the remaining history. */
                return;
            }

            _this.getDevice(l.unique, function(device) {
                if (device && device[l.type] && device[l.type][l.name]) {
                    if (device[l.type][l.name].time > l.timestamp) {
                        fetchpartialhistory(_this, l);
                    }
                }
            });
        }

        if (l.options.startid !== undefined) {
            /* startid has highest priority */
        } else if (l.options.starttime !== undefined) {
            /* if no startid, we will fetch data to given start time */
        } else if (l.options.num !== undefined) {
            /* fallback to number of elements */
        }
        var options = l.options;
        if (l.options.num === undefined || l.options.num === 0) {
            options.num = 1000;
        }

        _this.getVariableHistory(l.unique, l.type, l.name, options, function(data) {
            var d, device;

            l.fetching = false;

            if (typeof options.fetched === 'function') {
                var num;
                if (data) {
                    num = data.length;
                }
                options.fetched(num);
            }

            if (!data || !Array.isArray(data)) {
                /* Error */
                maybefetchagain('fetchfailed');
                return;
            }

            if (_this.devicecache[l.unique]) {
                device = clonedevice(_this.devicecache[l.unique].device);
            } else {
                device = undefined;
            }

            if (data.length === 0) {
                //l.listener(device, l.unique, l.type, l.name, undefined);
                return;
            }

            /* Sort by id */
            data = data.sort(function (a, b) {
                return a.id - b.id;
            });

            var num = 0;
            for (d in data) {
                if (num < l.seqnosskipped) {
                    /* Already seen */
                } else if (l.options.startid && data[d].id < l.options.startid) {
                    /* Duplicate, ignore */
                } else {
                    if (!l.timestamp || data[d].time > l.timestamp) {
                        l.timestamp = data[d].time;
                    }
                    if (!l.options.startid || data[d].id + 1 > l.options.startid) {
                        l.options.startid = data[d].id + 1;
                    }

                    l.listener(device, l.unique, l.type, l.name, data[d]);
                }
                num++;
            }
            l.seqnosskipped = 0;
            l.count += data.length;

            maybefetchagain('fetchchunk');
        });
    }

    function fetchhistoryifinteresting(thsq, device, unique, diff) {
        var type, name;

        for (type in diff) {
            if (diff.hasOwnProperty(type)) {
                for (name in diff[type]) {
                    if (diff[type].hasOwnProperty(name)) {
                        var key = unique + '-' + type + '-' + name;
                        if (thsq.historylisteners[key]) {
                            var seqdiff = device[type][name].seq - thsq.historylisteners[key].seq;
                            if (seqdiff === 0) {
                                /* We were triggered, but there was no new data. So we'll do nothing. */
                            } else if (seqdiff === 1) {
                                /* We received one more item, so we don't need to fetch any history */
                                device[type][name].timestamp = device[type][name].time;
                                thsq.historylisteners[key].map(function(l) {
                                    l.listener(device, l.unique, l.type, l.name, device[type][name]);
                                    if (l.seqnosskipped === undefined) {
                                        l.seqnosskipped = 0;
                                    }
                                    l.seqnosskipped++;
                                });
                            } else {
                                /* We need to fetch the missing history between the previous and this itme */
                                thsq.historylisteners[key].map(function(l) {
                                    fetchpartialhistory(thsq, l);
                                });
                            }
                            if (thsq.historylisteners[key]) {
                                thsq.historylisteners[key].seq = device[type][name].seq;
                            }
                        }
                    }
                }
            }
        }
    }

    function update(thsq, message) {
        if (message) {
            var i, data, deviceid, unique, device, diff, delta;
            try {
                if (message.type === 'message') {
                    data = JSON.parse(message.data);
                    thsq.stats.messagedata += message.data.length;
                } else if (message.type === 'utf8') {
                    data = JSON.parse(message.utf8Data);
                    thsq.stats.messagedata += message.utf8Data.length;
                } else {
                    return;
                }
            } catch (error) {
                thsq.disconnect();
                return;
            }
            thsq.stats.messages++;
            for (i in data) {
                if (data.hasOwnProperty(i)) {
                    if (i === 'devices') {
                        for (deviceid in data.devices) {
                            if (data.devices.hasOwnProperty(deviceid)) {
                                device = data.devices[deviceid];
                                device.id = deviceid;
                                unique = deviceunique(device);
                                diff = updateDeviceCache(thsq, unique, device);
                                thsq.emit('device-updated', device, unique, diff);
                                thsq.claimeduniqueids[unique] = deviceid;
                                fetchhistoryifinteresting(thsq, device, unique, diff);
                            }
                        }
                    } else if (i === 'delta') {
                        for (deviceid in data.delta) {
                            if (data.delta.hasOwnProperty(deviceid)) {
                                delta = data.delta[deviceid];
                                convertToUniqueFromId(thsq, deviceid, function (unique) {
                                    diff = addDeltaToDeviceCache(thsq, unique, delta);
                                    device = clonedevice(thsq.devicecache[unique].device);
                                    device.id = deviceid;
                                    unique = deviceunique(device);
                                    if (delta.d === undefined) {
                                        delta.d = {};
                                    }
                                    if (delta.s === undefined) {
                                        delta.s = {};
                                    }
                                    if (delta.meta === undefined) {
                                        delta.meta = {};
                                    }
                                    thsq.emit('device-updated', device, unique, diff);
                                    thsq.claimeduniqueids[unique] = deviceid;
                                    fetchhistoryifinteresting(thsq, device, unique, delta);
                                });
                            }
                        }
                    } else if (i === 'events') {
                        for (deviceid in data.events) {
                            if (data.events.hasOwnProperty(deviceid)) {
                                if (data.events[deviceid].type === 'device-removed') {
                                    convertToUniqueFromId(thsq, deviceid, function (unique) {
                                        delete thsq.claimeduniqueids[unique];
                                        updateDeviceCache(thsq, unique);
                                        thsq.emit('device-removed', unique);
                                    });
                                } else if (data.events[deviceid].type === 'device-claimed') {
                                    thsq.claimeduniqueids[deviceunique(data.events[deviceid].device)] = deviceid;
                                    addDeltaToDeviceCache(thsq, deviceunique(data.events[deviceid].device),
                                        data.events[deviceid].device);
                                    thsq.emit('device-claimed', data.events[deviceid].device, deviceunique(data.events[deviceid].device));

                                } else if (data.events[deviceid].type === 'user-event') {
                                    thsq.emit('user-event', data.events[deviceid]);
                                }
                            }
                        }
                    } else {
                    }
                }
            }
        }
    }
    var inittimeout;
    Thsq.prototype.init = function (options, callback) {
        var _this = this;

        if (_this.exiting) {
            /* We are currently executing an exit(), so we wait until we are properly exited before we try to init. */
            clearTimeout(inittimeout);
            inittimeout = setTimeout(function () {
                _this.init(options, callback);
            }, 1000);
            return;
        }

        if (typeof options === 'function') {
            callback = options;
            options = undefined;
        }

        if (!options) {
            options = {};
        }

        if (!_this.historylisteners) {
            _this.historylisteners = {}; /* indexed by unique-type-name */
        }

        if (options.request) {
            /* Overriding request with custom configuration */
            request = options.request;
        }

        _this.exit();
        _this.exited = false;
        _this.options = options;

        this.setServer();

        function deleteconn() {
            var wasconnected = _this.connected;
            _this.connected = false;
            _this.connecting = false;
            if (nodejs) {
                if (_this.connection) {
                    _this.connection.close();
                    if (typeof _this.connection.removeAllListeners === 'function') {
                        _this.connection.removeAllListeners();
                    }
                    delete _this.connection;
                }
                if (_this.socket) {
                    _this.socket.abort();
                    if (typeof _this.socket.removeAllListeners === 'function') {
                        _this.socket.removeAllListeners();
                    }
                    delete _this.socket;
                }
                if (wasconnected) {
                    _this.emit('server-disconnected');
                }
            } else if (browser) {
                if (_this.connection) {
                    delete _this.connection.onopen;
                    delete _this.connection.onerror;
                    delete _this.connection.onclose;
                    delete _this.connection.onmessage;
                    _this.connection.close();
                    delete _this.connection;
                }
                if (wasconnected) {
                    _this.emit('server-disconnected');
                }
            }
        }
        function doping() {
            if (nodejs) {
                if (_this.connection) {
                    _this.connection.ping();
                }
            }
        }

        function doconnect() {
            var protocol, host, parameter;

            if (_this.options.noupdates === true) {
                return;
            }

            if (_this.connected || _this.connecting || _this.exited || _this.paused) {
                return;
            }
            try {
                _this.connecting = true;

                var port = '';
                if ((_this.host && _this.host.split('//')[0] === 'https:') ||
                       (typeof location !== 'undefined' && location.protocol === 'https:')) {
                    protocol = 'wss://';
                    if (typeof location !== 'undefined' && location.port && location.port !== 443) {
                        port = ':' + location.port;
                    }
                } else {
                    protocol = 'ws://';
                    if (typeof location !== 'undefined' && location.port && location.port !== 80) {
                        port = ':' + location.port;
                    }
                }

                if (_this.host) {
                    host = _this.host.split('//')[1];
                } else if (typeof location !== 'undefined') {
                    host = location.hostname;
                }

                if (host.split(':').length === 2) {
                    port = '';
                }
                if (_this.app) {
                    parameter = _this.app;
                } else {
                    parameter = '';
                }

                var sendconnectionopions = function () {
                    let thsq;
                    if (typeof npm_pkg_version !== 'undefined') {
                        thsq = npm_pkg_version;
                    } else {
                        thsq = 'browser';
                    }

                    _this.connection.send(JSON.stringify({
                        delta: true,
                        devicedatafilter: _this.options.devicedatafilter,
                        throttletime: _this.options.throttletime,
                        lastseen: _this.lastseen,
                        thsq: thsq,
                        application: _this.options.application
                    }));
                };

                var onopen = function () {
                    _this.connected = true;
                    _this.connecting = false;
                    _this.emit('server-connected');
                    sendconnectionopions();
                };

                var onerror = function (error) {
                    deleteconn();
                };

                var onclose = function () {
                    deleteconn();
                };

                var onmessage = function (message) {
                    update(_this, message);
                };

                if (nodejs) {
                    if (_this.socket && typeof _this.socket.removeAllListeners === 'function') {
                        _this.socket.removeAllListeners();
                    }
                    _this.socket = new WebSocketClient();
                    _this.socket.on('connectFailed', function(error) {
                        //console.log('Connect Error '  + error.toString());
                        deleteconn();
                    });

                    _this.socket.on('connect', function(connection) {
                        _this.connection = connection;
                        connection.on('error', onerror);
                        connection.on('close', onclose);
                        connection.on('message', onmessage);
                        onopen();
                    });

                    var cookie = jar.getCookieString(_this.host);
                    if (!cookie) {
                        _this.session();
                    } else {
                        _this.socket.connect(protocol + host + '/0/session/' + parameter, 'thsq-user', null,
                                             { Cookie: cookie });
                    }
                } else if (browser) {
                    _this.connection = new WebSocket(protocol + host + port + '/0/session/' + parameter, 'thsq-user');

                    _this.connection.onopen = onopen;
                    _this.connection.onerror = onerror;
                    _this.connection.onclose = onclose;
                    _this.connection.onmessage = onmessage;
                }
            } catch (e) {
                console.log('doconnect error', e);
                deleteconn();
            }
        }

        this.disconnect = deleteconn;
        this.reconnect = doconnect;

        const startup = function () {
            _this.session(function (s) {
                if (s === 'user-ok') {
                    if (typeof callback === 'function') {
                        var devices, id, deviceid;
                        devices = {};
                        for (id in _this.devicecache) {
                            deviceid = deviceunique(_this.devicecache[id].device);
                            devices[deviceid] = _this.devicecache[id].device;
                        }
                        callback(devices);
                    }
                    doconnect();
                } else {
                    if (typeof callback === 'function') {
                        callback(undefined);
                    }
                    doconnect();
                }
                _this.inited = true;
                callinitcallbacks(_this);
            });
        };

        const readcachefile = function (callback) {
            /* If we have no devices in the device cache, we read the cache
               file. If we have devices in the cache, we are initializing
               ourselves after pausing, so we don't need to read the cache.
               If so, we just continue with the startup process. */
            if (Object.keys(_this.devicecache).length === 0) {
                fs.readFile(options.cachefile, function (err, jsondata) {
                    let data;
                    try {
                        data = JSON.parse(jsondata);
                    } catch (e) {
                        data = {};
                    }
                    let identifier = thsq.getServer() + '-' + thsq.getApp() + '-' + thsq.options.token;
                    if (data[identifier] && data[identifier].devicecache && data[identifier].claimeduniqueids && data[identifier].lastseen) {
                        thsq.devicecache = data[identifier].devicecache;
                        thsq.claimeduniqueids = data[identifier].claimeduniqueids;
                        thsq.lastseen = data[identifier].lastseen;
                    }
                    callback();
                });
            } else {
                callback();
            }
        };

        const readbrowsercache = function (callback) {
            /* If we have no devices in the device cache, we read the browser
               cache. If we have devices in the cache, we are initializing
               ourselves after pausing, so we don't need to read the cache.
               If so, we just continue with the startup process. */
            if (Object.keys(_this.devicecache).length === 0) {
                let jsondata = localStorage['devicecache'];
                let data;
                try {
                    data = JSON.parse(jsondata);
                } catch (e) {
                    data = {};
                }
                if (data.devicecache && data.claimeduniqueids && data.lastseen &&
                    Object.keys(data.devicecache).length > 0 &&
                    Object.keys(data.claimeduniqueids).length) {
                    thsq.devicecache = data.devicecache;
                    thsq.claimeduniqueids = data.claimeduniqueids;
                    thsq.lastseen = data.lastseen;
                    console.log('Read ' + Object.keys(data.devicecache).length + ' devices from cache, last seen on', new Date(data.lastseen));
                    callback();
                    return;
                }
            }
            thsq.devicecache = {};
            thsq.claimeduniqueids = {};
            callback();
        };


        if (options.cachefile) {
            readcachefile(startup);
        } else if (options.browsercache) {
            readbrowsercache(startup);
        } else {
            startup();
        }

        this.scaninterval = setInterval(function () {
            _this.scanBLE(10000);
        }, 10500);

        this.reconnecttimer = setInterval(function () {
            doconnect();
        }, 10000);

        this.pingtimer = setInterval(function () {
            doping();
        }, 60000);

        this.periodic = setInterval(function () {
            for (var unique in _this.nearbydevices) {
                if (_this.nearbydevices.hasOwnProperty(unique)) {
                    if (_this.nearbydevices[unique].time < Date.now() - 20000) {
                        var device = getDeviceFromCache(_this, unique);
                        if (deviceunique(device)) {
                            _this.emit('device-nearby-gone', device, unique);
                        }
                        delete _this.nearbydevices[unique];
                    }
                }
            }

            /* Write to browser cache every minute */
            if (thsq.options.browsercache &&
                Object.keys(thsq.devicecache).length > 0 &&
                (localStorage['devicecache-lastseen'] === undefined || localStorage['devicecache-lastseen'] < Date.now() - 60 * 1000)) {
                let cache = {
                    devicecache: clone(thsq.devicecache),
                    claimeduniqueids: clone(thsq.claimeduniqueids),
                    lastseen: lastupdate(thsq.devicecache),
                };
                let jsoncache = JSON.stringify(cache);
                try {
                    localStorage['devicecache'] = jsoncache;
                    localStorage['devicecache-lastseen'] = Date.now();
                } catch (e) {
                    if (e.name === 'QuotaExceededError') {
                        /* Too much data to store */
                        delete localStorage['devicecache'];
                        localStorage['devicecache-lastseen'] = Date.now();
                    }
                }
            }

        }, 1000);
    };

    Thsq.prototype.destroy = function (done) {
        this.removeAllListeners();
        this.exit(done);
    };

    Thsq.prototype.setDeviceDataFilter = function (filter) {
        if (this.options) {
            if (filter && filter !== this.options.devicedatafilter) {
                this.options.devicedatafilter = filter;
                this.disconnect();
                this.reconnect();
            } if (!filter && this.options.devicedatafilter) {
                delete this.options.devicedatafilter;
                this.disconnect();
                this.reconnect();
            }
        }
    };

    function lastupdate(devicecache) {
        let time = 0;
        for (let u in devicecache) {
            for (let t in devicecache[u].device) {
                for (let v in devicecache[u].device[t]) {
                    if (devicecache[u].device[t][v] && devicecache[u].device[t][v].time > time) {
                        time = devicecache[u].device[t][v].time;
                    }
                }
            }
        }
        return time;
    }

    function exit(thsq, callback) {
        let cachefile = thsq.options.cachefile;
        if (cachefile !== undefined) {
            let filename = thsq.options.cachefile;
            let identifier = thsq.getServer() + '-' + thsq.getApp() + '-' + thsq.options.token;
            let cache = {
                devicecache: clone(thsq.devicecache),
                claimeduniqueids: clone(thsq.claimeduniqueids),
                lastseen: lastupdate(thsq.devicecache),
            };
            fs.readFile(thsq.options.cachefile, function (data) {
                let cachedata = {};
                if (data && data.length > 0) {
                    try {
                        cachedata = JSON.parse(data);
                    } catch (e) {
                        console.log('error parsing cache file ' + thsq.options.cachefile + ':', e);
                    }
                }
                cachedata[identifier] = cache;
                fs.writeFile(filename, JSON.stringify(cachedata), callback);
            });
        }
        thsq.emit('exit');
        for (var unique in thsq.nearbydevices) {
            if (thsq.nearbydevices.hasOwnProperty(unique)) {
                var device = getDeviceFromCache(thsq, unique);
                if (deviceunique(device)) {
                    thsq.emit('device-nearby-gone', device, unique);
                }
                delete thsq.nearbydevices[unique];
            }
        }
        thsq.options = {};
        thsq.fetchqueues = {};
        thsq.fetching = {};
        thsq.fetchingnow = {};
        thsq.fetchingtimeout = {};
        thsq.usercache = {};
        thsq.devicecache = {};
        thsq.devicesfetching = {};
        thsq.claimeduniqueids = {};
        thsq.deviceuniqueids = {};
        thsq.nearbydevices = {};
        thsq.nearbyerrordevices = {};
        thsq.nearbydevicesrssi = {};
        thsq.stats = { messages: 0, messagedata: 0 };
        clearInterval(thsq.pingtimer);
        clearInterval(thsq.reconnecttimer);
        clearInterval(thsq.scaninterval);
        clearInterval(thsq.periodic);
        var queue;
        for (queue in thsq.fetchingtimeout) {
            if (thsq.fetchingtimeout[queue]) {
                clearTimeout(thsq.fetchingtimeout[queue]);
            }
        }
        if (thsq.sendudpinterval) {
            clearInterval(thsq.sendudpinterval);
        }
        if (typeof thsq.disconnect === 'function') {
            thsq.disconnect();
        }
        thsq.exited = true;
        thsq.inited = false;

        /* If options.cachefile is set, we call the callback after writing the
           cache to the filesystem. Otherwise we call the callback here. */
        if (cachefile === undefined) {
            callback();
        }
    }

    var exittimeout;

    Thsq.prototype.exit = function (callback) {
        var _this = this;

        if (_this.exited) {
            return;
        }

        /* Call any init callbacks that may be needed to be called before we exit */
        callinitcallbacks(_this);

        _this.exiting = true;

        clearTimeout(exittimeout);

        /* If there are no outstanding requests, then exit immediately. Otherwise, wait until there are no outstanding requests, then exit */

        var fetching = false;
        for (var q in _this.fetching) {
            fetching = fetching || _this.fetching[q];
        }

        if (!fetching) {
            exit(_this, function () {
                _this.exiting = false;
                if (typeof callback === 'function') {
                    callback();
                }
            });
        } else {
            exittimeout = setTimeout(function () {
                _this.exit(callback);
            }, 1000);
        }
    };

    Thsq.prototype.pause = function () {
        if (!this.paused) {
            this.paused = true;
            this.lastseen = Date.now();
            this.disconnect();
        }
    };

    Thsq.prototype.unpause = function () {
        this.paused = false;
        this.reconnect();
    };

    checkLoaded = function () {
        return true;
    };

    fcmMessage = function (token, thsqid) {
        setTimeout(function () {
            thsq.fcmToken = token;

            /*console.log("fcmMessage: " + thsqid);*/
            if (typeof thsq.fcmCallback === 'function') {
                thsq.fcmCallback(token, thsqid);
            }
        }, 0);
    };

    fcmToken = function (token) {
        setTimeout(function () {
            thsq.fcmToken = token;

            /*console.log("token: " + token);*/
            if (typeof thsq.fcmCallback === 'function') {
                thsq.fcmCallback(token, null);
            }
        }, 0);
    };

    hostMessage = function (str) {
        setTimeout(function () {
            if (thsq && (str === 'background' || str === 'stopped')) {
                thsq.dontPing = true;
                thsq.disconnect();
            } else {
                thsq.dontPing = false;
                thsq.reconnect();
                thsq.sendPing();
            }
        }, 0);
    };

    udpCallback = function (str) {
        setTimeout(function () {
            var array, i, auth;

            auth = str.toLowerCase();

            if (auth.length === 32) {
                addNearbyDevice(thsq, auth);
            } else if (auth.length === 64) {
                // Assume this is a hex-encoded 16 byte auth code
                array = new Uint8Array(auth.length / 2);
                for (i = 0; i < auth.length; i += 2) {
                    array[i / 2] = parseInt(auth.charAt(i), 16) * 16 +
                    parseInt(auth.charAt(i + 1), 16);
                }
                addNearbyDevice(thsq, String.fromCharCode.apply(null, array).toLowerCase());
            }
        }, 0);
    };

    bleCallback = function (str, rssi) {
        function hextostring(hex) {
            var i, string, val;

            string = '';
            for (i = 0; i < hex.length; i += 2) {
                val = parseInt(hex.substring(i, i + 2), 16);
                if (val === 0) {
                    return string;
                }
                string += String.fromCharCode(val);
            }
            return string;
        }
        setTimeout(function () {
            var auth, data, id, platform, state;

            if (str.indexOf('020106050954485351') === 0 &&
                (str.indexOf('FFFFFF') === 20 || str.indexOf('FF9103') === 20)) {
                data = str.substring(26);
            } else if (str.indexOf('ffff') === 0 || str.indexOf('9103') === 0) {
                data = str.substring(4);
            }
            if (data) {
                if (data.length === 32) {
                    auth = data.substring(0, 32).toLowerCase();
                    addNearbyDevice(thsq, auth, state, platform, rssi);
                    thsq.emit('ble-received', data);
                } else {
                    if (data.indexOf('1') === 0) {
                        auth = data.substring(2, 2 + 32).toLowerCase();
                        state = parseInt(data.substring(1, 2), 16);
                        addNearbyDevice(thsq, auth, state, platform, rssi);
                    } else if (data.indexOf('0') === 0) {
                        state = parseInt(data.substring(1, 2), 16);
                        id = data.substring(2, 2 + 4).toLowerCase();
                        platform = hextostring(data.substring(6).toLowerCase());
                        addNearbyDevice(thsq, id, state, platform, rssi);
                    } else if (data.indexOf('2') === 0) {
                        state = parseInt(data.substring(1, 2), 16);
                        id = data.substring(2, 2 + 4).toLowerCase();
                        addNearbyDevice(thsq, id, state, undefined, undefined, data.substring(6, 6 + 16));
                    }
                    thsq.emit('ble-received', data);
                }
            }
        }, 0);
    };

    var blewasenabled = true;
    bleEnabled = function (ble, location) {
        setTimeout(function () {
            var locationok = false;
            var bleok = false;

            if (ble === 'poweredOn') {
                bleok = true;
            } else if (String(ble) === String(true)) {
                bleok = true;
            }

            if (String(location) === String(true)) {
                locationok = true;
            }

            if (bleok === true && locationok === true) {
                if (blewasenabled === false) {
                    blewasenabled = true;
                    thsq.emit('ble-enabled');
                }
            } else {
                if (blewasenabled === true) {
                    blewasenabled = false;
                    thsq.emit('ble-disabled', bleok, locationok);
                }
            }
        }, 0);
    };

    /* Code for BLE transfer data reception and transmission below */
    Thsq.prototype.enableBLETransfer = function (key) {

        var _this = this;

        this.bletransferreceiverstate = {
            iv: undefined,
            nextseqno: 0,
            data: undefined,
            aescbc: undefined,
            startime: undefined,
            datalen: 0
        };

        if (typeof key === 'string') {
            /* The key is provided in hex format, so we convert it to an array */
            key = key.match(/.{1,2}/g).map(byte => parseInt(byte, 16));
        }


        function handleincomingbledata(data) {
            /* Handle a chunk of incoming BLE data. If we don't have an ongoing transfer, we check if we have received an initial packet. Those are identified by a 0 in the second byte, and that the first byte is the SHA256 sum of the packet, with the checksum byte set to 0. */

            if (data.length !== 16) {
                /* Too short packet */
                return;
            }

            /* Helper function for computing the checksum byte of a chunk. The checksum is the first byte of the HMAC-SHA256 of the SHA256 hash of the chunk and the key. */
            function checksum(chunk) {
                var hash = sha256.array(chunk);
                var hmac = sha256.hmac(key, hash);
                return Number.parseInt(hmac.substr(0, 2), 16);
            }

            /* Helper function for decrypting an encrypted packet */
            function decryptpacket(data) {
                return _this.bletransferreceiverstate.aescbc.decrypt(data);
            }

            /* Helper function for sending an ack for a data chunk */
            function sendack(data) {
                var ackdata = sha256.hmac(key, data);
                thsq.sendBLE(ackdata.slice(0, 16));
            }

            /* Helper function for converting from Uint8Array to string */
            function uint8arraytostring(data) {
                let decoder;

                if (nodejs) {
                    const TextDecoder = require('util').TextDecoder;
                    decoder = new TextDecoder();
                } else {
                    decoder = new TextDecoder();
                }
                const str = decoder.decode(data);
                return str;
            }

            /* Helper function that checks the integrity of an uncrypted packet */
            function checkunencryptedpacket(data, seqno) {
                /* Check if this is a good packet */

                if ((data[1] & 0x0f) !== (seqno & 0x0f)) {
                    /* This was not the packet we expected */
                    return false;
                }

                /* Copy the data into a new array. */
                var datacopy = new Uint8Array(16);
                datacopy.set(data);

                /* Set the first byte to 0, so that our checksum computation is made over the same data as was originally done */
                datacopy[0] = 0;

                var sum = checksum(datacopy);


                if (sum !== data[0]) {
                    /* This was not a correct packet. */
                    return false;
                }

                return true;
            }

            /* Helper function that determines if this is the final chunk */
            function isfinalchunk(data) {
                return (data[1] & 0x10) === 0x10;
            }


            /* If we have a receiver state, but it is too old, we start anew */
            if (_this.bletransferreceiverstate.time < Date.now() - 20000) {
                _this.bletransferreceiverstate.nextseqno = 0;
            }
            _this.bletransferreceiverstate.time = Date.now();
            if (_this.bletransferreceiverstate.timeout) {
                clearTimeout(_this.bletransferreceiverstate.timeout);
            }
            _this.bletransferreceiverstate.timeout = setTimeout(function () {
                thsq.emit('ble-transfer-timedout');
                delete _this.bletransferreceiverstate.timeout;
            }, 10000);

            if (_this.bletransferreceiverstate.nextseqno === 0) {
                if (checkunencryptedpacket(data, 0)) {
                    _this.bletransferreceiverstate.nextseqno = 1;
                    _this.bletransferreceiverstate.aescbc = new aesjs.ModeOfOperation.cbc(key, data);
                    _this.bletransferreceiverstate.data = [];
                    sendack(data);
                }
            } else {
                var unencrypted = decryptpacket(data);
                if (checkunencryptedpacket(unencrypted, _this.bletransferreceiverstate.nextseqno)) {

                    if (_this.bletransferreceiverstate.nextseqno === 1) {
                        /* Discard the first chunk, which consists of random data, except for the last two bytes, that contain the length of the message */
                        _this.bletransferreceiverstate.datalen = (unencrypted[14] << 8) + unencrypted[15];

                        thsq.emit('ble-transfer-started');

                    } else {
                        _this.bletransferreceiverstate.data.push(unencrypted.slice(2, 16));
                    }
                    sendack(unencrypted);

                    if (isfinalchunk(unencrypted)) {
                        var string = '';
                        var data = new Uint8Array(_this.bletransferreceiverstate.data.length * 14);
                        var i, j;
                        for (i = 0; i < _this.bletransferreceiverstate.data.length; i += 1) {
                            for (j = 0; j < _this.bletransferreceiverstate.data[i].length; j += 1) {
                                data[j + i * 14] = _this.bletransferreceiverstate.data[i][j];
                            }
                        }
                        /* Trim the data to the length that we have received */
                        var trimmed = data.slice(0, _this.bletransferreceiverstate.datalen);
                        thsq.emit('ble-transfer-received', trimmed, uint8arraytostring(trimmed));
                        _this.bletransferreceiverstate.nextseqno = 0;

                        clearTimeout(_this.bletransferreceiverstate.timeout);
                        delete _this.bletransferreceiverstate.timeout;

                        /* Stop sending beacons, but give the final ack some time to linger */
                        setTimeout(function () {
                            thsq.sendBLE('');
                        }, 8000);
                    } else {
                        _this.bletransferreceiverstate.nextseqno++;
                    }

                }
            }
        }

        var lastpacket;
        thsq.on('ble-received', function (hex) {
            if (hex && hex !== lastpacket) {
                var arraydata = new Uint8Array(hex.match(/.{1,2}/g).map(byte => parseInt(byte, 16)));
                handleincomingbledata(arraydata);
                lastpacket = hex;
            }
        });
    };

    Thsq.prototype.sendBLETransfer = function (options) {

        /* Split the data into 14-byte chunks, send each chunk and wait for the ack for that chunk. Once all chunks have been acked, call the callback.

        The signature of the callback is callback(err);

        The signature of the progress callback is progress(currentchunk / totalchunks);
        */

        var key = options.key;
        var data = options.data;
        var progress = options.progress;
        var callback = options.callback;

        if (typeof key === 'string') {
            /* The key is provided in hex format, so we convert it to an array */
            key = key.match(/.{1,2}/g).map(byte => parseInt(byte, 16));
        }

        const chunklen = 14;

        /* Helper function for computing the checksum byte of a chunk. The checksum is the first byte of the HMAC-SHA256 of the SHA256 hash of the chunk and the key. */
        function checksum(chunk) {
            var hash = sha256.array(chunk);
            var hmac = sha256.hmac(key, hash);
            return Number.parseInt(hmac.substr(0, 2), 16);
        }

        /* Helper function for splitting data into chunks of a certain size and prepend the data with the chunk header. The chunk header consists of two bytes: the first is a checksum of the entire chunk, which is computed with the checksum byte set to 0. The checksum is the first byte of a sha256 hash of the chunk. The second consists of flags and a sequence number. */
        function chunkify(data, size) {
            var i;
            var result = [];
            for (i = 0; i < data.length; i += size) {
                var lastchunk;
                var chunk = new Uint8Array(16);
                var datachunk = data.slice(i, i + size);

                /* Copy data into chunk at offset 2 */
                chunk.set(datachunk, 2);

                if (i >= data.length - size) {
                    lastchunk = true;
                } else {
                    lastchunk = false;
                }

                chunk[1] = ((i / size) & 0x0f) | (lastchunk? 0x10: 0x00);
                chunk[0] = checksum(chunk);

                result.push(chunk);
            }

            var numchunks = result.length;
            var resultbuf = new Uint8Array(16 * numchunks);
            for (i = 0; i < numchunks; i += 1) {
                resultbuf.set(result[i], i * 16);
            }
            return resultbuf;
        }

        /* Helper function for converting from string to Uint8Array */
        function stringtouint8array(str) {
            let encoder;
            if (nodejs) {
                const TextEncoder = require('util').TextEncoder;
                encoder = new TextEncoder();
            } else {
                encoder = new TextEncoder();
            }
            const uint8array = encoder.encode(str);
            return uint8array;
        }

        if (typeof data === 'string') {
            data = stringtouint8array(data);
        }

        /* The first data chunk is a 14-byte random initialization vector. */
        var iv = new Uint8Array(chunklen);
        var i;
        for (i = 0; i < iv.length; i += 1) {
            iv[i] = Math.floor(Math.random() * 256);
        }

        /* The second data chunk is a 14-byte random chunk, to avoid an attacker guessing the first block. */
        var firstchunk = new Uint8Array(chunklen);
        for (i = 0; i < firstchunk.length; i += 1) {
            firstchunk[i] = Math.floor(Math.random() * 256);
        }

        /* The last byte of the first encrypted chunk is the total length of the final message */
        firstchunk[12] = data.length >> 8;
        firstchunk[13] = data.length & 0xff;

        /* Combine the iv and the data into a single array. */
        var dataarray = new Uint8Array(iv.length + firstchunk.length + data.length);
        dataarray.set(iv);
        dataarray.set(firstchunk, iv.length);
        dataarray.set(data, iv.length + firstchunk.length);

        var allchunks = chunkify(dataarray, chunklen);

        /* Encrypt the chunks */

        /* The initialization vector is the first 16 bytes from the chunkified chunks array. */
        var aesiv = [];
        for (i = 0; i < 16; i += 1) {
            aesiv[i] = allchunks[i];
        }

        /* Use AES CBC, with the key from above, and the first chunk as IV */
        var aescbc = new aesjs.ModeOfOperation.cbc(key, aesiv);

        /* Convert the chunks that we should send into an array that the aes-js code understands. Make sure to not include the first chunk in the data to be encrypted, as it is the IV, which is sent in plain text before the encrypted data. */
        var unencrypteddata = new Array(allchunks.length - 16);
        for (i = 16; i < allchunks.length; i += 1) {
            unencrypteddata[i - 16] = allchunks[i];
        }
        /* Encrypt the bytes */
        var encrypteddata = aescbc.encrypt(unencrypteddata);

        var blechunksenderstate = {
            encrypteddata: undefined,
            seqno: 0,
        };

        blechunksenderstate.seqno = 0;
        blechunksenderstate.encrypteddata = encrypteddata;
        blechunksenderstate.unencrypteddata = allchunks;


        /* We call this function to clean up and call the callback */
        function done(err) {
            /* We are done - unregister our event handler */
            thsq.off('ble-received', handlebleacks);

            /* Unregister timeout */
            clearTimeout(blechunksenderstate.timeouthandler);

            /* Stop sending beacons */
            thsq.sendBLE('');

            /* Call our callback */
            if (typeof callback === 'function') {
                callback(err);
            }
        }

        /* This function handles the timeout */
        function timeouthandler() {
            done('timeout');
        }
        /* This function is called whenever a BLE packet is received. */
        var lastbledata;
        function handlebleacks(bledata) {
            if (bledata) {
                bledata = bledata.toLowerCase();
            }
            if (lastbledata === bledata) {
                return;
            }

            lastbledata = bledata;

            if (blechunksenderstate.encrypteddata === undefined ||
                    blechunksenderstate.unencrypteddata === undefined) {
                /* We haven't begun sending data yet, so we don't need to process potential acks. */
                return;
            }

            /* Check if the ack matches the data that we previously sent. */

            /* Pick out the data that it is supposed to match. This is 16 bytes from the unencrypted array, indexed by the seqno that we previously sent. */

            var datatobeacked = new Uint8Array(16);
            var i;
            for (i = 0; i < 16; i += 1) {
                datatobeacked[i] = blechunksenderstate.unencrypteddata[blechunksenderstate.seqno * 16 + i];
            }
            var ackdata = sha256.hmac(key, datatobeacked).slice(0, 16);
            if (ackdata === bledata) {

                /* We received an ack for the last data that we sent, so we can send the next chunk */
                blechunksenderstate.seqno += 1;

                /* Report our progress */
                if (typeof progress === 'function') {
                    var lastseqno = encrypteddata.length / 16 + 1;
                    progress(blechunksenderstate.seqno / lastseqno);
                }


                /* The first chunk that we sent was not part of the encrypted data (it was the unencrypted initial iv packet), so we need to subtract 1 from the seqno to create the correct offset. */
                var offset = (blechunksenderstate.seqno - 1) * 16;

                if (offset < encrypteddata.length) {
                    thsq.sendBLE(blechunksenderstate.encrypteddata.slice(offset, offset + 16));


                    /* Update the timeout */
                    clearTimeout(blechunksenderstate.timeouthandler);
                    blechunksenderstate.timeouthandler = setTimeout(timeouthandler, 10000);
                } else {
                    done();
                }
            }
        }

        thsq.on('ble-received', handlebleacks);

        /* Send the initial packet and wait for the correct ack. Transmission is done via the handlebleacks() function above. */
        if (!thsq.sendBLE(aesiv)) {
            setTimeout(function () {
                done('Could not send Bluetooth data');
            }, 100);
            return;
        }

        /* Set a timeout of 10 seconds to send the data */
        blechunksenderstate.timeouthandler = setTimeout(timeouthandler, 10000);
    };

    thsq = new Thsq();

})();

if (typeof exports !== 'undefined') {
    module.exports = thsq;

    if (typeof global.it === 'function') {
        // Only used during automated testing
        module.exports.bleCallback = bleCallback;
    }
}
